/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package examples.invoice.applications;

import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.jorm.lib.MapperJCA;
import org.objectweb.perseus.connector.ra.fos.FosAttributeControler;
import org.objectweb.perseus.connector.ra.fos.FosConnectionFactory;
import org.objectweb.perseus.connector.ra.fos.FosManagedConnectionFactory;
import org.objectweb.perseus.cache.api.CacheManager;
import org.objectweb.transaction.api.ConnectionFactoryWrapper;
import org.objectweb.transaction.ra.gen.FactoryImpl;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.HandlerFactory;
import org.objectweb.util.monolog.api.LevelFactory;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.util.monolog.file.monolog.PropertiesConfAccess;
import org.objectweb.util.monolog.wrapper.common.Configurable;

import java.io.FileInputStream;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Properties;
import javax.naming.Context;
import javax.resource.cci.ConnectionFactory;

/**
 * @author	P. Dechamboux
 */
public class JormInit {
	// JormInit related constants and variables
	public static final String LOGGER_FACTORY_CLASS_NAME
		= "loggerFactoryClassName";
	private static JormInit singleton = new JormInit();
    private static CacheManager cacheManager = new TrivialCache();
	protected LoggerFactory loggerFactory = null;
	protected Logger logger = null;
	// Mapper related constants and variables
	private HashMap mapperMap = new HashMap();
	// RA related constants and variables
    public static final String RESOURCE_JDBC_MCF_CLASSNAME
        = "jdbc.resource.mcf.classname";
    public static final String RESOURCE_JDBC_SUBMAPPER
        = "jdbc.submapper";
	public static final String RESOURCE_FOS_MCF_CLASSNAME
		= "fos.resource.mcf.classname";
	public static final String DEFAULT_FOS_URL = "fos.ConnectionURL";
	private HashMap raMap = new HashMap();
	private String jdbcMcfCN = null;
	private String fosMcfCN = null;
	private static final Boolean MAPPED = new Boolean(true);
	private HashMap mappedMap = new HashMap();

	/**
	 * Retrieves the JormInit singleton.
	 */
	public static JormInit getInstance() {
		return singleton;
	}

	/**
	 * Retrieves the LoggerFactory
	 */
	public LoggerFactory getLoggerFactory() {
		return loggerFactory;
	}

	/**
	 * Initializes the log system.
	 */
	public void initLogSystem(Properties prop) throws Exception {
		String a = prop.getProperty(Configurable.LOG_CONFIGURATION_FILE, null);
		if (a != null) {
			String b = prop.getProperty(
				Configurable.LOG_CONFIGURATION_FILE_USE_CLASSPATH, null);
			Properties p = new Properties();
			if (b == null && "true".equalsIgnoreCase(b)) {
				// Load from the classpath
				p.load(ClassLoader.getSystemResourceAsStream(a));
			} else {
				//Load from file system
				p.load(new FileInputStream(a));
			}
			// recall this method with the new properties
			initLogSystem(p);
		} else {
			a = prop.getProperty("log.config.classname", null);
			if (a == null) {
				throw new Exception(
					"Impossible to find the 'log.config.classname' property");
			}
			loggerFactory = (LoggerFactory) Class.forName(a).newInstance();
			PropertiesConfAccess.load(prop, loggerFactory,
									  (HandlerFactory) loggerFactory, (LevelFactory) loggerFactory);
			logger = loggerFactory.getLogger("examples.invoice.applications.JormInit");
			logger.log(BasicLevel.INFO, "Log system initialized.");
		}
	}

	/**
	 * Initializes the RA environment
	 */
	public void initResourceAdapter(Properties prop) throws Exception {
		// Initializes JDBC environment
		PMapper mapper;
		jdbcMcfCN = prop.getProperty(RESOURCE_JDBC_MCF_CLASSNAME);
		if (jdbcMcfCN == null)
			logger.log(BasicLevel.INFO,
					   "Cannot initialize JDBC RA environment: MCF=" + jdbcMcfCN);
		else {
			Object ra = Class.forName(jdbcMcfCN).newInstance();
			logger.log(BasicLevel.DEBUG, "JDBC Ra=" + ra);
            String submapper = prop.getProperty(RESOURCE_JDBC_SUBMAPPER);
			// configure it
			Context ctx = new org.objectweb.naming.ContextImpl();
			Enumeration e = prop.keys();
			while (e.hasMoreElements()) {
				String key = (String) e.nextElement();
				ctx.rebind(key, prop.getProperty(key, ""));
			}
			ctx.bind("org.objectweb.util.monolog.api.logger",
					 loggerFactory.getLogger("examples.invoice.ra.jdbc"));
			((FactoryImpl) ra).init(ctx);
			//Fetch the ConnectionFactory
			if (ra instanceof ConnectionFactory)
				raMap.put("jdbc.default", ra);
			else if (ra instanceof ConnectionFactoryWrapper)
				raMap.put("jdbc.default",
						  ((ConnectionFactoryWrapper) ra).getCCIConnectionFactory());
			else
				throw new Exception("Impossible to use this resource adapter:"
									+ "not a ConnectionFactory or a ConnecitonFactoryWrapper");
			logger.log(BasicLevel.INFO, "jdbc.default: connection factory=" + raMap.get("jdbc.default"));
            if (submapper.equals("generic")) {
                mapper = getMapper("jdbc.default", "rdb");
            } else {
			    mapper = getMapper("jdbc.default", "rdb." + submapper);
            }
			mapper.setConnectionFactory(raMap.get("jdbc.default"));
		}
		// Initializes FOS environment
		fosMcfCN = prop.getProperty(RESOURCE_FOS_MCF_CLASSNAME);
		if (fosMcfCN == null)
			logger.log(BasicLevel.INFO,
					   "Cannot initialize FOS RA environment: MCF=" + fosMcfCN);
		else {
			FosAttributeControler ra
				= (FosAttributeControler) Class.forName(fosMcfCN).newInstance();
			logger.log(BasicLevel.DEBUG, "FOS Ra=" + ra);
			// configure it
			ra.setConnectionURL(prop.getProperty(DEFAULT_FOS_URL));
			ra.setLoggerFactory(loggerFactory);
			ra.setInitializeAtStartUp(false);
			ra.setLoggerBaseName("examples.invoice.ra");
			((FosManagedConnectionFactory) ra).start();
			raMap.put("fos.default", ((FosManagedConnectionFactory) ra).createConnectionFactory());
			logger.log(BasicLevel.INFO, "fos.default: connection factory=" + raMap.get("fos.default"));
			mapper = getMapper("fos.default", "fos");
			mapper.setConnectionFactory(raMap.get("fos.default"));
		}
	}

	public PMapper getMapper(String name, String mappername) throws PException {
		PMapper mapper = (PMapper) mapperMap.get(name);
		if (mapper != null) {
			if (mappername == null)
				return mapper;
			if (mapper.getMapperName().equals(mappername))
				return mapper;
			return null;
		}
		if (mappername == null)
			return null;
		mapper = new MapperJCA();
		mapper.setMapperName(mappername);
		mapper.start();
		mapperMap.put(name, mapper);
		((Loggable) mapper).setLoggerFactory(loggerFactory);
		logger.log(BasicLevel.INFO, "Mapper " + mappername
									+ " instancied and initialised.");
		return mapper;
	}

	Iterator registeredMapper() {
		return mapperMap.keySet().iterator();
	}

	public ConnectionFactory getJdbcCF(String name, String url, String un, String pw) {
		ConnectionFactory cf = (ConnectionFactory) raMap.get(name);
		if (cf != null) {
			if (!(cf instanceof FosConnectionFactory))
				return cf;
			return null;
		}
		if ((url == null) || (un == null) || (pw == null))
			return null;
		// Complete code for CF creation
		return null;
	}

	public ConnectionFactory getFosCF(String name, String url) {
		ConnectionFactory cf = (ConnectionFactory) raMap.get(name);
		if (cf != null) {
			if (cf instanceof FosConnectionFactory)
				return cf;
			return null;
		}
		if (url == null)
			return null;
		// Complete code for CF creation
		return null;
	}

	Iterator registeredRA() {
		return raMap.keySet().iterator();
	}

	public boolean canMap(String name) throws PException {
		if (mappedMap.get(name) != null)
			return false;
		if (getMapper(name, null) == null)
			return false;
		if ((getJdbcCF(name, null, null, null) == null) &&
			(getFosCF(name, null) == null))
			return false;
		return true;
	}

	public void setMapped(String name) {
		mappedMap.put(name, MAPPED);
	}

	public boolean isMapped(String name) {
		return mappedMap.get(name) != null;
	}

    public CacheManager getCacheManager() {
        return cacheManager;
    }
}
