/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package examples.invoice.persistclass;

import org.objectweb.jorm.api.PAccessor;
import org.objectweb.jorm.api.PGenClassAccessor;
import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PIndexedElem;
import org.objectweb.jorm.naming.api.PName;

import java.io.Serializable;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.math.BigDecimal;

/**
 * @author	P. Dechamboux
 */
public class Invoice implements PAccessor, InvoiceAccessor {
	private int number;
	private PName addressPn;
	private Address address = null;
	private PName pusetPn;
	private Set puset = null;
	private SetAccessor pusetAcc;
	private PBinding pusetBinding = null;
	private PBinding pBinding;
	private InvoiceHelper invoiceHelper;

	public PBinding getPBinding() {
		return pBinding;
	}

	/**
	 * Implements lazy dereferencing of address field.
	 */
	private Address getAddress() throws PException {
		try {
			if (address == null)
				address = invoiceHelper.getAddressHelper().getAddress(
					pBinding.getPClassMapping().getPMapper(),
					addressPn);
		} catch (Exception e) {
			e.printStackTrace();
			throw new PException(e, "Cannot getAddress within Invoice!");
		}
		return address;
	}

	/**
	 * Implements lazy dereferencing of puset field.
	 */
	private Set getPuSet() throws PException {
		if (pusetPn == null)
			return null;
		if (pusetBinding == null) {
			pusetBinding = pBinding.getPClassMapping().getGenClassMapping("puset").createPBinding();
			pusetBinding.bind(pusetPn);
			puset = new HashSet();
			pusetAcc = new SetAccessor(puset);
			Object conn = null;
			try {
				conn = pBinding.getPClassMapping().getPMapper().getConnection();
				pusetBinding.read(conn, pusetAcc);
			} catch (PException pe) {
				throw pe;
			} catch (Exception e) {
				throw new PException(e, "Cannot getPuSet-1");
			} finally {
				try {
					pBinding.getPClassMapping().getPMapper().closeConnection(conn);
				} catch (Exception e) {
					throw new PException(e, "Cannot getPuSet-2");
				}
			}
		}
		return puset;
	}

	/**
	 * Creates a new Invoice.
	 */
	public Invoice(InvoiceHelper ih, Object conn, PBinding pb, int n,
				   PName ad, PName[] pus) throws PException {
		invoiceHelper = ih;
		pBinding = pb;
		number = n;
		addressPn = ad;
		try {
			pusetBinding = pBinding.getPClassMapping().getGenClassMapping("puset").createPBinding();
			pusetPn = pusetBinding.export(conn, new Integer(number));
			puset = new HashSet();
			pusetAcc = new SetAccessor(puset);
			for (int i = 0; i < pus.length; i++) {
				ProdUnits pu = invoiceHelper.getProdUnitsHelper().getProdUnits(
					pBinding.getPClassMapping().getPMapper(), pus[i]);
				puset.add(pu);
			}
			pusetBinding.write(conn, pusetAcc);
			pBinding.export(conn, new Integer(number));
			pBinding.write(conn, this);
		} catch (PException pe) {
			pe.getNestedException().printStackTrace();
			throw pe;
		} catch (Exception e) {
			e.printStackTrace();
			throw new PException(e, "Pb while creating an Invoice.");
		}
	}

	/**
	 * Creates an Invoice that is already persistent.
	 */
	public Invoice(InvoiceHelper ih, Object conn, PBinding pb) throws PException {
		invoiceHelper = ih;
		pBinding = pb;
		pb.read(conn, this);
	}

	/**
	 * Turns this persistent object into a non-persistent one as well as the
	 * referenced "puset".
	 */
	public void delete(Object conn) throws PException {
		if (pusetBinding == null) {
			pusetBinding = pBinding.getPClassMapping().getGenClassMapping("puset").createPBinding();
		}
		pusetBinding.unexport(conn);
		pusetBinding.write(conn, pusetAcc);
		pusetBinding = null;
		pBinding.unexport(conn);
		pBinding.write(conn, this);
		pBinding = null;
	}

	/**
	 * Yields a printable version a the full Address.
	 */
	public String toPrintable() throws PException {
		String res = "ID: [" + number + "]\n\t\tVALUE: [\n\t\t\tnumber: " + number + ",\n\t\t\taddress: [" + getAddress().toPrintable() + "],\n\t\t\tpuset: {";
		Set s = getPuSet();
		Iterator it = s.iterator();
		String sep = "\n\t\t\t\t";
		while (it.hasNext()) {
			ProdUnits pu = (ProdUnits) it.next();
			res += sep + pu.toPrintable();
			sep = ",\n\t\t\t\t";
		}
		res += "\n\t\t\t}\n\t\t]";
		return res;
	}

	// IMPLEMENTATION OF PAccessor INTERFACE

	public Object getMemoryInstance() {
		return this;
	}

	// IMPLEMENTATION OF AddressAccessor INTERFACE (generated by JORM)

	// Accessors to the name field
	public void paSetNumber(int val) throws PException {
		number = val;
	}

	public int paGetNumber() throws PException {
		return number;
	}

	// Accessors to the size field
	public void paSetAddress(PName val, Object conn) throws PException {
		addressPn = val;
	}

	public PName paGetAddress(Object conn) throws PException {
		return addressPn;
	}

	//Accessors to the puset field
	public void paSetPuset(PName val, Object connection) throws PException {
		pusetPn = val;
	}

	public PName paGetPuset(Object connection) throws PException {
		return pusetPn;
	}

	/**
	 * Provided as the PGenClassAccessor for accessing elements of the "puset"
	 * HashSet (that are ProdUnits objects) as PIndexedElem in order to be able
	 * to store them with JORM.
	 */
	class SetAccessor implements PGenClassAccessor {
		private Set setData;

		SetAccessor(Set sd) {
			setData = sd;
		}

		public void paAdd(PIndexedElem elem, Object conn) throws PException {
			try {
				ProdUnits pu = invoiceHelper.getProdUnitsHelper().getProdUnits(
					pBinding.getPClassMapping().getPMapper(), elem.pieGetRefElem());
				setData.add(pu);
			} catch (Exception e) {
				throw new PException(e, "Pb to Add in Invoice/SetAccessor.");
			}
		}

		public Object getMemoryInstance() {
			return setData;
		}

		public PIndexedElem createPIndexedElem() {
			return new SetIndexedElem();
		}

		public boolean paDeltaSupported() {
			return false;
		}

		public int paGetNbElem() {
			return setData.size();
		}

		public Iterator paIterator() {
			return new SetAccessorIterator(setData.iterator());
		}

		public void paSetNbElem(int nbelem) {
		}

		/**
		 * Provided to present elements of the set (ProdUnits) as PIndexedElem
		 * that manages a reference shown as a PName.
		 */
		class SetIndexedElem implements PIndexedElem {
			private final static String ERRNOINDEX = "Set generic class has no index for its elements.";
			private final static String ERRTYPEUNSUPP = "This is a set of reference (PName).";
			private byte status = PIndexedElem.ELEM_CREATED;
			private PName ref;

			public byte getElemStatus() {
				return status;
			}

			public boolean pieGetBooleanElem() throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public Boolean pieGetObooleanElem() throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public byte pieGetByteElem() throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public Byte pieGetObyteElem() throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public byte pieGetByteIndexField(String fn) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public Byte pieGetObyteIndexField(String fn) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public char pieGetCharElem() throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public Character pieGetOcharElem() throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public char pieGetCharIndexField(String fn) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public Character pieGetOcharIndexField(String fn) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public short pieGetShortElem() throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public Short pieGetOshortElem() throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public short pieGetShortIndexField(String fn) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public Short pieGetOshortIndexField(String fn) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public int pieGetIntElem() throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public Integer pieGetOintElem() throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public int pieGetIntIndexField(String fn) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public Integer pieGetOintIndexField(String fn) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public long pieGetLongElem() throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public Long pieGetOlongElem() throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public long pieGetLongIndexField(String fn) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public Long pieGetOlongIndexField(String fn) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public float pieGetFloatElem() throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public Float pieGetOfloatElem() throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public double pieGetDoubleElem() throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public Double pieGetOdoubleElem() throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public String pieGetStringElem() throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public String pieGetStringIndexField(String fn) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public BigDecimal pieGetBigDecimalElem() throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public Date pieGetDateElem() throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public Date pieGetDateIndexField(String fn) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public char[] pieGetCharArrayElem() throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public byte[] pieGetByteArrayElem() throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public Serializable pieGetSerializedElem() throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public PName pieGetRefElem() throws PException {
				return ref;
			}

			public void pieSetBooleanElem(boolean value) throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public void pieSetObooleanElem(Boolean value) throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public void pieSetByteElem(byte value) throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public void pieSetObyteElem(Byte value) throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public void pieSetByteIndexField(String fn, byte value) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public void pieSetObyteIndexField(String fn, Byte value) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public void pieSetCharElem(char value) throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public void pieSetOcharElem(Character value) throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public void pieSetCharIndexField(String fn, char value) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public void pieSetOcharIndexField(String fn, Character value) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public void pieSetShortElem(short value) throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public void pieSetOshortElem(Short value) throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public void pieSetShortIndexField(String fn, short value) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public void pieSetOshortIndexField(String fn, Short value) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public void pieSetIntElem(int value) throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public void pieSetOintElem(Integer value) throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public void pieSetIntIndexField(String fn, int value) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public void pieSetOintIndexField(String fn, Integer value) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public void pieSetLongElem(long value) throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public void pieSetOlongElem(Long value) throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public void pieSetLongIndexField(String fn, long value) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public void pieSetOlongIndexField(String fn, Long value) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public void pieSetFloatElem(float value) throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public void pieSetOfloatElem(Float value) throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public void pieSetDoubleElem(double value) throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public void pieSetOdoubleElem(Double value) throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public void pieSetStringElem(String value) throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public void pieSetStringIndexField(String fn, String value) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public void pieSetBigDecimalElem(BigDecimal value) throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public void pieSetDateElem(Date value) throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public void pieSetDateIndexField(String fn, Date value) throws PException {
				throw new PException(ERRNOINDEX);
			}

			public void pieSetCharArrayElem(char[] value) throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public void pieSetByteArrayElem(byte[] value) throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public void pieSetSerializedElem(Serializable value) throws PException {
				throw new PException(ERRTYPEUNSUPP);
			}

			public void pieSetRefElem(PName value) throws PException {
				ref = value;
			}
		}

		/**
		 * Provided to iterate over referenced ProdUnits shown as SetIndexedElem.
		 */
		class SetAccessorIterator implements Iterator {
			Iterator pusetIt;
			SetIndexedElem ie;

			SetAccessorIterator(Iterator it) {
				pusetIt = it;
				ie = new SetIndexedElem();
			}

			public boolean hasNext() {
				return pusetIt.hasNext();
			}

			public Object next() {
				if (!hasNext())
					return null;
				try {
					ie.pieSetRefElem(((ProdUnits) pusetIt.next()).getPBinding().getPName());
				} catch (PException pe) {
					pe.printStackTrace();
					return null;
				}
				return ie;
			}

			public void remove() {
			}
		}
	}
}
