/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.facility.naming.polymorphid;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.naming.api.PBinder;
import org.objectweb.jorm.naming.api.PNamingContext;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.jorm.facility.naming.generator.LongGenMgr;

/**
 * This manager allows the retrieval/allocation of PBinders and PNamingContext
 * that manage polymorphic identifiers/references. These identifiers/references
 * are composed of two longs: the first is the class identifier and the second
 * one is the object identifier within this class (see PolymorphName.pd).
 * @author	P. Dechamboux
 */
public interface PolymorphIdMgr {
	/**
	 * Notifies the manager that a PType has been associated with this binder.
	 */
	void binderTypeDef(PolymorphIdBinderInfo b);

	/**
	 * Retrieves the LongGenMgr associated with this PolymorphIdMgr.
	 */
	LongGenMgr getLongGenMgr();

	/**
	 * Retrieves a PBinder that manages polymorphic identifiers for the class
	 * with the given name. If it does not exist, creates it.
	 * @param cn	The class name.
	 */
	PBinder getPBinder(String cn) throws PException;

	/**
	 * Retrieves a PBinder that manages polymorphic identifiers for the class
	 * with the given name. If it does not exist, creates it.
	 * @param cn	The class name.
	 */
	PBinder getPBinder(long id) throws PException;

	/**
	 * Retrieves a PBinder that manages polymorphic identifiers for the class
	 * with the given name. If it does not exist, creates it.
	 * @param cn	The class name.
	 */
	PBinder getPBinder(String cn, Object conn) throws PException;

	/**
	 * Retrieves the mapper associated with this PolymorphIdMgr.
	 */
	PMapper getPMapper();

	/**
	 * Retrieves a PNamingContext that manages polymorphic references for the
	 * class with the given name. This means that such references can point to
	 * object of this class or any of its sub-classes. If it does not exist,
	 * creates it.
	 * @param cn	The class name.
	 */
	PNamingContext getRefNC(String cn);

	/**
	 * Initializes a polymorphic identifier manager by mapping relevant classes
	 * to the given mapper, and setting up relevant initial persistent
	 * information.
	 */
	void init(PMapper pm, byte clact) throws PException;

	/**
	 * Notifies the manager that a PType has been associated with this binder.
	 */
	void ncTypeDef(PolymorphRefNC n);

	/**
	 * Allocates a new persisttent class identifier.
	 */
	long newClassId() throws PException;
}
