/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.lib;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PMappingStructuresManager;
import org.objectweb.jorm.compiler.api.PExceptionCompiler;
import org.objectweb.jorm.generator.api.Generator;
import org.objectweb.jorm.generator.api.MOPFactory;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.metainfo.api.MappingFactory;
import org.objectweb.jorm.verifier.api.MappingVerifier;
import org.objectweb.jorm.verifier.api.Verifier;
import org.objectweb.jorm.xml2mi.api.MappingParser;
import org.objectweb.jorm.xml2mi.api.Parser;
import org.objectweb.jorm.mi2xml.api.MappingDomtreeBuilder;
import org.objectweb.jorm.mi2xml.api.Writer;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.Collection;
import java.util.HashMap;

/**
 * Used to hold the configuration information related to a JORM compiler
 * instance. The configuration is loaded from the
 * @author	P. Dechamboux
 */
public class JormConfigurator {

	public static final String LOGGER_NAME =
			"org.objectweb.jorm.compiler.config";

	public static final String DEFAULT_JORM_CONFIGURATION_FILE =
			"jorm.properties";

	Properties properties = null;
	Properties logProperties = null;
	String pathToJormcOpts = null;
	HashMap mappers = new HashMap();
	LoggerFactory loggerFactory = null;
	Logger logger = null;

	/**
	 * Configures the JORM compiler by loading information related to the
	 * different mappers supported by this instance of JORM.
	 * @throws org.objectweb.jorm.api.PException
	 */
	public void configure(String propertyfile) throws PException {
		if (properties != null) {
			throw new PExceptionCompiler("JORM already configured.");
		}
		try {
			File pf = new File(propertyfile);
			properties = new Properties();
			properties.load(new FileInputStream(pf));
			configure(properties);
			pathToJormcOpts = pf.getParent();
		} catch (IOException e) {
			InputStream is = getClass().getClassLoader().getResourceAsStream(propertyfile);
			if (is == null) {
				throw new PExceptionCompiler("Problem while configuring the JORM: unable to locate [" + propertyfile + "]");
			}
			try {
				properties = new Properties();
				properties.load(is);
				configure(properties);
			} catch (Exception e2) {
				properties = null;
				throw new PExceptionCompiler(e2, "Problem while configuring the JORM: unable to locate [" + propertyfile + "]");
			}
		}
	}

	public void configure() throws PException {
		configure(DEFAULT_JORM_CONFIGURATION_FILE);
	}

	/**
	 * Configures the JORM compiler by loading information related to the
	 * different mappers supported by this instance of JORM.
	 * @param p is the properties containing the jorm configuration
	 */
	public void configure(Properties p) {
		properties = p;
		StringTokenizer st = new StringTokenizer(p.getProperty("jorm.mapper.list").trim(), ",");
		while (st.hasMoreTokens()) {
			String mappername = st.nextToken();
			String sms = properties.getProperty("jorm.mapper.submappers." + mappername);
			if (sms == null) {
				continue;
			}
			ArrayList sml = new ArrayList();
			mappers.put(mappername, sml);
			StringTokenizer st2 = new StringTokenizer(sms.trim(), ",", false);
			while (st2.hasMoreTokens()) {
				sml.add(st2.nextToken());
			}
		}
	}

	/**
	 * Configures the logger factory object with a properties object.
	 * @param propertyfile    a properties object
	 */
	public void configureLog(String propertyfile) throws PException {
		if (loggerFactory == null) {
			loggerFactory = Monolog.init(propertyfile);
		} else {
			Properties p = new Properties();
			try {
				File pf = new File(propertyfile);
				p.load(new FileInputStream(pf));
			} catch (IOException e) {
				try {
					p.load(getClass().getClassLoader().getResourceAsStream(propertyfile));
				} catch (Exception e2) {
					throw new PExceptionCompiler(e, "Problem while configuring the JORM: unable to locate [" + propertyfile + "]");
				}
			}
			Monolog.loadMonologConfiguration(p, loggerFactory);
		}
	}

	/**
	 * Lists the names of supported mappers.
	 * @return	This list Iterator.
	 */
	public Iterator knownMappers() throws PException {
		if (properties == null) {
			throw new PExceptionCompiler("JORM has not been configured.");
		}
		return mappers.keySet().iterator();
	}

	/**
	 * It assignes the logger factory.
	 * @param lf
	 */
	public void setLoggerFactory(LoggerFactory lf) {
		loggerFactory = lf;
		logger = lf.getLogger(LOGGER_NAME);
	}

	/**
	 * Gets the logger factory associated with this instance of JORM.
	 * @return  The logger factory.
	 */
	public LoggerFactory getLoggerFactory() throws PException {
		if (loggerFactory == null) {
			throw new PExceptionCompiler("JORM has not been configured.");
		}
		return loggerFactory;
	}

	/**
	 * Gets the MI manager to be used by JORM.
	 * @return	That MI manager.
	 */
	public Manager getMIManager() throws PException {
		if (properties == null) {
			throw new PExceptionCompiler("JORM has not been configured.");
		}
		String cn = properties.getProperty("jorm.mimanager");
        if (cn == null) {
            throw new PException(
                    "Impossible to load a meta info manager: not configured:"
                    + properties);
        }
		try {
            Manager m = (Manager) Class.forName(cn).newInstance();
            if (loggerFactory != null && m instanceof Loggable) {
                ((Loggable) m).setLoggerFactory(loggerFactory);
            }
            return m;
		} catch (InstantiationException e) {
			throw new PExceptionCompiler(e, "Problem while instantiating " + cn);
		} catch (IllegalAccessException e) {
			throw new PExceptionCompiler(e, "Problem while trying to access " + cn);
		} catch (ClassNotFoundException e) {
			throw new PExceptionCompiler(e, "Problem while loading class " + cn);
		}
	}

	/**
	 * Gets the parser to be used by JORM.
	 * @return	That parser.
	 */
	public Parser getParser() throws PException {
		if (properties == null) {
			throw new PExceptionCompiler("JORM has not been configured.");
		}
		String cn = properties.getProperty("jorm.parser");
        if (cn == null) {
            throw new PException(
                    "Impossible to load a parser: not configured:"
                    + properties);
        }
		try {
			return (Parser) Class.forName(cn).newInstance();
		} catch (InstantiationException e) {
			throw new PExceptionCompiler(e, "Problem while instantiating " + cn);
		} catch (IllegalAccessException e) {
			throw new PExceptionCompiler(e, "Problem while trying to access " + cn);
		} catch (ClassNotFoundException e) {
			throw new PExceptionCompiler(e, "Problem while loading class " + cn);
		}
	}

	/**
	 * Gets the verifier to be used by JORM.
	 * @return	That verifier.
	 */
	public Verifier getVerifier() throws PException {
		if (properties == null) {
			throw new PExceptionCompiler("JORM has not been configured.");
		}
		String cn = properties.getProperty("jorm.verifier");
        if (cn == null) {
            throw new PException(
                    "Impossible to load a verifier: not configured:"
                    + properties);
        }
		try {
			return (Verifier) Class.forName(cn).newInstance();
		} catch (InstantiationException e) {
			throw new PExceptionCompiler(e, "Problem while instantiating " + cn);
		} catch (IllegalAccessException e) {
			throw new PExceptionCompiler(e, "Problem while trying to access " + cn);
		} catch (ClassNotFoundException e) {
			throw new PExceptionCompiler(e, "Problem while loading class " + cn);
		}
	}

	/**
	 * Gets the generator to be used by JORM.
	 * @return	That generator.
	 */
	public Generator getGenerator() throws PException {
		if (properties == null) {
			throw new PExceptionCompiler("JORM has not been configured.");
		}
		String cn = properties.getProperty("jorm.generator");
        if (cn == null) {
            throw new PException(
                    "Impossible to load a generator: not configured:"
                    + properties);
        }
		try {
			Generator res = (Generator) Class.forName(cn).newInstance();
			return res;
		} catch (InstantiationException e) {
			throw new PExceptionCompiler(e, "Problem while instantiating " + cn);
		} catch (IllegalAccessException e) {
			throw new PExceptionCompiler(e, "Problem while trying to access " + cn);
		} catch (ClassNotFoundException e) {
			throw new PExceptionCompiler(e, "Problem while loading class " + cn);
		}
	}

	/**
	 * Gets the name to be used for the compiler options file.
	 * @return	That name.
	 */
	public String getJormcOptsFile() throws PException {
		if (properties == null) {
			throw new PExceptionCompiler("JORM has not been configured.");
		}
		return properties.getProperty("jorm.compiler.optsfile");
	}

	/**
	 * Gets the MI factory associated with the given mapper.
	 * @param mappername	The name of the concerned mapper.
	 * @return The MappingFactory object.
	 * @throws org.objectweb.jorm.api.PException
	 */
	public MappingFactory getMIFactory(String mappername) throws PException {
		if (properties == null) {
			throw new PExceptionCompiler("JORM has not been configured.");
		}
		String cn = properties.getProperty("jorm.mapper.mifactory." + mappername);
        if (cn == null) {
            throw new PException(
                    "Impossible to load a meta info factory: not configured:"
                    + properties);
        }
		try {
			MappingFactory mf = (MappingFactory) Class.forName(cn).newInstance();
			return mf;
		} catch (InstantiationException e) {
			throw new PExceptionCompiler(e, "Problem while instantiating " + cn);
		} catch (IllegalAccessException e) {
			throw new PExceptionCompiler(e, "Problem while trying to access " + cn);
		} catch (ClassNotFoundException e) {
			throw new PExceptionCompiler(e, "Problem while loading class " + cn);
		}
	}

	/**
	 * Gets the schema manager associated with the given mapper.
	 * @param mappername	The name of the concerned mapper.
	 * @return The PMappingStructuresManager object.
	 * @throws org.objectweb.jorm.api.PException
	 */
	public PMappingStructuresManager getSchMgr(String mappername) throws PException {
		if (properties == null) {
			throw new PExceptionCompiler("JORM has not been configured.");
		}
		String cn = properties.getProperty("jorm.mapper.schmgr." + mappername);
        if (cn == null) {
            throw new PException(
                    "Impossible to load a schema manager: not configured:"
                    + properties);
        }
		try {
			PMappingStructuresManager mf = (PMappingStructuresManager) Class.forName(cn).newInstance();
			return mf;
		} catch (InstantiationException e) {
			throw new PExceptionCompiler(e, "Problem while instantiating " + cn);
		} catch (IllegalAccessException e) {
			throw new PExceptionCompiler(e, "Problem while trying to access " + cn);
		} catch (ClassNotFoundException e) {
			throw new PExceptionCompiler(e, "Problem while loading class " + cn);
		}
	}

	/**
	 * Gets the class name of the GenClassMapping associated with the given mapper.
	 * @param mappername	The name of the concerned mapper.
	 * @return The string giving the class name.
	 * @throws org.objectweb.jorm.api.PException
	 */
	public Class getGcmClass(String mappername) throws PException {
		if (properties == null) {
			throw new PExceptionCompiler("JORM has not been configured.");
		}
		String cn = properties.getProperty("jorm.mapper.gcmapping." + mappername);
        if (cn == null) {
            throw new PException(
                    "Impossible to load a generic class mapping: not configured:"
                    + properties);
        }
		try {
			Class res = Class.forName(cn);
			Object trycreation = res.newInstance();
			return res;
		} catch (InstantiationException e) {
			throw new PExceptionCompiler(e, "Problem while trying to instantiate " + cn);
		} catch (IllegalAccessException e) {
			throw new PExceptionCompiler(e, "Problem while trying to access " + cn);
		} catch (ClassNotFoundException e) {
			throw new PExceptionCompiler(e, "Problem while loading class " + cn);
		}
	}

	/**
	 * Gets the MOP factory associated with the given mapper.
	 * @param mappername	The name of the concerned mapper.
	 * @return The MOPFactory object.
	 * @throws org.objectweb.jorm.api.PException
	 */
	public MOPFactory getMOPFactory(String mappername) throws PException {
		if (properties == null) {
			throw new PExceptionCompiler("JORM has not been configured.");
		}
		String cn = properties.getProperty("jorm.mapper.mopfactory." + mappername);
        if (cn == null) {
            throw new PException(
                    "Impossible to load a meta object factory: not configured:" 
                    + properties);
        }
		try {
			MOPFactory mf = (MOPFactory) Class.forName(cn).newInstance();
			return mf;
		} catch (InstantiationException e) {
			throw new PExceptionCompiler(e, "Problem while instantiating " + cn);
		} catch (IllegalAccessException e) {
			throw new PExceptionCompiler(e, "Problem while trying to access " + cn);
		} catch (ClassNotFoundException e) {
			throw new PExceptionCompiler(e, "Problem while loading class " + cn);
		}
	}

	/**
	 * Gets an InputStream for reading the global compiler parameters.
	 * @return  The allocated InputStream.
	 * @throws org.objectweb.jorm.api.PException
	 */
	public InputStream getGlobalJormcOptsFile() throws PException {
		InputStream res = null;
		String jormcOptsFile = getJormcOptsFile();
		if (pathToJormcOpts != null) {
			if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
				logger.log(BasicLevel.DEBUG, "JormcOptsFile=" + pathToJormcOpts
											 + File.separator + jormcOptsFile);
			}
			File f = new File(
					pathToJormcOpts + File.separator + jormcOptsFile);
			if (f.exists()) {
				try {
					res = new FileInputStream(f);
				} catch (FileNotFoundException e) {
				}
			}
		}
		if (res == null) {
			File f = new File(jormcOptsFile);
			if (f.exists()) {
				try {
					res = new FileInputStream(f);
				} catch (FileNotFoundException e) {
				}
			} else {
				res = getClass().getClassLoader()
						.getResourceAsStream(jormcOptsFile);
			}
			if (res == null) {
				throw new PExceptionCompiler("Problem while accessing global " +
											 "compiler parameters: unable to locate ["
											 + getJormcOptsFile() + "]");
			}
		}
		return res;
	}

	/**
	 * Gets the parser associated with the given mapper for mapping infos.
	 * @param mappername	The name of the concerned mapper.
	 * @return The MappingParser object.
	 * @throws org.objectweb.jorm.api.PException
	 */
	public MappingParser getMappingParser(String mappername) throws PException {
		if (properties == null) {
			throw new PExceptionCompiler("JORM has not been configured.");
		}
		String cn = properties.getProperty("jorm.mapper.parser." + mappername);
		try {
			MappingParser p = (MappingParser) Class.forName(cn).newInstance();
			return p;
		} catch (InstantiationException e) {
			throw new PExceptionCompiler(e, "Problem while instantiating " + cn);
		} catch (IllegalAccessException e) {
			throw new PExceptionCompiler(e, "Problem while trying to access " + cn);
		} catch (ClassNotFoundException e) {
			throw new PExceptionCompiler(e, "Problem while loading class " + cn);
		}
	}

	/**
	 * Gets the verifier associated with the given mapper.
	 * @param mappername	The name of the concerned mapper.
	 * @return The MappingVerifier object.
	 * @throws org.objectweb.jorm.api.PException
	 */
	public MappingVerifier getMappingVerifier(String mappername) throws PException {
		if (properties == null) {
			throw new PExceptionCompiler("JORM has not been configured.");
		}
		String cn = properties.getProperty("jorm.mapper.verifier." + mappername);
		try {
			MappingVerifier v = (MappingVerifier) Class.forName(cn).newInstance();
			return v;
		} catch (InstantiationException e) {
			throw new PExceptionCompiler(e, "Problem while instantiating " + cn);
		} catch (IllegalAccessException e) {
			throw new PExceptionCompiler(e, "Problem while trying to access " + cn);
		} catch (ClassNotFoundException e) {
			throw new PExceptionCompiler(e, "Problem while loading class " + cn);
		}
	}

	/**
	 * Gets the BasicDomWriter object.
	 * @return The BasicDomWriter object.
	 * @throws org.objectweb.jorm.api.PException
	 */
	public Writer getWriter() throws PException {
		if (properties == null) {
			throw new PExceptionCompiler("JORM has not been configured.");
		}
		String cn = properties.getProperty("jorm.writer");
		try {
			return (Writer) Class.forName(cn).newInstance();
		} catch (InstantiationException e) {
			throw new PExceptionCompiler(e, "Problem while instantiating " + cn);
		} catch (IllegalAccessException e) {
			throw new PExceptionCompiler(e, "Problem while trying to access " + cn);
		} catch (ClassNotFoundException e) {
			throw new PExceptionCompiler(e, "Problem while loading class " + cn);
		}
	}

	/**
	 * Gets the MappingDomTreeBuilder object associated with the given mapper.
	 * @param mappername	The name of the concerned mapper.
	 * @return The MappingDomTreeBuilder object.
	 * @throws org.objectweb.jorm.api.PException
	 */
	public MappingDomtreeBuilder getMappingDomtreeBuilder(String mappername)
			throws PException {
		if (properties == null) {
			throw new PExceptionCompiler("JORM has not been configured.");
		}
		String cn = properties.getProperty("jorm.mapper.writer." + mappername);
		try {
			MappingDomtreeBuilder mdg = (MappingDomtreeBuilder) Class.forName(cn).newInstance();
			return mdg;
		} catch (InstantiationException e) {
			throw new PExceptionCompiler(e, "Problem while instantiating " + cn);
		} catch (IllegalAccessException e) {
			throw new PExceptionCompiler(e, "Problem while trying to access " + cn);
		} catch (ClassNotFoundException e) {
			throw new PExceptionCompiler(e, "Problem while loading class " + cn);
		}
	}

	/**
	 * Gets the submappers associated with a given mapper.
	 * @param mappername    The concerned mapper.
	 * @return  The collection of the names of submappers.
	 * @throws org.objectweb.jorm.api.PException
	 */
	public Collection getSubMappers(String mappername) throws PException {
		if (properties == null) {
			throw new PExceptionCompiler("JORM has not been configured.");
		}
		return (Collection) mappers.get(mappername);
	}

	/**
	 * Removes this mapper from the list of the ones used for generation.
	 * @param mn    The name of the concerned mapper.
	 */
	public void removeMapper(String mn) {
		mappers.remove(mn);
	}

	/**
	 * Removes all mappers from the list of the ones used for generation.
	 */
	public void removeAllMappers() {
		mappers.clear();
	}

	/**
	 * Removes this submapper from the list of the ones used for generation.
	 * @param mn    The concerned mapper.
	 * @param smn   The submapper to be removed.
	 */
	public void removeSubMapper(String mn, String smn) {
		ArrayList sms = (ArrayList) mappers.get(mn);
		if (sms == null) {
			return;
		}
		int pos = sms.indexOf(smn);
		if (pos == -1) {
			return;
		}
		sms.remove(pos);
	}

	/**
	 * Adds this submapper to the list of the ones used for generation.
	 * @param mn    The concerned mapper.
	 * @param smn   The submapper to be added.
	 */
	public void addSubMapper(String mn, String smn) {
		ArrayList sms = (ArrayList) mappers.get(mn);
		if (sms == null) {
			sms = new ArrayList();
			mappers.put(mn, sms);
		}
		int pos = sms.indexOf(smn);
		if (pos == -1) {
			sms.add(smn);
		}
	}
}
