/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.adapter;

import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.type.api.PType;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.PreparedStatement;
import java.sql.Types;
import java.util.Date;

/**
 * MySql 4.0.12
 * @author S.Chassande-Barrioz
 */
public class MysqlAdapter extends BasicRdbAdapter {

	private final static String MYSQLINDEXEDLOCATE = "locate";

	public MysqlAdapter() {
		super();
	}

	public boolean supportBatchPreparedStatement() {
		return false;
	}

	/**
	 * This calculates the size of a result set. If this feature is not supported
	 * by a database or its jdbc driver, this method returns -1.
	 * @param rs the result set which the size must be returned
	 */
	public int fetchResultSetSize(ResultSet rs) throws SQLException {
		return -1;
	}

	// MAX LENGTH OF VARCHAR TYPE MYSQL IS LIMITED to 255
	private int varcharSize = 254;
	private int varcharSizeInPk = 200;

	public int varbinarySize = 254;
	public int varbinarySizeInPk = 200;

	public void setVarcharSize(int s) {
		varcharSize = s;
	}

	public void setVarcharSizeInPk(int s) {
		varcharSizeInPk = s;
	}

	public String getSqlType(PType pt, boolean usedInPK, int size)
			throws PExceptionProtocol {
		if (pt == null) {
			throw new PExceptionProtocol("No Sql Type name associated to a null PType");
		}
		// ALL COLUMN IN PK SHOULD BE DECLARED NOT NULL
		int realsize = size;
		switch (pt.getTypeCode()) {
		case PType.TYPECODE_STRING:
			if (realsize == PType.NOSIZE) {
				realsize = (usedInPK ? varcharSizeInPk : varcharSize);
			}
			return (usedInPK
					? "VARCHAR(" + realsize + ")  NOT NULL"
					: "LONGTEXT ");
		case PType.TYPECODE_SERIALIZED:
			return "LONGTEXT";
		case PType.TYPECODE_BIGDECIMAL:
			return "NUMERIC(127,127)";
		case PType.TYPECODE_BYTEARRAY:
			if (realsize == PType.NOSIZE) {
				realsize = (usedInPK ? varbinarySizeInPk : varbinarySize);
			}
			return (usedInPK
					? "VARCHAR(" + realsize + ") BINARY NOT NULL"
					: "LONGTEXT ");
		case PType.TYPECODE_CHARARRAY:
			if (realsize == PType.NOSIZE) {
				realsize = (usedInPK ? varbinarySizeInPk : varbinarySize);
			}
			return (usedInPK
					? "VARCHAR(" + realsize + ") NOT NULL"
					: "LONGTEXT ");
		case PType.TYPECODE_DATE:
			return "DATETIME";
		default:
			String res = super.getSqlType(pt, usedInPK, size);
			if (usedInPK) {
				res += " NOT NULL";
			}
			return res;
		}
	}

	public int getSqlTypeCode(PType pt, String sqlColumnType) throws PExceptionProtocol {
		if (pt == null) {
			throw new PExceptionProtocol("No Sql Type code associated to a null PType");
		}
		switch (pt.getTypeCode()) {
		case PType.TYPECODE_DATE:
			if (sqlColumnType == null) {
				return Types.TIMESTAMP;
			} else {
				String str = sqlColumnType.toUpperCase();
				if (str.startsWith("DATE")) {
					return Types.DATE;
				} else if (str.startsWith("TIMESTAMP")) {
					return Types.TIMESTAMP;
				} else if (str.startsWith("TIME")) {
					return Types.TIME;
				} else {
					return Types.TIMESTAMP;
				}
			}
		default:
			return super.getSqlTypeCode(pt, sqlColumnType);
		}
	}


	public String getSqlGetterName(PType pt, String resultSet, int idx,
								   String sql)
			throws PExceptionProtocol {
		if (pt == null) {
			throw new PExceptionProtocol("Invalid parameter");
		}
		switch (pt.getTypeCode()) {
		case PType.TYPECODE_OBJFLOAT:
			return "new Float(" + resultSet + ".getFloat(" + idx + "))";
		case PType.TYPECODE_OBJBOOLEAN:
			return "new Boolean(" + resultSet + ".getBoolean(" + idx + "))";
		case PType.TYPECODE_CHAR:
			return "(((String) (fieldval = " + resultSet + ".getString("
					+ idx + "))).length()==0 " +
					"? ' ' : ((String) fieldval).charAt(0))";
		case PType.TYPECODE_OBJCHAR:
			return "((fieldval = " + resultSet + ".getString(" + idx + "))==null " +
					"? null " +
					": new Character(" +
					"(((String) fieldval).length()==0 " +
					"? ' ' : ((String) fieldval).charAt(0))))";
		case PType.TYPECODE_DATE:
			if (sql == null) {
				return resultSet + ".getTimestamp(" + idx + ")";
			} else {
				String str = sql.toUpperCase();
				if (str.startsWith("TIMESTAMP") || str.equals("DATETIME")) {
					return resultSet + ".getTimestamp(" + idx + ")";
				} else if (str.startsWith("DATE")) {
					return resultSet + ".getDate(" + idx + ")";
				} else if (str.startsWith("TIME")) {
					return resultSet + ".getTime(" + idx + ")";
				} else {
					return resultSet + ".getTimestamp(" + idx + ")";
				}
			}
		default:
			return super.getSqlGetterName(pt, resultSet, idx, sql);
		}
	}

	public String getSqlSetterName(PType pt, String value, int idx,
								   String sqlColumnType) throws PExceptionProtocol {
		if (pt == null) {
			throw new PExceptionProtocol("No Sql setter associated to a null PType");
		}
		switch (pt.getTypeCode()) {
		case PType.TYPECODE_DATE:
			String type = "java.sql.Timestamp";
			String sqlSetterName = "setTimestamp";
			if (sqlColumnType != null) {
				String str = sqlColumnType.toUpperCase();
				if (str.startsWith("TIMESTAMP")
						|| str.equals("DATETIME")) {
					type = "java.sql.Timestamp";
					sqlSetterName = "setTimestamp";
				} else if (str.startsWith("DATE")) {
					type = "java.sql.Date";
					sqlSetterName = "setDate";
				} else if (str.startsWith("TIME")) {
					type = "java.sql.Time";
					sqlSetterName = "setTime";
				}
			}
			return sqlSetterName + "(" + idx + ", "
					+ "(" + value + "== null ? null : " +
					"(" + value + " instanceof " + type
					+ " ? (" + type + ")" + value
					+ " : new " + type + "(" + value + ".getTime()))))";
		default:
			return super.getSqlSetterName(pt, value, idx, sqlColumnType);
		}
	}

	public String getIndexedLocateExpression(String instring,
											 String substring,
											 String fromIndex) {
		return MYSQLINDEXEDLOCATE + "(" + substring + ", " + instring + ", " +
				fromIndex + ")";
	}

	public void setDate(PreparedStatement ps, int idx, Date o,
						String columnType)
			throws SQLException {
		if (columnType == null) {
			//default is a Timestamp
			if (o == null)
				ps.setTimestamp(idx, new java.sql.Timestamp(0));
			else if (o instanceof java.sql.Timestamp) {
				ps.setTimestamp(idx, (java.sql.Timestamp) o);
			} else {
				ps.setTimestamp(idx, new java.sql.Timestamp(o.getTime()));
			}
		} else if (columnType.startsWith("TIMESTAMP")
				|| columnType.equals("DATETIME")) {
			if (o == null)
				ps.setTimestamp(idx, new java.sql.Timestamp(0));
			else if (o instanceof java.sql.Timestamp) {
				ps.setTimestamp(idx, (java.sql.Timestamp) o);
			} else {
				ps.setTimestamp(idx, new java.sql.Timestamp(o.getTime()));
			}
		} else if (columnType.startsWith("TIME")) {
			if (o == null)
				ps.setTime(idx, new java.sql.Time(0));
			else if (o instanceof java.sql.Time) {
				ps.setTime(idx, (java.sql.Time) o);
			} else {
				ps.setTime(idx, new java.sql.Time(o.getTime()));
			}
		} else if (columnType.startsWith("DATE")) {
			if (o == null)
				ps.setDate(idx, new java.sql.Date(0));
			else if (o instanceof java.sql.Date) {
				ps.setDate(idx, (java.sql.Date) o);
			} else {
				ps.setDate(idx, new java.sql.Date(o.getTime()));
			}
		}

	}
}
