/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.adapter;

import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.type.api.PType;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.PreparedStatement;
import java.sql.Connection;
import java.util.Date;
import java.util.List;
import java.math.BigDecimal;
import java.io.Serializable;
import java.io.IOException;

/*
 * This interface permits to do the conversion between the jorm types PType and
 * the sql types. It is also possible to fetch the function name of getter and
 * setter methods associated to a PType.
 *
 * @author : S.Chassande-Barrioz
 */

public interface RdbAdapter {
	/**
	 * This method returns the SQL type linked to the PType
	 * specified in parameter.
	 */
	String getSqlType(PType pt, boolean usedInPK, int size) throws PExceptionProtocol;

	/**
	 * This method returns the sql getter method associated to the PType
	 * specified in parameter. The returned string is the fetching of the
	 * value on the resultset to the specified index:
	 * "@resultSet.getBoolean(@idx)"
	 */
	String getSqlGetterName(PType pt, String resultSet, int idx)
			throws PExceptionProtocol;

	/**
	 * This method returns the sql getter method associated to the PType
	 * specified in parameter. The returned string is the fetching of the
	 * value on the resultset to the specified index:
	 * "@resultSet.getBoolean(@idx)"
	 */
	String getSqlGetterName(PType pt, String resultSet, int idx,
							String sqlColumnType) throws PExceptionProtocol;

	/**
	 * This method returns the sql setter method associated to the PType
	 * specified in parameter. The returned string is the assignment of the
	 * value to the specified index: "setBoolean(@idx, @value)"
	 */
	String getSqlSetterName(PType pt, String value, int idx)
			throws PExceptionProtocol;

	/**
	 * This method returns the sql setter method associated to the PType
	 * specified in parameter. The returned string is the assignment of the
	 * value to the specified index: "setBoolean(@idx, @value)"
	 */
	String getSqlSetterName(PType pt, String value, int idx,
							String sqlColumnType) throws PExceptionProtocol;

	/**
	 * Indicates if the driver supports the batch of PreparedStatement
	 */
	boolean supportBatchPreparedStatement();

	/**
	 * This method returns a String value that represents a value. This method
	 * can be used for the constant parameters of SQL queries.
	 * @param value
	 * @param type is the PType of the @value
	 */
	String getValueAsSQLString(Object value, PType type);

	/**
	 * This method returns a String value that represents a boolean value. This
	 * method can be used for the constant parameters of SQL queries.
	 */
	String getValueAsSQLString(boolean value);

	/**
	 * This method returns a String value that represents a byte value. This
	 * method can be used for the constant parameters of SQL queries.
	 */
	String getValueAsSQLString(byte value);

	/**
	 * This method returns a String value that represents a char value. This
	 * method can be used for the constant parameters of SQL queries.
	 */
	String getValueAsSQLString(char value);

	/**
	 * This method returns a String value that represents a short value. This
	 * method can be used for the constant parameters of SQL queries.
	 */
	String getValueAsSQLString(short value);

	/**
	 * This method returns a String value that represents a int value. This
	 * method can be used for the constant parameters of SQL queries.
	 */
	String getValueAsSQLString(int value);

	/**
	 * This method returns a String value that represents a long value. This
	 * method can be used for the constant parameters of SQL queries.
	 */
	String getValueAsSQLString(long value);

	/**
	 * This method returns a String value that represents a float value. This
	 * method can be used for the constant parameters of SQL queries.
	 */
	String getValueAsSQLString(float value);

	/**
	 * This method returns a String value that represents a double value. This
	 * method can be used for the constant parameters of SQL queries.
	 */
	String getValueAsSQLString(double value);

	/**
	 * This method returns the SQL type code linked to the PType
	 * specified in parameter.
	 * @return An int constant defined in the java.sql.Types class
	 */
	int getSqlTypeCode(PType pt) throws PExceptionProtocol;

	/**
	 * This method returns the SQL type code linked to the PType
	 * specified in parameter.
	 * @return An int constant defined in the java.sql.Types class
	 */
	int getSqlTypeCode(PType pt, String sqlColumnType) throws PExceptionProtocol;

	/**
	 * This method calculates the size of a result set. If this feature is not
	 * supported by a database or its jdbc driver, this method returns -1.
	 * @param rs the result set which the size must be returned
	 */
	int fetchResultSetSize(ResultSet rs) throws SQLException;

	boolean getBoolean(ResultSet rs, int idx) throws SQLException;

	Boolean getOboolean(ResultSet rs, int idx) throws SQLException;

	char getChar(ResultSet rs, int idx) throws SQLException;

	Character getOchar(ResultSet rs, int idx) throws SQLException;

	byte getByte(ResultSet rs, int idx) throws SQLException;

	Byte getObyte(ResultSet rs, int idx) throws SQLException;

	short getShort(ResultSet rs, int idx) throws SQLException;

	Short getOshort(ResultSet rs, int idx) throws SQLException;

	int getInt(ResultSet rs, int idx) throws SQLException;

	Integer getOint(ResultSet rs, int idx) throws SQLException;

	long getLong(ResultSet rs, int idx) throws SQLException;

	Long getOlong(ResultSet rs, int idx) throws SQLException;

	float getFloat(ResultSet rs, int idx) throws SQLException;

	Float getOfloat(ResultSet rs, int idx) throws SQLException;

	double getDouble(ResultSet rs, int idx) throws SQLException;

	Double getOdouble(ResultSet rs, int idx) throws SQLException;

	String getString(ResultSet rs, int idx) throws SQLException;

	Date getDate(ResultSet rs, int idx) throws SQLException;

	Date getDate(ResultSet rs, int idx, String sqlColumnType) throws SQLException;

	BigDecimal getBigDecimal(ResultSet rs, int idx) throws SQLException;

	char[] getCharArray(ResultSet rs, int idx) throws SQLException;

	byte[] getByteArray(ResultSet rs, int idx) throws SQLException;

	Serializable getSerialized(ResultSet rs, int idx) throws SQLException, IOException, ClassNotFoundException;


	void setBoolean(PreparedStatement ps, int idx, boolean o)
			throws SQLException;

	void setOboolean(PreparedStatement ps, int idx, Boolean o)
			throws SQLException;

	void setChar(PreparedStatement ps, int idx, char c)
			throws SQLException;

	void setOchar(PreparedStatement ps, int idx, Character o)
			throws SQLException;

	void setByte(PreparedStatement ps, int idx, byte b)
			throws SQLException;

	void setObyte(PreparedStatement ps, int idx, Byte o)
			throws SQLException;

	void setShort(PreparedStatement ps, int idx, short o)
			throws SQLException;

	void setOshort(PreparedStatement ps, int idx, Short o)
			throws SQLException;

	void setInt(PreparedStatement ps, int idx, int o)
			throws SQLException;

	void setOint(PreparedStatement ps, int idx, Integer o)
			throws SQLException;

	void setLong(PreparedStatement ps, int idx, long o)
			throws SQLException;

	void setOlong(PreparedStatement ps, int idx, Long o)
			throws SQLException;

	void setFloat(PreparedStatement ps, int idx, float o)
			throws SQLException;

	void setOfloat(PreparedStatement ps, int idx, Float o)
			throws SQLException;

	void setDouble(PreparedStatement ps, int idx, double o)
			throws SQLException;

	void setOdouble(PreparedStatement ps, int idx, Double o)
			throws SQLException;

	void setString(PreparedStatement ps, int idx, String o)
			throws SQLException;

	void setDate(PreparedStatement ps, int idx, Date o, String columnType)
			throws SQLException;

	void setCharArray(PreparedStatement ps, int idx, char[] o)
			throws SQLException;

	void setByteArray(PreparedStatement ps, int idx, byte[] o)
			throws SQLException;

	void setBigDecimal(PreparedStatement ps, int idx, BigDecimal o)
			throws SQLException;

	void setSerialized(PreparedStatement ps, int idx, Serializable o)
			throws SQLException, IOException;

	String getSubstringName();

	String getConcatExpression(String op1, String op2);

	/**
	 * Returns the expression for searching the position of the first occurrence
	 * of a substring in a string.
	 * <p>Two main syntaxes are found: position(substr in str) and
	 * instr(str, substr).
	 * @param substring The substring searched
	 * @param instring The string in which to search the substring
	 * @return The corresponding relational expression.
	 */
	String getFirstLocateExpression(String substring, String instring)
			throws PException;

	/**
	 * Returns the expression for searching the position of the first occurrence
	 * of a substring in a string starting from a given index.
	 * <p>Two main syntaxes are found: instr() and locate()
	 * @param inputString The string in which to search the substring
	 * @param substring The substring searched
	 * @param fromIndex The index from which to start searching
	 * @return The corresponding relational expression.
	 */
	String getIndexedLocateExpression(String inputString, String substring,
									  String fromIndex)
			throws PException;

	String getQuery(String selectClause, List jts, String whereclause);

	/**
	 * Used to write table aliases, for which the syntax may vary (use of
	 * 'as', except for Oracle)
	 * @param alias The table alias
	 * @param sb The StringBuffer into which to write the alias.
	 */
	void writeTableAlias(String alias, StringBuffer sb);

	String handleOrderBy(String query);

	String getNextValInSequence(String seqName);

	String getCreateSequence(String seqName);

	boolean existTable(Connection connection, String tableName)
			throws SQLException;

	boolean existSequence(Connection connection, String seqName) 
			throws SQLException;
}
