/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.genclass;

import org.objectweb.jorm.api.PAccessor;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.api.PGenClassAccessor;
import org.objectweb.jorm.api.PIndexedElem;
import org.objectweb.jorm.api.PStateGraph;
import org.objectweb.jorm.genclass.lib.GenClassBinding;
import org.objectweb.jorm.genclass.lib.GenClassMapping;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.api.PNameGetter;

import java.io.IOException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Iterator;

/**
 * This class provides the implementation of PBinding for RDBMS.
 * @author	P. D?chamboux
 */

public class RdbGenClassBinding extends GenClassBinding {

	/**
	 * It constructs a RdbGenClassBinding.
	 */
	public RdbGenClassBinding() {
		super();
	}

	/**
	 * It constructs a RdbGenClassBinding.
	 */
	public RdbGenClassBinding(GenClassMapping mapping)
			throws PExceptionProtocol {
		super(mapping);
	}

	/**
	 * It verifies if the DSI attached to this binding exists within the DS.
	 * @param conn	The connection that can be used to refer to the DS.
	 * @return	It returns true if the DSI exists.
	 * @exception PExceptionIO	It is raised when a problem occured while
	 * 							accessing to the DS.
	 * @exception PExceptionProtocol	It is raised when it is called with a
	 * 									state different from the following ones:
	 * 									LIFECYCLE_ACTIVEFORIO
	 * 									LIFECYCLE_NOACC
	 */
	public boolean exist(Object conn) throws PException {
		byte nextstate = PStateGraph.nextStatePBinding(status, ACTION_EXIST);
		if (nextstate == LIFECYCLE_ERROR)
			throw new PExceptionProtocol(
					"Cannot perform exist on this GenClass PBinding");
		status = nextstate;
		return true;
	}

	/**
	 * It reads the data store instance designated by the PName of this
	 * binding and transfers each attribute value to the accessor object.
	 * @param conn	The connection that can be used to refer to the DS.
	 * @exception PExceptionIO	It is raised when a problem occured while
	 * 							accessing to the DS.
	 * @exception PExceptionNaming	It is raised when there is a problem for the
	 * 								associated binder to manage this PName.
	 * @exception PExceptionProtocol	It is raised when it is called with a
	 * 									state different from the following ones:
	 * 									LIFECYCLE_ACTIVEFORIO
	 */
	public void read(Object conn, PAccessor pa)
			throws PException {
		PGenClassAccessor paccessorGenClass = (PGenClassAccessor) pa;
		byte nextstate = PStateGraph.nextStatePBinding(status, ACTION_READ);
		if (nextstate == LIFECYCLE_ERROR)
			throw new PExceptionProtocol(
					"Cannot perform read on this GenClass PBinding");
		PreparedStatement pstmt = null;
		ResultSet rs = null;
		try {
			RdbGenClassMapping gcmapping = (RdbGenClassMapping) genClassMapping;
			Connection sqlconn = gcmapping.getConnection(conn);
			pstmt = sqlconn.prepareStatement(gcmapping.getSelectQuery());
			gcmapping.defineIdentifierVal(pName, pstmt, 1);
			rs = pstmt.executeQuery();
			int size = ((RdbGenClassMapping) genClassMapping)
					.typeConverter.fetchResultSetSize(rs);
			paccessorGenClass.paSetNbElem(size);
			PNameGetter png = gcmapping.getPNameGetter(rs);
			while (rs.next()) {
				gcmapping.passValuesToAccessor(conn, rs, paccessorGenClass, png);
			}
		} catch (SQLException e) {
			throw new PExceptionIO(e);
		} catch (IOException e) {
			throw new PExceptionIO(e);
		} finally {
			try {
				if (rs != null) {
					rs.close();
				}
				if (pstmt != null) {
					pstmt.close();
				}
			} catch (SQLException e) {
				throw new PExceptionIO(e);
			}
		}
		status = nextstate;
	}

	/**
	 * It transfers each attribute value taken from the PAccessor
	 * object to the data store instance designated by the PName
	 * associated with this binding.
	 * @param conn	The connection that can be used to refer to the DS.
	 * @exception PExceptionIO	It is raised when a problem occured while
	 * 							accessing to the DS.
	 * @exception PExceptionProtocol	It is raised when it is called with a
	 * 									state different from the following ones:
	 * 									LIFECYCLE_ACTIVEFORIO
	 * 									LIFECYCLE_NEWTOWRITE
	 * 									LIFECYCLE_DELTOWRITE
	 * 									LIFECYCLE_DELTOWRITENOACC
	 */
	public void write(Object conn, PAccessor pa) throws PException {
		PGenClassAccessor paccessorGenClass = (PGenClassAccessor) pa;
		byte nextstate = PStateGraph.nextStatePBinding(status, ACTION_WRITE);
		if (nextstate == LIFECYCLE_ERROR)
			throw new PExceptionProtocol(
					"Cannot perform write on this GenClass PBinding");
		if (((RdbGenClassMapping) genClassMapping).readOnly) {
			status = nextstate;
			return;
		}
		try {
			RdbGenClassMapping gcmapping = (RdbGenClassMapping) genClassMapping;
			Connection sqlconn = gcmapping.getConnection(conn);
			switch (status) {
			case LIFECYCLE_NEWTOWRITE:
				createGcObject(gcmapping, sqlconn, paccessorGenClass);
				break;
			case LIFECYCLE_DELTOWRITE:
				deleteGcObject(gcmapping, sqlconn);
				break;
			case LIFECYCLE_ACTIVEFORIO:
				if (paccessorGenClass.paDeltaSupported())
					updateGcObject(gcmapping, sqlconn, paccessorGenClass);
				else {
					deleteGcObject(gcmapping, sqlconn);
					createGcObject(gcmapping, sqlconn, paccessorGenClass);
				}
				break;
			default :
				throw new PExceptionProtocol("unmanaged status: " + status);
			}
		} catch (SQLException e) {
			throw new PExceptionIO(e);
		} catch (IOException e) {
			throw new PExceptionIO(e);
		}
		status = nextstate;
	}

	/**
	 *
	 */
	private void createGcObject(RdbGenClassMapping gcmapping,
								Connection sqlconn,
								PGenClassAccessor paccessorGenClass)
			throws SQLException, PException, IOException {
		int size = paccessorGenClass.paGetNbElem();
		if (size == 0) {
			// Empty GC: nothing to do
			return;
		}
		PreparedStatement pstmt_ins = null;
		PreparedStatement pstmt_upd = null;
		try {
			if (gcmapping.colocated) {
				pstmt_upd = sqlconn.prepareStatement(gcmapping.getUpdateQuery());
			}
			pstmt_ins = sqlconn.prepareStatement(gcmapping.getInsertQuery());
			// Inserts all elements of the MI of this generic class instance
			Iterator iterelem = paccessorGenClass.paIterator();
			int paramindex;
			boolean batchused = false;
			while (iterelem.hasNext()) {
				PIndexedElem pie = (PIndexedElem) iterelem.next();
				if (gcmapping.colocated) {
					paramindex = gcmapping.defineIdentifierVal(pName, pstmt_upd, 1);
					paramindex = gcmapping.defineIndexVal(pie, pstmt_upd, paramindex);
					gcmapping.defineElemVal(pie, pstmt_upd, paramindex);
					if (pstmt_upd.executeUpdate() > 0) {
						continue;
					}
				}
				paramindex = gcmapping.defineIdentifierVal(pName, pstmt_ins, 1);
				paramindex = gcmapping.defineElemVal(pie, pstmt_ins, paramindex);
				gcmapping.defineIndexVal(pie, pstmt_ins, paramindex);
				if (gcmapping.useBatch && !gcmapping.colocated) {
					batchused = true;
					pstmt_ins.addBatch();
				} else {
					pstmt_ins.execute();
				}
			}
			if (batchused && gcmapping.useBatch) {
				pstmt_ins.executeBatch();
			}
		} finally {
			if (pstmt_ins != null) {
				pstmt_ins.close();
			}
			if (pstmt_upd != null) {
				pstmt_upd.close();
			}
		}
	}

	/**
	 *
	 */
	private void deleteGcObject(RdbGenClassMapping gcmapping,
								Connection sqlconn)
			throws SQLException, PException, IOException {
		PreparedStatement pstmt = null;
		try {
			pstmt = sqlconn.prepareStatement(gcmapping.colocated
											 ? gcmapping.getUpdateNullAllQuery()
											 : gcmapping.getDeleteAllQuery());
			// Deletes all elements of this generic class instance
			gcmapping.defineIdentifierVal(pName, pstmt, 1);
			pstmt.execute();
		} finally {
			if (pstmt != null) {
				pstmt.close();
			}
		}
	}

	/**
	 *
	 */
	private void updateGcObject(RdbGenClassMapping gcmapping,
								Connection sqlconn,
								PGenClassAccessor paccessorGenClass)
			throws SQLException, PException, IOException {
		PreparedStatement pstmt_ins = null;
		PreparedStatement pstmt_del = null;
		PreparedStatement pstmt_upd = null;
		PreparedStatement pstmt_upd_null = null;
		try {
			Iterator iterelem = paccessorGenClass.paIterator();
			int paramindex;
			while (iterelem.hasNext()) {
				PIndexedElem pie = (PIndexedElem) iterelem.next();
				switch (pie.getElemStatus()) {
				case PIndexedElem.ELEM_CREATED:
					if (pstmt_ins == null) {
						pstmt_ins = sqlconn.prepareStatement(gcmapping.getInsertQuery());
					}
					if (gcmapping.colocated) {
						if (pstmt_upd == null) {
							pstmt_upd = sqlconn.prepareStatement(gcmapping.getUpdateQuery());
						}
						paramindex = gcmapping.defineElemVal(pie, pstmt_upd, 1);
						paramindex = gcmapping.defineIdentifierVal(pName, pstmt_upd, paramindex);
						gcmapping.defineIndexVal(pie, pstmt_upd, paramindex);
						if (pstmt_upd.executeUpdate() == 0) {
							paramindex = gcmapping.defineIdentifierVal(pName, pstmt_ins, 1);
							paramindex = gcmapping.defineElemVal(pie, pstmt_ins, paramindex);
							gcmapping.defineIndexVal(pie, pstmt_ins, paramindex);
							pstmt_ins.execute();
						}
					} else {
						paramindex = gcmapping.defineIdentifierVal(pName, pstmt_ins, 1);
						paramindex = gcmapping.defineElemVal(pie, pstmt_ins, paramindex);
						gcmapping.defineIndexVal(pie, pstmt_ins, paramindex);
						if (gcmapping.useBatch) {
							pstmt_ins.addBatch();
						} else {
							pstmt_ins.execute();
						}
					}
					break;
				case PIndexedElem.ELEM_DELETED:
					if (gcmapping.colocated) {
						if (pstmt_upd_null == null) {
							pstmt_upd_null = sqlconn.prepareStatement(gcmapping.getUpdateNullElemQuery());
						}
						paramindex = gcmapping.defineIdentifierVal(pName, pstmt_upd_null, 1);
						paramindex = gcmapping.defineIndexVal(pie, pstmt_upd_null, paramindex);
						gcmapping.defineElemVal(pie, pstmt_upd_null, paramindex);
						if (gcmapping.useBatch) {
							pstmt_upd_null.addBatch();
						} else {
							pstmt_upd_null.execute();
						}
					} else {
						if (pstmt_del == null) {
							pstmt_del = sqlconn.prepareStatement(gcmapping.getDeleteElemQuery());
						}
						paramindex = gcmapping.defineIdentifierVal(pName, pstmt_del, 1);
						paramindex = gcmapping.defineIndexVal(pie, pstmt_del, paramindex);
						gcmapping.defineElemVal(pie, pstmt_del, paramindex);
						if (gcmapping.useBatch) {
							pstmt_del.addBatch();
						} else {
							pstmt_del.execute();
						}
					}
					break;
				case PIndexedElem.ELEM_MODIFIED:
					if (pstmt_upd == null) {
						pstmt_upd = sqlconn.prepareStatement(gcmapping.getUpdateQuery());
					}
					if (gcmapping.colocated) {
						paramindex = gcmapping.defineIdentifierVal(pName, pstmt_upd, 1);
						paramindex = gcmapping.defineIndexVal(pie, pstmt_upd, paramindex);
						gcmapping.defineElemVal(pie, pstmt_upd, paramindex);
					} else {
						paramindex = gcmapping.defineElemVal(pie, pstmt_upd, 1);
						paramindex = gcmapping.defineIdentifierVal(pName, pstmt_upd, paramindex);
						gcmapping.defineIndexVal(pie, pstmt_upd, paramindex);
					}
					if (gcmapping.useBatch) {
						pstmt_upd.addBatch();
					} else {
						pstmt_upd.execute();
					}
					break;
				case PIndexedElem.ELEM_UNMODIFIED:
					continue;
				default :
					throw new PExceptionProtocol(
							"Unknown modification status for PIndexedElem.");
				}
			}
			if (gcmapping.useBatch) {
				if (pstmt_ins != null) {
					pstmt_ins.execute();
				}
				if (pstmt_upd_null != null) {
					pstmt_upd_null.execute();
				}
				if (pstmt_del != null) {
					pstmt_del.execute();
				}
				if (pstmt_upd != null) {
					pstmt_upd.execute();
				}
			}
		} finally {
			if (pstmt_ins != null) {
				pstmt_ins.close();
			}
			if (pstmt_upd_null != null) {
				pstmt_upd_null.close();
			}
			if (pstmt_del != null) {
				pstmt_del.close();
			}
			if (pstmt_upd != null) {
				pstmt_upd.close();
			}
		}
	}
}
