/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.genclass;

import org.objectweb.jorm.api.PBinding;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.api.PExceptionProtocol;
import org.objectweb.jorm.api.PGenClassAccessor;
import org.objectweb.jorm.api.PIndexedElem;
import org.objectweb.jorm.api.PMappingCallback;
import org.objectweb.jorm.api.ConnectionHolder;
import org.objectweb.jorm.genclass.api.FieldDesc;
import org.objectweb.jorm.genclass.lib.GenClassMapping;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNameGetter;
import org.objectweb.jorm.naming.lib.CTHelper;
import org.objectweb.jorm.type.api.PExceptionTyping;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.mapper.rdb.adapter.RdbAdapter;
import org.objectweb.transaction.api.ConnectionWrapper;
import org.objectweb.util.monolog.api.BasicLevel;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Iterator;

/**
 * This class provides the implementation of PClassMapping for RDBMS.
 * @author	P. D?chamboux
 */

public class RdbGenClassMapping extends GenClassMapping
		implements RdbGenClassProp {
	static final byte ID = 0x01;
	static final byte INDEX = 0x02;
	static final byte ELEM = 0x04;
	static final String SEPAND = " AND ";
	static final String SEPSCOL = ", ";
	static final String QUOTE = "'";
	static final String ASSIGNVAL = " = ?";
	static final String ASSIGN_NULL = " = null";
	/**
	 * The name of the table into which to store generic class object elements.
	 * It should be null in case of "legacy" mapping.
	 */
	protected String tableName = null;
	/**
	 * This is the type converter defined for a particular kind of RDB.
	 */
	protected RdbAdapter typeConverter = null;
	protected boolean useBatch = true;
	protected boolean colocated = false;
	protected boolean colocatedMaster = false;
	protected boolean readOnly = false;

	private String selectQuery = null;
	private String selectAllNamesQuery = null;
	private String insertQuery = null;
	private String deleteAllQuery = null;
	private String deleteElemQuery = null;
	private String updateQuery = null;
	private String updateNullAllQuery = null;
	private String updateNullElemQuery = null;

	/**
	 * It returns an SQL connection from the connection received from the
	 * application, which should be either an actual SQL one or a
	 * ConnectionWrapper one.
	 * @param conn	The connection received from the application.
	 * @exception PExceptionIO	It is raised if no valid SQL connection can
	 * 							be returned to access to the RDB.
	 */
	Connection getConnection(Object conn) throws PExceptionIO {
		if (conn == null)
			throw new PExceptionIO("Needs a connection to access to the RDB");
		else if (conn instanceof Connection)
			return (Connection) conn;
		else if (conn instanceof ConnectionHolder)
			try {
				return (Connection) ((ConnectionHolder) conn).getCHConnection();
			} catch (PException e) {
				throw new PExceptionIO(e, "Impossible to fetch a connection to access to the RDB:");
			}
		else if (conn instanceof ConnectionWrapper)
			return (Connection) ((ConnectionWrapper) conn).getNonCCIConnection();
		else
			throw new PExceptionIO("Wrong connection type to access to the RDB:" + conn);
	}

	String getSelectQuery() {
		if (selectQuery == null) {
			synchronized (this) {
				if (selectQuery == null) {
					StringBuffer sqlstring = new StringBuffer("SELECT ");
					defineAllNames(sqlstring, (byte) (INDEX | ELEM));
					sqlstring.append(" FROM " + tableName + " WHERE ");
					defineIdentifierClause(sqlstring);
					selectQuery = sqlstring.toString();
				}
			}
		}
		return selectQuery;
	}

	String getSelectAllNamesQuery() {
		if (selectAllNamesQuery == null) {
			synchronized (this) {
				if (selectAllNamesQuery == null) {
					StringBuffer sqlstring = new StringBuffer("SELECT ");
					defineAllNames(sqlstring, ID);
					sqlstring.append(" FROM " + tableName);
					selectAllNamesQuery = sqlstring.toString();
				}
			}
		}
		return selectAllNamesQuery;
	}

	String getInsertQuery() {
		if (insertQuery == null) {
			synchronized (this) {
				if (insertQuery == null) {
					StringBuffer sqlstring = new StringBuffer("INSERT INTO " + tableName);
					// Specify the column names because the order of the fields is not
					// always the same as in the meta information.
					sqlstring.append("(");
					defineNames(getIdentifierFields(), sqlstring, true, (short) 0);
					defineNames(getElemFields(), sqlstring, false,
								(short) (getElemFields().length == 1 ? 0 : 1));
					defineNames(getIndexFields(), sqlstring, false, (short) 0);
					sqlstring.append(") ");
					sqlstring.append("VALUES (");
					int nq = getIdentifierFields().length;
					nq += (getElemFields().length == 1 ? 1 : getElemFields().length - 1);
					nq += getIndexFields().length;
					String sep = "";
					for (; nq > 0; nq--) {
						sqlstring.append(sep);
						sep = SEPSCOL;
						sqlstring.append("?");
					}
					sqlstring.append(")");
					insertQuery = sqlstring.toString();
				}
			}
		}
		return insertQuery;
	}

	String getDeleteAllQuery() {
		if (deleteAllQuery == null) {
			synchronized (this) {
				if (deleteAllQuery == null) {
					StringBuffer sqlstring
							= new StringBuffer("DELETE FROM " + tableName + " WHERE ");
					defineIdentifierClause(sqlstring);
					deleteAllQuery = sqlstring.toString();
				}
			}
		}
		return deleteAllQuery;
	}

	String getDeleteElemQuery() {
		if (deleteElemQuery == null) {
			synchronized (this) {
				if (deleteElemQuery == null) {
					StringBuffer sqlstring
							= new StringBuffer("DELETE FROM " + tableName + " WHERE ");
					defineIdentifierClause(sqlstring);
					defineIndexClause(sqlstring);
					defineElemClause(sqlstring, false);
					deleteElemQuery = sqlstring.toString();
				}
			}
		}
		return deleteElemQuery;
	}

	String getUpdateQuery() {
		if (updateQuery == null) {
			synchronized (this) {
				if (updateQuery == null) {
					StringBuffer sqlstring
							= new StringBuffer("UPDATE " + tableName + " SET ");
					if (colocated) {
						defineIdentifierSet(sqlstring);
						defineIndexSet(sqlstring);
						sqlstring.append(" WHERE ");
						defineElemClause(sqlstring, true);
					} else {
						defineElemSet(sqlstring);
						sqlstring.append(" WHERE ");
						defineIdentifierClause(sqlstring);
						defineIndexClause(sqlstring);
					}
					updateQuery = sqlstring.toString();
				}
			}
		}
		return updateQuery;
	}

	String getUpdateNullAllQuery() {
		if (updateNullAllQuery == null) {
			synchronized (this) {
				if (updateNullAllQuery == null) {
					StringBuffer sqlstring
							= new StringBuffer("UPDATE " + tableName + " SET ");
					defineIdentifierSetNull(sqlstring);
					defineIndexSetNull(sqlstring);
					sqlstring.append(" WHERE ");
					defineIdentifierClause(sqlstring);
					updateNullAllQuery = sqlstring.toString();
				}
			}
		}
		return updateNullAllQuery;
	}

	String getUpdateNullElemQuery() {
		if (updateNullElemQuery == null) {
			synchronized (this) {
				if (updateNullElemQuery == null) {
					StringBuffer sqlstring
							= new StringBuffer("UPDATE " + tableName + " SET ");
					defineIdentifierSetNull(sqlstring);
					defineIndexSetNull(sqlstring);
					sqlstring.append(" WHERE ");
					defineIdentifierClause(sqlstring);
					defineIndexClause(sqlstring);
					defineElemClause(sqlstring, false);
					updateNullElemQuery = sqlstring.toString();
				}
			}
		}
		return updateNullElemQuery;
	}

	// IMPLEMENTATION OF METHODS FROM THE PClassMapping INTERFACE

	/**
	 * It creates a new PBinding.
	 * @return	The new PBinding with lifecycle state set to
	 * 			LIFECYCLE_NOTBOUND.
	 */
	public PBinding createPBinding() throws PException {
		return new RdbGenClassBinding(this);
	}

	/**
	 * It returns an iterator over all pname of persitent objects availlable in
	 * this class. When you use the returned iterator you must leave the
	 * connection opened.
	 * @param conn is the connection which permits to access to the support
	 * @return the iterator over PName objects
	 * @exception PExceptionProtocol	A problem occured during the request
	 * 									that is not related to I/O.
	 * @exception PExceptionIO	A problem occured during the request.
	 */
	public Iterator getPNameIterator(Object conn)
			throws PExceptionProtocol, PExceptionIO {
		Connection sqlconn = getConnection(conn);
		try {
			PreparedStatement ps = sqlconn.prepareStatement(getSelectAllNamesQuery());
			return new RdbGenClassPNGIterator(ps, ps.executeQuery(),
											  identifierFields, this);
		} catch (SQLException e) {
			throw new PExceptionIO(e);
		}
	}

	/**
	 * It specifies if the given kind of PMapper is supported by this
	 * PClassMapping. In this case, all kinds of RDB mappers are supported.
	 * @param mappername	The name defining the PMapper kind.
	 * @return	true if this kind of mapper is supported.
	 */
	public boolean isConform(String mappername) {
		return mappername.equals("rdb");
	}

	// IMPLEMENTATION OF METHODS FROM THE RdbGenClassProp INTERFACE

	/**
	 * Defines the table name property of a generic class instance mapped to
	 * a RDB.
	 * @param tn	The associated type.
	 */
	public void setTableName(String tn) {
		tableName = tn;
	}

	public void setColocated(boolean colocated) {
		this.colocated = colocated;
	}

	public void setColocatedMaster(boolean cm) {
		this.colocatedMaster = cm;
	}

	public void setReadOnly(boolean readonly) {
		this.readOnly = readonly;
	}

	/**
	 * Defines the type converter property of a generic class instance mapped to
	 * a RDB.
	 * @param tc	The associated type converter.
	 */
	public void setTypeConverter(RdbAdapter tc) {
		typeConverter = tc;
		if (typeConverter != null) {
			useBatch = typeConverter.supportBatchPreparedStatement();
		}
	}

	// IMPLEMENTATION OF LOCAL METHODS

	/**
	 * It appends to the given sqlstring the definition of columns that
	 * corresponds to the given fields.
	 * @param fields	The fields for which the columns must be defined.
	 * @param sqlstring	The SQL string to which to append these column
	 * 					definitions.
	 * @param start		The first field from which to start for defining the
	 * 					columns.
	 */
	void defineColumns(FieldDesc[] fields, StringBuffer sqlstring, short start)
			throws PExceptionProtocol {
		RdbFieldDesc field;
		for (short i = start; i < fields.length; i++) {
			field = (RdbFieldDesc) fields[i];
			sqlstring.append(field.columnName);
			sqlstring.append(" ");
			sqlstring.append(typeConverter.getSqlType(field.type, true, field.size));
			sqlstring.append(", ");
		}
	}

	/**
	 * It appends to the given sqlstring the names of the columns that
	 * composes the primary key.
	 * @param fields	The fields for which the columns must belong to the
	 * 					primary key.
	 * @param sqlstring	The SQL string to which to append these column names.
	 * @param firstname	It specifies that the first name has not already been
	 * 					defined for the primary key.
	 * @return 	The returned boolean value specifies if a column name was
	 * 			already added into the sqlstring. The value depends on the
	 * 			firstname parameter and if column has been added during the
	 * 			method execution.
	 */
	boolean defineNames(FieldDesc[] fields, StringBuffer sqlstring,
						boolean firstname, short start) {
		RdbFieldDesc field;
		for (short i = start; i < fields.length; i++) {
			field = (RdbFieldDesc) fields[i];
			if (firstname)
				firstname = false;
			else
				sqlstring.append(", ");
			sqlstring.append(field.columnName);
		}
		return firstname;
	}

	/**
	 * It appends to the given sqlstring the names of all columns that
	 * composes the indexed element (some or all those that compose an element).
	 * @param sqlstring		The SQL string to which to append all column names.
	 * @param whichnames	It specifies the names to appends. This is an OR
	 * 						operation between the ID, ELEM, and INDEX flags.
	 */
	void defineAllNames(StringBuffer sqlstring, byte whichnames) {
		boolean firstname = true;
		if ((whichnames & ID) == ID) {
			firstname = defineNames(identifierFields, sqlstring, firstname,
									(short) 0);
		}
		if ((whichnames & ELEM) == ELEM) {
			//In case where the element is a reference to a class the elemFields
			// array contains at least two element which the first is the real
			// type of the element. Then the column description starts to the
			// index 1.
			firstname = defineNames(elemFields, sqlstring, firstname,
									(short) (elemFields.length == 1 ? 0 : 1));
		}
		if ((whichnames & INDEX) == INDEX) {
			firstname = defineNames(indexFields, sqlstring, firstname,
									(short) 0);
		}
	}

	/**
	 * It appends to the given sqlstring the expression for matching a GC
	 * identifier represented by the given PName.
	 * @param sqlstring	The involved SQL string.
	 */
	void defineIdentifierClause(StringBuffer sqlstring) {
		if (identifierFields.length == 1) {
			sqlstring.append("(");
			sqlstring.append(((RdbFieldDesc) identifierFields[0]).columnName);
			sqlstring.append(ASSIGNVAL);
			sqlstring.append(")");
			return;
		}
		// The identifier is a composite name
		for (short i = 0; i < identifierFields.length; i++) {
			if (i != 0) {
				sqlstring.append(" ");
				sqlstring.append(SEPAND);
				sqlstring.append(" ");
			}
			sqlstring.append("(");
			sqlstring.append(((RdbFieldDesc) identifierFields[i]).columnName);
			sqlstring.append(ASSIGNVAL);
			sqlstring.append(")");
		}
	}

	/**
	 * It appends to the given sqlstring the expression for matching a GC
	 * identifier represented by the given PName.
	 * @param sqlstring	The involved SQL string.
	 */
	void defineIndexClause(StringBuffer sqlstring) {
		for (short i = 0; i < indexFields.length; i++) {
			sqlstring.append(" ");
			sqlstring.append(SEPAND);
			sqlstring.append(" ");
			sqlstring.append("(");
			sqlstring.append(((RdbFieldDesc) indexFields[i]).columnName);
			sqlstring.append(ASSIGNVAL);
			sqlstring.append(")");
		}
	}

	/**
	 * It appends to the given sqlstring the expression for matching a GC
	 * identifier represented by the given PName.
	 * @param sqlstring	The involved SQL string.
	 */
	void defineElemSet(StringBuffer sqlstring) {
		if (elemFields.length == 1) {
			sqlstring.append(((RdbFieldDesc) elemFields[0]).columnName);
			sqlstring.append(ASSIGNVAL);
		} else if (elemFields.length == 2) {
			sqlstring.append(((RdbFieldDesc) elemFields[1]).columnName);
			sqlstring.append(ASSIGNVAL);
		} else {
			// This is a composite reference
			for (int i = 1; i < elemFields.length; i++) {
				if (i != 1) {
					sqlstring.append(SEPSCOL);
				}
				sqlstring.append(((RdbFieldDesc) elemFields[i]).columnName);
				sqlstring.append(ASSIGNVAL);
			}
		}
	}

	/**
	 * It appends to the given sqlstring the expression for setting the GC
	 * identifier to null.
	 * @param sqlstring	The involved SQL string.
	 */
	void defineIdentifierSetNull(StringBuffer sqlstring) {
		if (identifierFields.length == 1) {
			sqlstring.append(((RdbFieldDesc) identifierFields[0]).columnName);
			sqlstring.append(ASSIGN_NULL);
		} else {
			// This is a composite identifier
			for (int i = 1; i < identifierFields.length; i++) {
				if (i != 1) {
					sqlstring.append(SEPSCOL);
				}
				sqlstring.append(((RdbFieldDesc) identifierFields[i]).columnName);
				sqlstring.append(ASSIGN_NULL);
			}
		}
	}

	/**
	 * It appends to the given sqlstring the expression for setting the GC
	 * identifier.
	 * @param sqlstring	The involved SQL string.
	 */
	void defineIdentifierSet(StringBuffer sqlstring) {
		if (identifierFields.length == 1) {
			sqlstring.append(((RdbFieldDesc) identifierFields[0]).columnName);
			sqlstring.append(ASSIGNVAL);
		} else {
			// This is a composite identifier
			for (int i = 0; i < identifierFields.length; i++) {
				if (i != 0) {
					sqlstring.append(SEPSCOL);
				}
				sqlstring.append(((RdbFieldDesc) identifierFields[i]).columnName);
				sqlstring.append(ASSIGNVAL);
			}
		}
	}

	/**
	 * It appends to the given sqlstring the expression for setting a GC
	 * index to null.
	 * @param sqlstring	The involved SQL string.
	 */
	void defineIndexSetNull(StringBuffer sqlstring) {
		for (int i = 1; i < indexFields.length; i++) {
			sqlstring.append(SEPSCOL);
			sqlstring.append(((RdbFieldDesc) indexFields[i]).columnName);
			sqlstring.append(ASSIGN_NULL);
		}
	}

	/**
	 * It appends to the given sqlstring the expression for setting the GC
	 * indexexes.
	 * @param sqlstring	The involved SQL string.
	 */
	void defineIndexSet(StringBuffer sqlstring) {
		for (int i = 1; i < indexFields.length; i++) {
			sqlstring.append(SEPSCOL);
			sqlstring.append(((RdbFieldDesc) indexFields[i]).columnName);
			sqlstring.append(ASSIGNVAL);
		}
	}

	/**
	 * It appends to the given sqlstring the expression for matching a GC
	 * identifier represented by the given PName.
	 * @param pname		The PName representing the identifier of the GC to
	 * 					access.
	 * @exception PExceptionProtocol	It is raised when an encoding type for
	 * 									one of the identifier fields is wrong.
	 * @exception PExceptionTyping		It is raised when an encoding type for
	 * 									one of the fields is wrong.
	 */
	int defineIdentifierVal(PName pname, PreparedStatement pstmt, int paramindex)
			throws PException, SQLException, IOException {
		if (identifierFields.length == 1) {
			if (!pname.codingSupported(
					CTHelper.ptc2ct(identifierFields[0].type.getTypeCode()))) {
				throw new PExceptionNaming("Coding format not supported: "
										   + CTHelper.ptc2ct(identifierFields[0].type.getTypeCode()));
			}
			defineValEncode(pname, pstmt, paramindex,
							identifierFields[0],
							((RdbFieldDesc) identifierFields[0]).columnType);
			paramindex++;
		} else {
			// The identifier is a composite name
			PNameGetter png = (PNameGetter) pname.encodeAbstract();
			for (short i = 0; i < identifierFields.length; i++) {
				defineValPng(png, pstmt, paramindex,
							 identifierFields[i],
							 ((RdbFieldDesc) identifierFields[i]).compositeName);
				paramindex++;
			}
		}
		return paramindex;
	}

	/**
	 * It appends to the given sqlstring the expression for defining the list
	 * of values.
	 * @param pie		The PIndexedElem giving access to the index of this
	 * 					element.
	 * @exception PExceptionProtocol	It is raised when an encoding type for
	 * 									one of the identifier fields is wrong.
	 * @exception PExceptionTyping		It is raised when an encoding type for
	 * 									one of the fields is wrong.
	 */
	int defineElemVal(PIndexedElem pie, PreparedStatement pstmt, int paramindex)
			throws PException, SQLException, IOException {
		if (elemFields.length == 1) {
			defineValBasic(pie, pstmt, paramindex,
						   elemFields[0].type.getTypeCode());
			paramindex++;
		} else if (elemFields.length == 2) {
			defineValEncode(pie.pieGetRefElem(), pstmt, paramindex,
							elemFields[1],
							((RdbFieldDesc) elemFields[1]).columnType);
			paramindex++;
		} else {
			// This is a composite reference
			PNameGetter png = (PNameGetter) pie.pieGetRefElem().encodeAbstract();
			for (int i = 1; i < elemFields.length; i++) {
				defineValPng(png, pstmt, paramindex,
							 elemFields[i],
							 ((RdbFieldDesc) elemFields[i]).compositeName);
				paramindex++;
			}
		}
		return paramindex;
	}

	/**
	 * It appends to the given sqlstring the expression for matching a GC
	 * identifier represented by the given PName.
	 * @param sqlstring	The involved SQL string.
	 */
	void defineElemClause(StringBuffer sqlstring, boolean isfirst) {
		if (elemFields.length == 1) {
			if (!isfirst) {
				sqlstring.append(" ");
				sqlstring.append(SEPAND);
				sqlstring.append(" ");
			}
			sqlstring.append("(");
			sqlstring.append(((RdbFieldDesc) elemFields[0]).columnName);
			sqlstring.append(ASSIGNVAL);
			sqlstring.append(")");
		} else {
			// The element is a reference (composite or single)
			for (short i = 1; i < elemFields.length; i++) {
				if (!isfirst || i > 1) {
					sqlstring.append(" ");
					sqlstring.append(SEPAND);
					sqlstring.append(" ");
				}
				sqlstring.append("(");
				sqlstring.append(((RdbFieldDesc) elemFields[i]).columnName);
				sqlstring.append(ASSIGNVAL);
				sqlstring.append(")");
			}
		}
	}


	/**
	 * It appends to the given sqlstring the expression for defining the list
	 * of values.
	 * @param pie		The PIndexedElem giving access to the index of this
	 * 					element.
	 * @exception PExceptionProtocol	It is raised when an encoding type for
	 * 									one of the identifier fields is wrong.
	 * @exception PExceptionTyping		It is raised when an encoding type for
	 * 									one of the fields is wrong.
	 */
	int defineIndexVal(PIndexedElem pie, PreparedStatement pstmt, int paramindex)
			throws PException, SQLException, IOException {
		for (int i = 0; i < indexFields.length; i++) {
			defineValIndex(pie, pstmt, paramindex,
						   indexFields[i],
						   ((RdbFieldDesc) indexFields[i]).name);
			paramindex++;
		}
		return paramindex;
	}

	/**
	 * @param pie		The PIndexedElem being the getter for the index fields
	 * 					of a PIndexedElem. It is null if the involved getter
	 * 					is PName one.
	 */
	void defineValBasic(PIndexedElem pie, PreparedStatement pstmt,
						int paramindex, int typecode)
			throws PException, SQLException, IOException {
		switch (typecode) {
		case PType.TYPECODE_BOOLEAN:
			typeConverter.setBoolean(pstmt, paramindex, pie.pieGetBooleanElem());
			break;
		case PType.TYPECODE_OBJBOOLEAN:
			typeConverter.setOboolean(pstmt, paramindex, pie.pieGetObooleanElem());
			break;
		case PType.TYPECODE_BYTE:
			typeConverter.setByte(pstmt, paramindex, pie.pieGetByteElem());
			break;
		case PType.TYPECODE_OBJBYTE:
			typeConverter.setObyte(pstmt, paramindex, pie.pieGetObyteElem());
			break;
		case PType.TYPECODE_CHAR:
			typeConverter.setChar(pstmt, paramindex, pie.pieGetCharElem());
			break;
		case PType.TYPECODE_OBJCHAR:
			typeConverter.setOchar(pstmt, paramindex, pie.pieGetOcharElem());
			break;
		case PType.TYPECODE_SHORT:
			typeConverter.setShort(pstmt, paramindex, pie.pieGetShortElem());
			break;
		case PType.TYPECODE_OBJSHORT:
			typeConverter.setOshort(pstmt, paramindex, pie.pieGetOshortElem());
			break;
		case PType.TYPECODE_INT:
			typeConverter.setInt(pstmt, paramindex, pie.pieGetIntElem());
			break;
		case PType.TYPECODE_OBJINT:
			typeConverter.setOint(pstmt, paramindex, pie.pieGetOintElem());
			break;
		case PType.TYPECODE_LONG:
			typeConverter.setLong(pstmt, paramindex, pie.pieGetLongElem());
			break;
		case PType.TYPECODE_OBJLONG:
			typeConverter.setOlong(pstmt, paramindex, pie.pieGetOlongElem());
			break;
		case PType.TYPECODE_FLOAT:
			typeConverter.setFloat(pstmt, paramindex, pie.pieGetFloatElem());
			break;
		case PType.TYPECODE_OBJFLOAT:
			typeConverter.setOfloat(pstmt, paramindex, pie.pieGetOfloatElem());
			break;
		case PType.TYPECODE_DOUBLE:
			typeConverter.setDouble(pstmt, paramindex, pie.pieGetDoubleElem());
			break;
		case PType.TYPECODE_OBJDOUBLE:
			typeConverter.setOdouble(pstmt, paramindex, pie.pieGetOdoubleElem());
			break;
		case PType.TYPECODE_STRING:
			typeConverter.setString(pstmt, paramindex, pie.pieGetStringElem());
			break;
		case PType.TYPECODE_DATE:
			typeConverter.setDate(pstmt, paramindex, pie.pieGetDateElem(), ((RdbFieldDesc) elemFields[0]).columnType);
			break;
		case PType.TYPECODE_BYTEARRAY:
			typeConverter.setByteArray(pstmt, paramindex, pie.pieGetByteArrayElem());
			break;
		case PType.TYPECODE_CHARARRAY:
			typeConverter.setCharArray(pstmt, paramindex, pie.pieGetCharArrayElem());
			break;
		case PType.TYPECODE_BIGDECIMAL:
			typeConverter.setBigDecimal(pstmt, paramindex, pie.pieGetBigDecimalElem());
			break;
		case PType.TYPECODE_SERIALIZED:
			typeConverter.setSerialized(pstmt, paramindex, pie.pieGetSerializedElem());
			break;
		default :
			throw new PExceptionTyping("Wrong field type.");
		}
	}

	/**
	 * @param pname		The PIndexedElem being the getter for the index fields
	 * 					of a PIndexedElem. It is null if the involved getter
	 * 					is PName one.
	 */
	void defineValEncode(PName pname, PreparedStatement pstmt,
						 int paramindex, FieldDesc fd, String sqlcoltype)
			throws PException, SQLException, IOException {
		switch (fd.type.getTypeCode()) {
		case PType.TYPECODE_BYTE:
			typeConverter.setByte(pstmt, paramindex, pname.encodeByte());
			break;
		case PType.TYPECODE_OBJBYTE:
			typeConverter.setObyte(pstmt, paramindex, pname.encodeObyte());
			break;
		case PType.TYPECODE_CHAR:
			typeConverter.setChar(pstmt, paramindex, pname.encodeChar());
			break;
		case PType.TYPECODE_OBJCHAR:
			typeConverter.setOchar(pstmt, paramindex, pname.encodeOchar());
			break;
		case PType.TYPECODE_SHORT:
			typeConverter.setShort(pstmt, paramindex, pname.encodeShort());
			break;
		case PType.TYPECODE_OBJSHORT:
			typeConverter.setOshort(pstmt, paramindex, pname.encodeOshort());
			break;
		case PType.TYPECODE_INT:
			typeConverter.setInt(pstmt, paramindex, pname.encodeInt());
			break;
		case PType.TYPECODE_OBJINT:
			typeConverter.setOint(pstmt, paramindex, pname.encodeOint());
			break;
		case PType.TYPECODE_LONG:
			typeConverter.setLong(pstmt, paramindex, pname.encodeLong());
			break;
		case PType.TYPECODE_OBJLONG:
			typeConverter.setOlong(pstmt, paramindex, pname.encodeOlong());
			break;
		case PType.TYPECODE_STRING:
			typeConverter.setString(pstmt, paramindex, pname.encodeString());
			break;
		case PType.TYPECODE_DATE:
			typeConverter.setDate(pstmt, paramindex, pname.encodeDate(), ((RdbFieldDesc) fd).columnType);
			break;
		case PType.TYPECODE_BYTEARRAY:
			typeConverter.setByteArray(pstmt, paramindex, pname.encode());
			break;
		case PType.TYPECODE_CHARARRAY:
			typeConverter.setCharArray(pstmt, paramindex, pname.encodeCharArray());
			break;
		default :
			throw new PExceptionTyping("Wrong field type.");
		}
	}

	/**
	 * @param png		The PIndexedElem being the getter for the index fields
	 * 					of a PIndexedElem. It is null if the involved getter
	 * 					is PName one.
	 */
	void defineValPng(PNameGetter png, PreparedStatement pstmt,
					  int paramindex, FieldDesc fd, String compfield)
			throws PException, SQLException, IOException {

		switch (fd.type.getTypeCode()) {
		case PType.TYPECODE_BYTE:
			typeConverter.setByte(pstmt, paramindex, png.pngetByteField(compfield, null));
			break;
		case PType.TYPECODE_OBJBYTE:
			typeConverter.setObyte(pstmt, paramindex, png.pngetObyteField(compfield, null));
			break;
		case PType.TYPECODE_CHAR:
			typeConverter.setChar(pstmt, paramindex, png.pngetCharField(compfield, null));
			break;
		case PType.TYPECODE_OBJCHAR:
			typeConverter.setOchar(pstmt, paramindex, png.pngetOcharField(compfield, null));
			break;
		case PType.TYPECODE_SHORT:
			typeConverter.setShort(pstmt, paramindex, png.pngetShortField(compfield, null));
			break;
		case PType.TYPECODE_OBJSHORT:
			typeConverter.setOshort(pstmt, paramindex, png.pngetOshortField(compfield, null));
			break;
		case PType.TYPECODE_INT:
			typeConverter.setInt(pstmt, paramindex, png.pngetIntField(compfield, null));
			break;
		case PType.TYPECODE_OBJINT:
			typeConverter.setOint(pstmt, paramindex, png.pngetOintField(compfield, null));
			break;
		case PType.TYPECODE_LONG:
			typeConverter.setLong(pstmt, paramindex, png.pngetLongField(compfield, null));
			break;
		case PType.TYPECODE_OBJLONG:
			typeConverter.setOlong(pstmt, paramindex, png.pngetOlongField(compfield, null));
			break;
		case PType.TYPECODE_STRING:
			typeConverter.setString(pstmt, paramindex, png.pngetStringField(compfield, null));
			break;
		case PType.TYPECODE_DATE:
			typeConverter.setDate(pstmt, paramindex, png.pngetDateField(compfield, null), ((RdbFieldDesc) fd).columnType);
			break;
		case PType.TYPECODE_BYTEARRAY:
			typeConverter.setByteArray(pstmt, paramindex, png.pngetByteArrayField(compfield, null));
			break;
		case PType.TYPECODE_CHARARRAY:
			typeConverter.setCharArray(pstmt, paramindex, png.pngetCharArrayField(compfield, null));
			break;
		default :
			throw new PExceptionTyping("Wrong field type.");
		}
	}

	/**
	 * @param pie		The PIndexedElem being the getter for the index fields
	 * 					of a PIndexedElem. It is null if the involved getter
	 * 					is PName one.
	 */
	void defineValIndex(PIndexedElem pie, PreparedStatement pstmt,
						int paramindex, FieldDesc fd, String indexfield)
			throws PException, SQLException, IOException {
		switch (fd.type.getTypeCode()) {
		case PType.TYPECODE_BYTE:
			typeConverter.setByte(pstmt, paramindex, pie.pieGetByteIndexField(indexfield));
			break;
		case PType.TYPECODE_OBJBYTE:
			typeConverter.setObyte(pstmt, paramindex, pie.pieGetObyteIndexField(indexfield));
			break;
		case PType.TYPECODE_CHAR:
			typeConverter.setChar(pstmt, paramindex, pie.pieGetCharIndexField(indexfield));
			break;
		case PType.TYPECODE_OBJCHAR:
			typeConverter.setOchar(pstmt, paramindex, pie.pieGetOcharIndexField(indexfield));
			break;
		case PType.TYPECODE_SHORT:
			typeConverter.setShort(pstmt, paramindex, pie.pieGetShortIndexField(indexfield));
			break;
		case PType.TYPECODE_OBJSHORT:
			typeConverter.setOshort(pstmt, paramindex, pie.pieGetOshortIndexField(indexfield));
			break;
		case PType.TYPECODE_INT:
			typeConverter.setInt(pstmt, paramindex, pie.pieGetIntIndexField(indexfield));
			break;
		case PType.TYPECODE_OBJINT:
			typeConverter.setOint(pstmt, paramindex, pie.pieGetOintIndexField(indexfield));
			break;
		case PType.TYPECODE_LONG:
			typeConverter.setLong(pstmt, paramindex, pie.pieGetLongIndexField(indexfield));
			break;
		case PType.TYPECODE_OBJLONG:
			typeConverter.setOlong(pstmt, paramindex, pie.pieGetOlongIndexField(indexfield));
			break;
		case PType.TYPECODE_STRING:
			typeConverter.setString(pstmt, paramindex, pie.pieGetStringIndexField(indexfield));
			break;
		case PType.TYPECODE_DATE:
			typeConverter.setDate(pstmt, paramindex, pie.pieGetDateIndexField(indexfield), ((RdbFieldDesc) fd).columnType);
			break;
/*        case PType.TYPECODE_BYTEARRAY:
            byte[] ba = pie.pieGetByteArrayIndexField(indexfield);
            if (ba != null) {
                pstmt.setBytes(paramindex, ba);
            } else {
                pstmt.setNull(paramindex, typeConverter.getSqlTypeCode(PTypeSpace.BYTEARRAY, sqlcoltype));
            }
            break;
        case PType.TYPECODE_CHARARRAY:
            char[] ca = pie.pieGetCharArrayIndexField(indexfield);
            if (ca != null) {
                pstmt.set ???(paramindex, ca);
            } else {
                pstmt.setNull(paramindex, typeConverter.getSqlTypeCode(PTypeSpace.CHARARRAY, sqlcoltype));
            }
            break;*/
		default :
			throw new PExceptionTyping("Wrong field type.");
		}
	}

	/**
	 * It intialises the class mapping. It especially assigns a mapper as well
	 * as a meta-object describing either a class or a generic class.
	 * @param mapper	The mapper to be assigned to this class mapping.
	 * @param metaclass	The meta-object describing the associated class. It may
	 * 					be null.
	 */
	public void init(PMappingCallback mapper, MetaObject metaclass) throws PException {
		super.init(mapper, metaclass);
	}

	/**
	 * It passes the values of the index and the generic class element through
	 * a PIndexedElem to the accessor associated to this generic class instance.
	 * @param rs		The ResultSet that points to the current indexed element
	 * 					to add to the MI through the given PGenClassAccessor
	 * 					accessor.
	 * @param accessor	The relevant generic class accessor associated with a
	 * 					valid MI.
	 * @param png		The PNameGetter that gives access to the values of the
	 * 					fields of a composite name, when the elements of this
	 * 					generic class instance are composite references. It is
	 * 					null if it is not a composite name.
	 * @exception PExceptionProtocol	It is raised if there is a problem with
	 * 									the DSI structure with respect to the MI
	 * 									one.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							accessing the SQL ResultSet.
	 */
	void passValuesToAccessor(Object conn, ResultSet rs,
							  PGenClassAccessor accessor, PNameGetter png)
			throws PException {
		PIndexedElem pie = accessor.createPIndexedElem();
		if (logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG, "passValuesToAccessor");
		try {
			// The begining of the index depends of the size of the element:
			// If the element is the primitive element then the size is 1
			// If the element is a reference field then the first element of the
			// elemFields array is just a tag.
			short indexStart = (short)
					(elemFields.length == 1 ? 2 : elemFields.length);

			for (short i = 0; i < indexFields.length; i++) {
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG,
							   "index[" + i + "]:" + indexFields[i].name);
				switch (indexFields[i].type.getTypeCode()) {
				case PType.TYPECODE_BYTE:
					pie.pieSetByteIndexField(indexFields[i].name,
											 typeConverter.getByte(rs, i + indexStart));
					break;
				case PType.TYPECODE_OBJBYTE:
					pie.pieSetObyteIndexField(indexFields[i].name,
											  typeConverter.getObyte(rs, i + indexStart));
					break;
				case PType.TYPECODE_CHAR:
					pie.pieSetCharIndexField(indexFields[i].name,
											 typeConverter.getChar(rs, i + indexStart));
					break;
				case PType.TYPECODE_OBJCHAR:
					pie.pieSetOcharIndexField(indexFields[i].name,
											  typeConverter.getOchar(rs, i + indexStart));
					break;
				case PType.TYPECODE_SHORT:
					pie.pieSetShortIndexField(indexFields[i].name,
											  typeConverter.getShort(rs, i + indexStart));
					break;
				case PType.TYPECODE_OBJSHORT:
					pie.pieSetOshortIndexField(indexFields[i].name,
											   typeConverter.getOshort(rs, i + indexStart));
					break;
				case PType.TYPECODE_INT:
					pie.pieSetIntIndexField(indexFields[i].name,
											typeConverter.getInt(rs, i + indexStart));
					break;
				case PType.TYPECODE_OBJINT:
					pie.pieSetOintIndexField(indexFields[i].name,
											 typeConverter.getOint(rs, i + indexStart));
					break;
				case PType.TYPECODE_LONG:
					pie.pieSetLongIndexField(indexFields[i].name,
											 typeConverter.getLong(rs, i + indexStart));
					break;
				case PType.TYPECODE_OBJLONG:
					pie.pieSetOlongIndexField(indexFields[i].name,
											  typeConverter.getOlong(rs, i + indexStart));
					break;
				case PType.TYPECODE_STRING:
					pie.pieSetStringIndexField(indexFields[i].name,
											   typeConverter.getString(rs, i + indexStart));
					break;
				case PType.TYPECODE_DATE:
					pie.pieSetDateIndexField(indexFields[i].name,
											 typeConverter.getDate(rs, i + indexStart, ((RdbFieldDesc) indexFields[i]).columnType));
					break;
				default :
					throw new PExceptionProtocol(
							"Wrong type for an index field of a GenClass");
				}
			}
			if (logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG,
						   "element type:" + elemFields[0].type.getJormName());
			switch (elemFields[0].type.getTypeCode()) {
			case PType.TYPECODE_BOOLEAN:
				pie.pieSetBooleanElem(typeConverter.getBoolean(rs, 1));
				break;
			case PType.TYPECODE_OBJBOOLEAN:
				pie.pieSetObooleanElem(typeConverter.getOboolean(rs, 1));
				break;
			case PType.TYPECODE_CHAR:
				pie.pieSetCharElem(typeConverter.getChar(rs, 1));
				break;
			case PType.TYPECODE_OBJCHAR:
				pie.pieSetOcharElem(typeConverter.getOchar(rs, 1));
				break;
			case PType.TYPECODE_BYTE:
				pie.pieSetByteElem(typeConverter.getByte(rs, 1));
				break;
			case PType.TYPECODE_OBJBYTE:
				pie.pieSetObyteElem(typeConverter.getObyte(rs, 1));
				break;
			case PType.TYPECODE_SHORT:
				pie.pieSetShortElem(typeConverter.getShort(rs, 1));
				break;
			case PType.TYPECODE_OBJSHORT:
				pie.pieSetOshortElem(typeConverter.getOshort(rs, 1));
				break;
			case PType.TYPECODE_INT:
				pie.pieSetIntElem(typeConverter.getInt(rs, 1));
				break;
			case PType.TYPECODE_OBJINT:
				pie.pieSetOintElem(typeConverter.getOint(rs, 1));
				break;
			case PType.TYPECODE_LONG:
				pie.pieSetLongElem(typeConverter.getLong(rs, 1));
				break;
			case PType.TYPECODE_OBJLONG:
				pie.pieSetOlongElem(typeConverter.getOlong(rs, 1));
				break;
			case PType.TYPECODE_FLOAT:
				pie.pieSetFloatElem(typeConverter.getFloat(rs, 1));
				break;
			case PType.TYPECODE_OBJFLOAT:
				pie.pieSetOfloatElem(typeConverter.getOfloat(rs, 1));
				break;
			case PType.TYPECODE_DOUBLE:
				pie.pieSetDoubleElem(typeConverter.getDouble(rs, 1));
				break;
			case PType.TYPECODE_OBJDOUBLE:
				pie.pieSetOdoubleElem(typeConverter.getOdouble(rs, 1));
				break;
			case PType.TYPECODE_STRING:
				pie.pieSetStringElem(typeConverter.getString(rs, 1));
				break;
			case PType.TYPECODE_DATE:
				pie.pieSetDateElem(typeConverter.getDate(rs, 1, ((RdbFieldDesc) elemFields[0]).columnType));
				break;
			case PType.TYPECODE_CHARARRAY:
				pie.pieSetCharArrayElem(typeConverter.getCharArray(rs, 1));
				break;
			case PType.TYPECODE_BYTEARRAY:
				pie.pieSetByteArrayElem(typeConverter.getByteArray(rs, 1));
				break;
			case PType.TYPECODE_SERIALIZED:
				pie.pieSetSerializedElem((Serializable)
						(new ObjectInputStream(rs.getBlob(1).getBinaryStream()))
						.readObject());
				break;
			case PType.TYPECODE_BIGDECIMAL:
				pie.pieSetBigDecimalElem(typeConverter.getBigDecimal(rs, 1));
				break;
			case PType.TYPECODE_REFERENCE:
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG,
							   "reference size:" + (elemFields.length - 1));
				if (elemFields.length == 2) {
					pie.pieSetRefElem(decodeSingle(elemFields[1], rs, 1));
				} else {
					pie.pieSetRefElem(elemNameCoder.decodeAbstract(png, null));
				}
				break;
			default :
				throw new PExceptionProtocol(
						"Wrong type for an element field of a GenClass");
			}
		} catch (SQLException e) {
			throw new PExceptionIO(e);
		} catch (IOException e) {
			throw new PExceptionIO(e);
		} catch (ClassNotFoundException e) {
			throw new PExceptionIO(e);
		}
		accessor.paAdd(pie, conn);
	}

	/**
	 * It retrieves a PNameGetter that gives access to the values of the
	 * fields of a composite name, when the elements of this generic class
	 * instance are composite references. It yields null if it is not a
	 * composite name.
	 * @return	The allocated PNameGetter.
	 */
	PNameGetter getPNameGetter(ResultSet rs) {
		if (elemFields.length > 2)
			return new RdbGenClassNameGetter(null, rs, elemFields);
		return null;
	}

	PName decodeSingle(FieldDesc fd, ResultSet resultSet, int idx)
			throws PExceptionProtocol, PExceptionNaming, SQLException {
		switch (fd.type.getTypeCode()) {
		case PType.TYPECODE_BYTE:
			return elemNameCoder.decodeByte(typeConverter.getByte(resultSet, idx));
		case PType.TYPECODE_CHAR:
			return elemNameCoder.decodeChar(typeConverter.getChar(resultSet, idx));
		case PType.TYPECODE_SHORT:
			return elemNameCoder.decodeShort(typeConverter.getShort(resultSet, idx));
		case PType.TYPECODE_INT:
			return elemNameCoder.decodeInt(typeConverter.getInt(resultSet, idx));
		case PType.TYPECODE_LONG:
			return elemNameCoder.decodeLong(typeConverter.getLong(resultSet, idx));
		case PType.TYPECODE_OBJBYTE:
			return elemNameCoder.decodeObyte(typeConverter.getObyte(resultSet, idx));
		case PType.TYPECODE_OBJCHAR:
			return elemNameCoder.decodeOchar(typeConverter.getOchar(resultSet, idx));
		case PType.TYPECODE_OBJSHORT:
			return elemNameCoder.decodeOshort(typeConverter.getOshort(resultSet, idx));
		case PType.TYPECODE_OBJINT:
			return elemNameCoder.decodeOint(typeConverter.getOint(resultSet, idx));
		case PType.TYPECODE_OBJLONG:
			return elemNameCoder.decodeOlong(typeConverter.getOlong(resultSet, idx));
		case PType.TYPECODE_STRING:
			return elemNameCoder.decodeString(typeConverter.getString(resultSet, idx));
		case PType.TYPECODE_DATE:
			return elemNameCoder.decodeDate(typeConverter.getDate(resultSet, idx, ((RdbFieldDesc) fd).columnType));
		case PType.TYPECODE_BYTEARRAY:
			return elemNameCoder.decode(typeConverter.getByteArray(resultSet, idx));
		case PType.TYPECODE_CHARARRAY:
			return elemNameCoder.decodeCharArray(typeConverter.getCharArray(resultSet, idx));
		default:
			throw new PExceptionProtocol(
					"The jorm naming does not manage this type "
					+ fd.type.getJormName());
		}
	}
}
