/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.generator;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.compiler.lib.JormCompilerParameter;
import org.objectweb.jorm.generator.api.MOP;
import org.objectweb.jorm.generator.api.MOPFactory;
import org.objectweb.jorm.metainfo.api.Mapping;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.Package;
import org.objectweb.jorm.metainfo.api.Reference;
import org.objectweb.jorm.metainfo.api.ReferenceMapping;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbClassMultiMapping;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbExternalTable;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbClassMapping;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbPrimitiveElementMapping;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbTable;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbJoin;

import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;

/**
 *
 */
public class RdbMOPFactory implements MOPFactory {
	// IMPLEMENTATION OF METHODS FROM THE MOPFactory INTERFACE

	/**
	 * This method must return a new instance of a MappingMOP
	 */
	public MOP createMappingMOP(JormCompilerParameter cp, String mappername) throws PException {
		RdbMappingMOP res = new RdbMappingMOP();
		return res;
	}

	/**
	 * This method must return a new instance of a BindingMOP
	 */
	public MOP createBindingMOP(JormCompilerParameter cp, String mappername) throws PException {
		RdbBindingMOP res = new RdbBindingMOP();
		return res;
	}

	/**
	 * Creates a generation context that can be shared between binding and
	 * mapping generation for a particular class.
	 */
	public Object createClassGenerationContext(Mapping mo) throws PException {
		RdbGenInfos rgi = new RdbGenInfos();
		org.objectweb.jorm.mapper.rdb.metainfo.RdbClassMapping cm = (org.objectweb.jorm.mapper.rdb.metainfo.RdbClassMapping) mo.getClassMapping();

		// define the main table
		RdbGenTable mainTable = new RdbGenTable(cm.getRdbTable());
		mainTable.genInfos = rgi;
		rgi.genTables.put(mainTable.tableName, mainTable);
		rgi.referenceTable = mainTable;
		rgi.readableColumns.addAll(mainTable.tableColumns);

		if (cm instanceof org.objectweb.jorm.mapper.rdb.metainfo.RdbClassMultiMapping) {
			RdbClassMultiMapping rcm = (RdbClassMultiMapping) cm;
			for (Iterator it = rcm.getRdbExternalTables().iterator(); it.hasNext();) {
				RdbExternalTable moTable = (RdbExternalTable) it.next();
				RdbGenTable curTable = new RdbGenTable(moTable, mainTable);
				curTable.genInfos = rgi;
				rgi.genTables.put(curTable.tableName, curTable);
				for (int i = 0; i < curTable.tableColumns.size(); i++) {
					RdbGenColumn rgc = (RdbGenColumn) curTable.tableColumns.get(i);
					if (rgc.joinCol == null) {
						rgi.readableColumns.add(rgc);
					}
				}
			}
		}
		// define GenRef of the class identifier
		int ndId = 0;
		NameDef nd = (NameDef) cm.getIdentifierMapping().getLinkedMO();
		//System.out.println("Define GenRef of class");
		rgi.genId = defineGenRef(nd, null, ndId++, cm, rgi);

		for (Iterator it = rgi.genTables.values().iterator(); it.hasNext();) {
			((RdbGenTable) it.next()).setGenerateUpdate(rgi);
		}

		// define GenRef for each reference field
		for (Iterator it = cm.getReferenceMappings().iterator(); it.hasNext();) {
			nd = (NameDef) ((ReferenceMapping) it.next()).getLinkedMO();
			String fn = ((Reference) nd.getParent()).getName();
			rgi.genRefs.put(fn, defineGenRef(nd, fn, ndId++, cm, rgi));
		}
		return rgi;
	}

	// PRIVATE METHODS

	/**
	 * Defines the RdbGenDef associated either with the identifier of this class
	 * mapping, or with one of its references.
	 * @param nd    The NameDef associated with the identifier or a reference.
	 * @param cnid  The composite name identifier allocated for code generation
	 *              of references.
	 * @return  The created RdbGenRef describing the identifier or a reference.
	 */
	private RdbGenRef defineGenRef(NameDef nd, String fn, int cnid,
								   RdbClassMapping cm, RdbGenInfos rgi)
			throws PException {
		RdbGenRef rgr = new RdbGenRef();
		rgr.nd = nd;
		rgr.fieldName = fn;
		if (nd.isFieldName()) {
			RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
					cm.getPrimitiveElementMapping(nd.getFieldName());
			if (pem == null) {
				throw new PException(
						"Unable to find the mapping of the field "
						+ nd.getFieldName());
			}
			RdbGenTable rgt = (RdbGenTable)
					rgi.genTables.get(((RdbTable) pem.getParent()).getName());
			rgr.refColumn = rgt.getColumn(pem.getName());
			if (rgt.joins != null) {
				getJoin(rgt, pem, nd.getFieldName(), rgr);
			}
		} else { // this is a NameRef
			//System.out.println("nd("+fn+") cnId=" + cnid);
			rgr.cnId = cnid;
			rgr.cnName = nd.getNameRef().getCompositeName().getName();
			rgr.cnPackage = ((Package) nd.getNameRef().getCompositeName().getParent()).getName();
			rgr.cnFieldColumns = new TreeMap();
			Iterator itp = nd.getNameRef().getProjection().entrySet().iterator();
			while (itp.hasNext()) {
				Map.Entry me = (Map.Entry) itp.next();
				String cnfield = (String) me.getKey();
				String clafn = (String) me.getValue();
				RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
						cm.getPrimitiveElementMapping(clafn);
				if (pem == null) {
					throw new PException(
							"Unable to find the mapping of the field "
							+ me.getValue());
				}
				RdbGenTable rgt = (RdbGenTable)
						rgi.genTables.get(((RdbTable) pem.getParent()).getName());
				RdbGenColumn c = rgt.getColumn(pem.getName());
				rgr.cnFieldColumns.put(cnfield, c);
				if (rgr.join == null && rgt.joins != null) {
					getJoin(rgt, pem, clafn, rgr);
				}
			}
		}
		return rgr;
	}

	private void getJoin(RdbGenTable rgt,
						 RdbPrimitiveElementMapping pem,
						 String fn,
						 RdbGenRef rgr) {
		//System.out.println("getJoin: fn: " + rgr.fieldName + " / col: " + pem.getName() + " / rgt: " + rgt.tableName);
		if (rgt.joins.size() == 1) {
			//System.out.println("return: fn: " + rgr.fieldName + " / col: " + pem.getName() + " / join: " + ((RdbGenJoin) rgt.joins.get(0)).joinIdx);
			rgr.join = (RdbGenJoin) rgt.joins.get(0);
		}
		Iterator it = pem.getPrimitiveElementByRdbJoin().entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry me2 = (Map.Entry) it.next();
			PrimitiveElement pe = (PrimitiveElement) me2.getValue();
			if (pe.getName().equals(fn)) {
				//System.out.println("return fn: " + rgr.fieldName + " / col: " + pem.getName() + " / joinName: " + ((RdbJoin) me2.getKey()).getName() + " / join: " + rgt.jn2join.get(((RdbJoin) me2.getKey()).getName()));
				rgr.join = (RdbGenJoin)
						rgt.jn2join.get(((RdbJoin) me2.getKey()).getName());
			}
		}
	}
}
