/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.metainfo.lib;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.Package;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.ScalarField;
import org.objectweb.jorm.metainfo.api.ClassRef;
import org.objectweb.jorm.metainfo.api.ClassProject;
import org.objectweb.jorm.metainfo.lib.BasicMetaObject;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.HashMap;
import java.util.Map;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

/**
 * BasicClass is an implementation of the Class interface defined
 * in the meta information system. This object is defined by a list
 * of fields, a name, a list of mapping.
 * @author	X. Spengler
 */
public class BasicClass extends BasicMetaObject implements Class, Loggable {
	/**
	 * abstractClass boolean defines if the current class is an abstract
	 * class or not.
	 */
	private boolean abstractClass;
	private String fileName;
	/**
	 * specifies that this class is inherited by one of those that are defined
	 * within this MetaInfo Manager.
	 */
	private boolean isInherited = false;

	/**
	 * name is the name of the current class object.
	 */
	private String name;

	/**
	 * nameDefs defines a list of NameDef that the user can potentially
	 * used for the mapping.
	 * key: the name of the NameDef object
	 * value: a NameDef object
	 */

	private ArrayList nameDefs;

	/**
	 * classRefs defines the list of the fields reprensented by
	 * a reference to a Class.
	 * key: the name of the ClassRef object
	 * value: the ClassRef object
	 */
	private Map classRefs;

	/**
	 * genClassRefs defines the list of the fields which are defined
	 * by a Generic Class.
	 * key: the name of the GenClassRef object
	 * value: the GenClassRef object
	 */
	private Map genClassRefs;

	/**
	 * primitiveElements defines the list of the fields with a primitive type.
	 * key: the name of the PrimitiveElement object
	 * value: the PrimitiveElement object
	 */
	private Map primitiveElements;

	/**
	 * hiddenFields defines the list of the hiddenfields.
	 * These fields are scalar fields.
	 * key: the name of the hiddenField object
	 * value: the scalarfield object
	 */
	private Map hiddenFields;

	/**
	 * inheritedClasses represents the list of the inherited classes by
	 * the current class object.
	 * element: a Class MetaObject object
	 */
	private ArrayList inheritedClasses;

	/**
	 * This HashMap associates a project name with a vector of ClassProject
	 * objects.
	 * The notion of project allows the developer to define multiple mappings
	 * for a class per mapper type.
	 */
	private HashMap classProjects;

	/**
	 * a simple logger to log
	 */
	private transient Logger logger;

	/**
	 * Builds a new Class object.
	 * This object is mainly defined by its name, if the class is abstract or
	 * not, and its parent (a schema).
	 *
	 * @param className     the name of the current class
	 * @param parent    the parent of the current object
	 */
	public BasicClass(String className,
					  MetaObject parent) {
		super(parent);
		this.name = className;
		this.abstractClass = false;

		classRefs = new HashMap();
		genClassRefs = new HashMap();
		nameDefs = new ArrayList();
		primitiveElements = new HashMap();
		hiddenFields = new HashMap();
		inheritedClasses = new ArrayList();
		classProjects = new HashMap();
		this.fileName = getFQName().replace('.',
											System.getProperty("file.separator").charAt(0)) + ".pd";
	}

	///////////////////////////////////////////////////////////////////
	//                      from Class interface
	///////////////////////////////////////////////////////////////////

	/**
	 * Allows to retrieve the file name of this class MO.
	 * @return  The file name.
	 */
	public String getFileName() {
		return fileName;
	}

	/**
	 * Allows to assign the file name of this class MO.
	 * @param filename  The file name.
	 */
	public void setFileName(String filename) {
		this.fileName = filename;
	}

	/**
	 * Allows to know if the current class is an abstract class or not.
	 *
	 * @return	true, if the current class is an abstract class, and false,
	 *          if the current class is not abstract
	 */
	public boolean isAbstract() {
		return abstractClass;
	}

	public void setAbstract(boolean isAbstract) {
		abstractClass = isAbstract;
	}

	/**
	 * Allows to know the name of the current class object.
	 * @return	the string representation of the class name
	 */
	public String getName() {
		return name;
	}

	public String getFQName() {
		String packName = null;
		MetaObject parent = getParent();
		if (parent instanceof Package)
			packName = ((Package) parent).getName();
		if (packName == null || packName.length() == 0)
			return name;
		else
			return packName + "." + name;
	}

	/**
	 * Allows to obtain an existing Field.
	 * A TypedElement is a class field, and this method finds a field with its
	 * name (fieldName). This method is a generic method to return a
	 * TypedElement object. The field can be hidden or not.
	 *
	 * @param fieldName	the name of the class field
	 * @return	a field object. If the field does not exist, null is returned.
	 */
	public TypedElement getTypedElement(String fieldName) {
		if (logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG, "try to return field (" + fieldName + ")");

		TypedElement res = null;
		// this method finds in the 3 structures where is the field
		// fieldName could be a PrimitiveElement
		res = (TypedElement) primitiveElements.get(fieldName);
		if (res != null) {
			if (logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG, "the field is a PrimitiveElement object");
			return res;
		}
		res = (TypedElement) hiddenFields.get(fieldName);
		if (res != null) {
			if (logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG, "the field is an hiddenfield object");
			return res;
		}
		res = (TypedElement) classRefs.get(fieldName);
		// fieldName could be a reference to a class
		if (classRefs.get(fieldName) != null) {
			if (logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG, "the field is a ClassRef object");
			return res;
		}
		res = (TypedElement) genClassRefs.get(fieldName);
		// fieldName could be a reference to a generic class
		if (res != null) {
			if (logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG, "the field is a GenClassRef object");
			return (GenClassRef) genClassRefs.get(fieldName);
		}
		Iterator it = iterateInheritedClass();
		while (it.hasNext() && res == null) {
			res = ((Class) it.next()).getTypedElement(fieldName);
		}
		// attention, the fieldname does not exist. where is it ?
		if (res == null && logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.WARN,
					   "the field (" + fieldName + ") is not defined in the current class");
		return res;
	}

	/**
	 * Removes a typed element.
	 * @param fieldName is the name of the field to be removed
	 * @return the removed element or null if the field has not been found.
	 */
	public TypedElement removeTypedElement(String fieldName) {
		TypedElement res = (TypedElement) primitiveElements.remove(fieldName);
		if (res != null) {
			return res;
		}
		res = (TypedElement) hiddenFields.remove(fieldName);
		if (res != null) {
			return res;
		}
		res = (TypedElement) classRefs.remove(fieldName);
		// fieldName could be a reference to a class
		if (classRefs.get(fieldName) != null) {
			return res;
		}
		res = (TypedElement) genClassRefs.remove(fieldName);
		return res;
	}

	/**
	 * Creates a new NameDef object for the current Class object.
	 * Here we can define if the namedef is defined outside jorm or not.
	 * In the case where the namedef is already defined for the current class,
	 * the old one is used and returned.
	 *
	 * @return	a new object used to describe the PName projection for
	 *          the current Class
	 */
	public NameDef createNameDef() {
		if (logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG,
					   "create a new NameDef () for the current Class '" + name + "'");
		NameDef newNameDef = new BasicNameDef(this);
		((Loggable) newNameDef).setLogger(logger);
		nameDefs.add(newNameDef);
		return newNameDef;
	}

	/**
	 * register the namedef in the currentclass with the mapping name passed as parameter.
	 * @param mappingname the name of the mapping
	 * @param nd the namedef to register.
	 */
	/*public void registerNameDef(String mappingName, NameDef nd) {
if (nameDefs.get(mappingName) != null) {

logger.log(BasicLevel.ERROR,
"attention, the mappingName (" + mappingName +
") is already binded to an existing NameDef in the class (" + name + ")");
} else {
nameDefs.put(mappingName, nd);
}
}*/

	/**
	 * Returns a NameDef object.
	 * If the namedef does not exist, null is returned.
	 *
	 * @param name   the name of a NameDef object.
	 * @return  a namedef object, else null is returned.
	 */
	public NameDef getNameDef(String name) {
		Iterator iter = nameDefs.iterator();
		while (iter.hasNext()) {
			NameDef nameDef = (NameDef) iter.next();
			if (nameDef.getName().equals(name)) {
				return nameDef;
			}
		}
		return null;
	}

	/**
	 * Returns an iterator on existing NameDef for the current Class.
	 * This method returns an empty iterator if no namedef exists in this Class.
	 *
	 * @return	an iterator on NameDef, or an empty iterator if no namedef is
	 *          defined
	 */

	public Iterator iterateNameDef() {
		return nameDefs.iterator();
	}

	/**
	 * Allows to obtain a new PrimitiveElement from a class field name an its
	 * type.
	 * A PrimitiveElement is a Class field. If the class field already exists,
	 * it is returned.
	 *
	 * @param fieldName	the name of the primitive field
	 * @param type      the PType of the primitive field
	 * @return	a new PrimitiveElement object, or the existing one if already
	 *          defined
	 */
	public PrimitiveElement createPrimitiveElement(String fieldName,
												   PType type, int size) {
		if (logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG, "create a new PrimitiveElement (" +
										 fieldName + ") for the current Class (" + name + ")");

		// try to get the fieldname from the current structure
		PrimitiveElement result = (PrimitiveElement)
				primitiveElements.get(fieldName);
		if (result != null) {
			// be careful, the primitive element already exists, we return it.
			logger.log(BasicLevel.WARN,
					   "Try to create an existing PrimitiveElement (" +
					   fieldName + "), return existing one");
		} else {
			// oh, the primitive element does not exist, we must create it.
			result = new BasicPrimitiveElement(fieldName, type, size, this);
			((Loggable) result).setLogger(logger);
			primitiveElements.put(fieldName, result);
		}
		return result;
	}

	/**
	 * Creates a new scalar field and adds it to the list of hidden fields.
	 * If the field already exists, it is returned, else a new one is built.
	 *
	 * @param fieldName	the name of the field to create
	 * @param type	    the type of the field to create
	 * @return	a new ScalarField object if this object does not already exist
	 */
	public ScalarField createHiddenField(String fieldName, PType type, int size) {
		if (logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG,
					   "create a new HiddenField (" + fieldName + ") for the current class (" + name + ")");

		// verify if the field is not yet in the database
		ScalarField result = (ScalarField) hiddenFields.get(fieldName);
		if (result != null) {
			logger.log(BasicLevel.WARN,
					   "Try to create an existing HiddenField (" +
					   fieldName + "), return existing one");
		} else {
			// else create a new one
			result = new BasicScalarField(fieldName, type, size, this);
			((Loggable) result).setLogger(logger);
			hiddenFields.put(fieldName, result);
		}
		return result;
	}

	/**
	 * Allows to obtain a new ClassRef from a field name an its class.
	 * A ClassRef object is a field which is represented by a reference to a
	 * Class. If the field already exists, it is returned.
	 *
	 * @param fieldName	the name of the field
	 * @param class_	the string representation of the class which represents
	 *                  the reference of the field
	 * @return	a new ClassRef object, or the existing one if already defined
	 */
	public ClassRef createClassRef(String fieldName, Class class_) {
		if (logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG, "create a new ClassRef (" + fieldName +
										 ") for the current Class (" + name + ")");

		// just verify if the fielname does not exist in the current structure.
		ClassRef result = (ClassRef) classRefs.get(fieldName);
		if (result != null) {
			// be careful, the object exists, we return it
			logger.log(BasicLevel.WARN,
					   "attention, try to create an existing ClassRef (" +
					   fieldName + "), return existing one");
		} else {
			// we have to create a new object
			result = new BasicClassRef(fieldName, class_, this);
			((Loggable) result).setLogger(logger);
			classRefs.put(fieldName, result);
		}
		return result;
	}

	/**
	 * Allows to obtain a new GenClassRef from a field name an its generic
	 * class.
	 * A GenClassRef object is a field which is represented by a
	 * reference to a generic class. If the field already exists,
	 * it is returned.
	 *
	 * @param fieldName	the name of the field
	 * @param genName	the GenClass object which represents the reference of
	 *                  the field
	 * @return	a new GenClassRef object, or the existing one if already
	 *          defined
	 */
	public GenClassRef createGenClassRef(String fieldName, String genName) {
		if (logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG, "create a new GenClassRef (" +
										 genName + ") from field (" + fieldName + ") from Class (" +
										 name + ")");

		// verify if the fieldname already exists
		GenClassRef result = (GenClassRef) genClassRefs.get(fieldName);
		if (result != null) {
			// be careful, the object already exists, we return it
			logger.log(BasicLevel.WARN,
					   "attention, try to create an existing GenClassRef (" + fieldName +
					   "), return existing one");
		} else {
			// we have to create a new object and return it
			result =
					new BasicGenClassRef(genName, fieldName, this);
			((Loggable) result).setLogger(logger);
			genClassRefs.put(fieldName, result);
		}
		return result;
	}

	/**
	 * Returns an iterator on existing class fields.
	 * If no field is defined, an empty iterator is returned.
	 * This iterator contains PrimitiveElements, ClassRef and GenClassRef
	 * objects. The top level object is TypedElement.
	 *
	 * @return	an iterator on fields, or null if there is no defined field
	 */
	public Iterator iterateField() {
		return listField().iterator();
	}

	/**
	 * returns an ArrayList which contains the current fields.
	 * @return  a list of current fields
	 */
	protected List listField() {
		ArrayList allFields = new ArrayList();
		allFields.addAll(primitiveElements.values());
		allFields.addAll(classRefs.values());
		allFields.addAll(genClassRefs.values());
		Collections.sort(allFields, FieldComparator.instance);
		return allFields;
	}

	/**
	 * Returns an iterator on existing class fields + fields from super classes.
	 * This iterator contains PrimitiveElement, ClassRef and GenClassRef
	 * objects. This iterator does not contains the hiddenfield of the class.
	 *
	 * @return	an iterator on all defined fields, plus fields from super
	 */
	public Iterator iterateAllField() {
		return getAllFields().iterator();
	}

	public Collection getAllFields() {
		List allFields = listField();
		// and add the fields from the inherited classes
		Iterator iter = iterateInheritedClass();
		while (iter.hasNext())
			allFields.addAll(((BasicClass) iter.next()).listField());
		Collections.sort(allFields, FieldComparator.instance);
		return allFields;
	}


	/**
	 * Provides an iterator over the all the hiddenfield of the class including inherited fields.
	 * @return the iterator embeding scalarfield object describing the hiddenfields.
	 */
	public Iterator iterateAllHiddenField() {
		return getAllHiddenFields().iterator();
	}

	public Collection getAllHiddenFields() {
		ArrayList allFields = new ArrayList();
		Iterator iter = hiddenFields.values().iterator();

		while (iter.hasNext()) {
			allFields.add(iter.next());
		}
		// and add the fields from the inherited classes
		iter = iterateInheritedClass();
		while (iter.hasNext()) {
			Class aClass = (Class) iter.next();
			Iterator itera = aClass.iterateAllHiddenField();
			while (itera.hasNext()) {
				allFields.add(itera.next());
			}
		}
		return allFields;
	}

	/**
	 * return an iterator over the hiddenfield of the class definition.
	 *
	 * @return the iterator containing the hiddenfield.
	 */
	public Iterator iterateHiddenField() {
		return hiddenFields.values().iterator();
	}

	/**
	 * retrieve an hiddenfield from its name.
	 *
	 * @return the hiddenfield or null if the field does not exist.
	 */
	public ScalarField getHiddenField(String fieldName) {
		if (logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG,
					   "try to return an hidden field (" + fieldName + ")");
		// this method finds in the 3 structures where is the field
		return (ScalarField) hiddenFields.get(fieldName);

	}

	/**
	 * Allows to obtain all the inherited classes from the current class object.
	 * This iterator contains Class objects. In the case where there is no
	 * inherited class, an empty iterator is returned.
	 *
	 * @return	an iterator on inherited classes (Class objets). If no object
	 *          exists, an empty iterator is returned.
	 */
	public Iterator iterateInheritedClass() {
		return inheritedClasses.iterator();
	}

	/**
	 * Adds a new inherited class to the list of the inherited classes for the
	 * current class object.
	 * If the class already exists, nothing is done.
	 *
	 * @param class_	the inherited class to add to inherited classes list
	 */
	public void addInheritedClass(Class class_) {
		if (class_ != null) {
			inheritedClasses.add(class_);
			class_.setInherited();
		}
	}

	/**
	 * Returns the number of inherited classes for the current Class object.
	 *
	 * @return  the number of inherited classes
	 */
	public int getInheritedClassNumber() {
		return inheritedClasses.size();
	}

	/**
	 * Constructs an Map with inherited class
	 *
	 * @param table the hashtable to build
	 */
	public void addAllInheritedClass(Map table) {
		Iterator iter = iterateInheritedClass();
		String all[] = new String[inheritedClasses.size() + 1];
		int count = 0;
		all[count] = name;
		count++;

		while (iter.hasNext()) {
			Class theClass = ((Class) iter.next());
			theClass.addAllInheritedClass(table);

			all[count] = theClass.getFQName();
			count++;

		}
		table.put(name, all);
	}

	/**
	 * Returns the number of fields which are references to classes.
	 *
	 * @return  the number of fieldRef
	 */
	public int getFieldRefNumber() {
		return classRefs.size();
	}

	/**
	 * Specifies that this class is inherited by another one within the
	 * set of compiled ones.
	 */
	public void setInherited() {
		isInherited = true;
	}

	/**
	 * Tests if this class is inherited by another one.
	 */
	public boolean isInherited() {
		return isInherited;
	}

	/**
	 * Returns a ClassProject object.
	 *
	 * @param  projectName       a project name.
	 * @return a ClassProject object.
	 */
	public ClassProject getClassProject(String projectName) {
		return (ClassProject) classProjects.get(projectName);
	}

	/**
	 * Returns a collection of ClassProject objects.
	 * @return	 a collection.
	 */
	public Collection getClassProjects() {
		return classProjects.values();
	}

	/**
	 * return an iterator over the ClassProject objects.
	 *
	 * @return an iterator.
	 */
	public Iterator iterateClassProjects() {
		return classProjects.values().iterator();
	}

	/**
	 * Adds a ClassProject object.
	 *
	 * @param  projectName       a project name,
	 *         classProject	    a ClassProject object.
	 */
	public void addClassProject(String projectName, ClassProject classProject) {
		if (!classProjects.containsKey(projectName)) {
			classProjects.put(projectName, classProject);
		}
	}

	/**
	 * Creates a new ClassProject object.
	 *
	 * @param projectName	the name of the project.
	 * @return a new ClassProject object, or an existing one.
	 */
	public ClassProject createClassProject(String projectName) {
		if (logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG,
					   "create a new ClassProject for the current Class (" + name +
					   ") projectName=<" + projectName + ">");

		// verifies whether the classProject already exists
		ClassProject cp = (ClassProject) classProjects.get(projectName);
		if (cp == null) {
			cp = new BasicClassProject(projectName, this);
			cp.setLogger(this.logger);
			classProjects.put(projectName, cp);
			return cp;
		} else {
			logger.log(BasicLevel.DEBUG,
					   "return the existing ClassProject " + projectName);
			return cp;
		}
	}

	/**
	 * Removes a ClassProject instance.
	 * @param projectname is the name of removed ClassProject
	 * @return The removed ClassProject
	 */
	public ClassProject removeClassProject(String projectname) {
		return (ClassProject) classProjects.remove(projectname);
	}

	///////////////////////////////////////////////////////////////////
	//                      from Loggable interface
	///////////////////////////////////////////////////////////////////

	/**
	 * Defines a logger object.
	 *
	 * @param logger	the logger object
	 */
	public void setLogger(Logger logger) {
		this.logger = logger;
		for (Iterator it = iterateNameDef(); it.hasNext();) {
			((Loggable) it.next()).setLogger(logger);
		}
		for (Iterator it = this.iterateClassProjects(); it.hasNext();) {
			((Loggable) it.next()).setLogger(logger);
		}
		for (Iterator it = this.iterateAllField(); it.hasNext();) {
			((Loggable) it.next()).setLogger(logger);
		}
		for (Iterator it = this.iterateAllHiddenField(); it.hasNext();) {
			((Loggable) it.next()).setLogger(logger);
		}
	}

	/**
	 * Defines the logger factory to obtain new logger.
	 *
	 * @param loggerFactory	the LoggerFactory object to obtain a logger object
	 */
	public void setLoggerFactory(LoggerFactory loggerFactory) {
	}

	public Logger getLogger() {
		return logger;
	}

	public LoggerFactory getLoggerFactory() {
		return null;
	}
}

