/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */
package org.objectweb.jorm.metainfo.lib;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassMapping;
import org.objectweb.jorm.metainfo.api.ClassRef;
import org.objectweb.jorm.metainfo.api.CompositeName;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.NameRef;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.Reference;
import org.objectweb.jorm.metainfo.api.Package;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.metainfo.api.PrimitiveElementMapping;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.MappingPrinter;
import org.objectweb.jorm.metainfo.api.GenClassMapping;
import org.objectweb.jorm.metainfo.api.Mapping;
import org.objectweb.jorm.metainfo.api.ClassProject;
import org.objectweb.jorm.metainfo.api.MappingFactory;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.wrapper.printwriter.PrintStreamImpl;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;
import java.util.Collection;

/**
 * This class permits to print the jorm meta information in a PrintStream
 * (System.out for example).
 *
 * @author S.Chassande-Barrioz
 */
public class MetaInfoPrinter {

	ArrayList mappingprinters = null;
	public final static String TAB = "    ";

	public MetaInfoPrinter() {
		mappingprinters = new ArrayList();
	}

	public MetaInfoPrinter(Manager m) {
		Collection c = m.getMappingFactories();
		mappingprinters = new ArrayList(c.size());
		Iterator it = c.iterator();
		while (it.hasNext()) {
			mappingprinters.add(
					((MappingFactory) it.next()).createMappingPrinter());
		}
	}

	public void addMappingPrinter(MappingPrinter vmp) {
		mappingprinters.add(vmp);
	}

	public void print(String p, MetaObject mo, Logger logger) {
		print(p, mo, new PrintStreamImpl(logger));
	}

	public void print(String p, MetaObject mo, PrintStream out) {
		if (mo instanceof Package)
			print(p, (Package) mo, out);
		else if (mo instanceof Class)
			print(p, (Class) mo, out);
		else if (mo instanceof CompositeName)
			print(p, (CompositeName) mo, out);
		else if (mo instanceof TypedElement)
			print(p, (TypedElement) mo, out);
		else if (mo instanceof NameDef)
			print(p, (NameDef) mo, out);
		else if (mo instanceof NameRef)
			print(p, (NameRef) mo, out);
		else if (mo instanceof ClassMapping)
			print(p, (ClassMapping) mo, out);
		else if (mo instanceof GenClassMapping)
			print(p, (GenClassMapping) mo, out);
		else if (mo instanceof PrimitiveElementMapping)
			print(p, (PrimitiveElementMapping) mo, out);
		else
			out.print("Unknown meta object: " + mo);
	}

	public void print(String p, Manager mgr, PrintStream out) {
		out.println("Print the meta information of the mnanager: " + mgr);
		for (Iterator it = mgr.getPackages().iterator(); it.hasNext();) {
			print(p, (Package) it.next(), out);
		}
	}

	public void print(String p, Package sc, PrintStream out) {
		for (Iterator it = sc.getClasses().iterator(); it.hasNext();) {
			print(p, (Class) it.next(), out);
		}
		for (Iterator it = sc.iterateCompositeName(); it.hasNext();) {
			print(p, (CompositeName) it.next(), out);
		}
	}

	public void print(String p, CompositeName cn, PrintStream out) {
		out.println(p + "* CompositeName " + cn.getFQName());
		Iterator it = cn.iterateAllField();
		while (it.hasNext()) {
			print(p, ((TypedElement) it.next()), out);
		}
	}

	public void print(String p, Class clazz, PrintStream out) {
		out.println(p + "* Class " + clazz.getFQName());
		out.println(p + "Persistent fields");

		Iterator it = clazz.iterateAllField();
		while (it.hasNext()) {
			print(p + TAB, (TypedElement) it.next(), out);
		}
		it = clazz.iterateAllHiddenField();
		out.println(p + "Hidden persistent fields");
		while (it.hasNext()) {
			print(p + TAB, (TypedElement) it.next(), out);
		}
		it = clazz.iterateNameDef();
		out.println(p + "NameDef of the class " + clazz.getName());
		while (it.hasNext()) {
			print(p + TAB, (NameDef) it.next(), out);
		}
		it = clazz.getClassProjects().iterator();
		while (it.hasNext()) {
			ClassProject cp = (ClassProject) it.next();
			String msg = p + "Project: " + cp.getProjectName() + ",";
			Iterator mit = cp.getMappings().iterator();
			if (!mit.hasNext()) {
				out.println(msg + " no mapping defined ");
			}
			while (mit.hasNext()) {
				Mapping m = (Mapping) mit.next();
				out.println(msg + " Mapper: " + m.getMapperName());
				print(p, m.getClassMapping(), out);
				Iterator gcmit = m.getGenClassMappings().iterator();
				while (gcmit.hasNext()) {
					print(p + TAB, (GenClassMapping) gcmit.next(), out);
				}
			}
		}
	}

	public void print(String p, TypedElement te, PrintStream out) {
		if (te == null) {
			out.print("ERROR: null field");
			return;
		}
		PType type = te.getType();
		String strtype;
		if (type == null)
			strtype = " type is null";
		else
			strtype = te.getType().getJormName();
		out.print(p + "- Field  " + strtype + " " + te.getName());
		if (te instanceof Reference)
			print(p, (Reference) te, out);
		else if (te instanceof PrimitiveElement)
			print(p, (PrimitiveElement) te, out);
		else
			out.print("ERROR: Unknon typedlement" + te);
	}

	public void print(String p, PrimitiveElement pe, PrintStream out) {
		if (pe.isScalar()) {
			out.println(" (scalar)");
		} else {
			out.println();
		}
	}

	public void print(String p, Reference ref, PrintStream out) {
		if (ref instanceof ClassRef)
			print((ClassRef) ref, out);
		else if (ref instanceof GenClassRef)
			print(p, (GenClassRef) ref, out);
		out.println(p + "NameDef of the reference " + ref.getName());
		for (Iterator it = ref.getRefNameDef().iterator(); it.hasNext();) {
			print(p + TAB, (NameDef) it.next(), out);
		}
	}

	public void print(String p, NameDef nd, PrintStream out) {
		if (nd.isFieldName())
			out.println(p + "fieldName: " + nd.getFieldName());
		else if (nd.isNameRef()) {
			print(p, nd.getNameRef(), out);
		}
	}

	public void print(String p, NameRef nr, PrintStream out) {
		out.println(p + "NameRef name: " + nr.getName() + " / Compsosite name: " + nr.getCompositeName());
		out.println(p + TAB + "Projection: " + nr.getCompositeName());
		Map proj = nr.getProjection();
		for (Iterator it = proj.entrySet().iterator(); it.hasNext();) {
			Map.Entry me = (Map.Entry) it.next();
			out.println(p + TAB + "composite field name: " + me.getKey()
						+ " / class field Name: " + me.getValue());
		}
	}

	public void print(ClassRef cr, PrintStream out) {
		out.println("ClassRef: " + cr.getClassName());
	}

	public void print(String p, GenClassRef gcr, PrintStream out) {
		out.println(p + "GenClass " + gcr.getName() + "hidden persistent fields");
		for (Iterator it = gcr.getHiddenFields().iterator(); it.hasNext();) {
			print(p + TAB, (TypedElement) it.next(), out);
		}
		out.println(p + "NameDef of the GenClass " + gcr.getName());
		for (Iterator it = gcr.getIdNameDef().iterator(); it.hasNext();) {
			print(p + TAB, (NameDef) it.next(), out);
		}
		out.println(p + "=====Gen class " + gcr.getName() + " element: begin");
		if (gcr.isPrimitive()) {
			out.println(p + "GenClassRef " + gcr.getName() + " has a PrimitveElement");
			print(p + TAB, (TypedElement) gcr.getPrimitiveElement(), out);
		} else if (gcr.isClassRef()) {
			out.println(p + "GenClassRef " + gcr.getName() + " has a PrimitveElement");
			print(p + TAB, (TypedElement) gcr.getClassRef(), out);
		} else if (gcr.isGenClassRef()) {
			out.println(p + "GenClassRef " + gcr.getName() + " has a GenClassRef");
			print(p + TAB, (TypedElement) gcr.getGenClassRef(), out);
		} else
			out.print("ERROR: unknown inner element type of the gen class: " + gcr.getName());
		out.println(p + "=====Gen class " + gcr.getName() + " element: end");

		out.println(p + "ClassMapping of the genclass " + gcr.getName());
	}

	public void print(String p, PrimitiveElementMapping vm, PrintStream out) {
		for (Iterator it = mappingprinters.iterator(); it.hasNext();) {
			MappingPrinter mp = (MappingPrinter) it.next();
			if (mp.canPrint(vm)) {
				mp.print(p + TAB, vm, out);
				return;
			}
		}
		out.println(p + "WARN: Unable to print the value mapping: " + vm);
	}

	public void print(String p, ClassMapping cm, PrintStream out) {
		for (Iterator it = mappingprinters.iterator(); it.hasNext();) {
			MappingPrinter mp = (MappingPrinter) it.next();
			if (mp.canPrint(cm)) {
				mp.print(p + TAB, cm, out);
				return;
			}
		}
		out.println(p + "WARN: Unable to print the class mapping: " + cm);
	}

	public void print(String p, GenClassMapping gcm, PrintStream out) {
		for (Iterator it = mappingprinters.iterator(); it.hasNext();) {
			MappingPrinter mp = (MappingPrinter) it.next();
			if (mp.canPrint(gcm)) {
				mp.print(p + TAB, gcm, out);
				return;
			}
		}
		out.println(p + "WARN: Unable to print the genclass mapping: " + gcm);
	}

}
