/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.naming.lib;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNamingContext;
import org.objectweb.jorm.naming.api.PNameManager;
import org.objectweb.jorm.type.api.PType;

import java.util.Date;

/**
 * This basic implementation of a PName delegates all calls to its methods
 * to the PNamingContext it references. The only call wich is not delegated is
 * to isNull and getPNamingContext.
 * @author	P. D?chamboux
 */
public abstract class BasicPName implements PName {

	public String toString() {
		return super.toString()
				+ " / pnc: " + (pnc == null
				? "no assigned "
				: (pnc.getPType() == null
				? "null PType"
				: pnc.getPType().getJormName()));
	}

	/**
	 * The PNamingContext that manages this PName
	 */
	protected transient PNameManager pnc = null;

	public void setPNamingContext(PNamingContext pnc) {
		this.pnc = pnc;
	}

	public boolean codingSupported(short codingtype) {
		return pnc.codingSupported(codingtype);
	}

	public byte[] encode() throws PExceptionNaming {
		return pnc.encode(this);
	}

	public Object encodeAbstract() throws PExceptionNaming, UnsupportedOperationException {
		return pnc.encodeAbstract(this);
	}

	public byte encodeByte() throws PExceptionNaming, UnsupportedOperationException {
		return pnc.encodeByte(this);
	}

	public Byte encodeObyte() throws PExceptionNaming, UnsupportedOperationException {
		return pnc.encodeObyte(this);
	}

	public char encodeChar() throws PExceptionNaming, UnsupportedOperationException {
		return pnc.encodeChar(this);
	}

	public Character encodeOchar() throws PExceptionNaming, UnsupportedOperationException {
		return pnc.encodeOchar(this);
	}

	public int encodeInt() throws PExceptionNaming, UnsupportedOperationException {
		return pnc.encodeInt(this);
	}

	public Integer encodeOint() throws PExceptionNaming, UnsupportedOperationException {
		return pnc.encodeOint(this);
	}

	public long encodeLong() throws PExceptionNaming, UnsupportedOperationException {
		return pnc.encodeLong(this);
	}

	public Long encodeOlong() throws PExceptionNaming, UnsupportedOperationException {
		return pnc.encodeOlong(this);
	}

	public short encodeShort() throws PExceptionNaming, UnsupportedOperationException {
		return pnc.encodeShort(this);
	}

	public Short encodeOshort() throws PExceptionNaming, UnsupportedOperationException {
		return pnc.encodeOshort(this);
	}

	public String encodeString() throws PExceptionNaming {
		return pnc.encodeString(this);
	}

	public char[] encodeCharArray() throws PExceptionNaming {
		return pnc.encodeCharArray(this);
	}

	public Date encodeDate() throws PExceptionNaming {
		return pnc.encodeDate(this);
	}

	public PName export(Object conn, PNamingContext pnc) throws PException {
		return pnc.export(conn, this);
	}

	public PName export(Object conn, PNamingContext pnc, Object hints)
			throws PException {
		return pnc.export(conn, this, hints);
	}

	public PNameManager getPNameManager() {
		return pnc;
	}

	public PType getPType() {
		return pnc.getPType();
	}

	public PName resolve(Object conn) throws UnsupportedOperationException, PException {
		if (pnc instanceof PNamingContext) {
			return ((PNamingContext) pnc).resolve(conn, this);
		}
		throw new UnsupportedOperationException();
	}

	public void unexport(Object conn) throws PException {
		pnc.unexport(conn, this);
	}

	public void unexport(Object conn, Object hints) throws PException {
		pnc.unexport(conn, this, hints);
	}
}
