/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.naming.lib;

import org.objectweb.jorm.naming.api.PExceptionCoder;
import org.objectweb.jorm.naming.api.PNCStringCoder;

import java.text.DateFormat;
import java.text.ParseException;
import java.util.Date;

/**
 * The coding format supported by this coder is the following one:
 *		- char is coded as is
 * 		  (example 'A' <-> "A"
 * 		- byte, short, int, long are coded to hexadecimal fixed size strings
 * 		  (example 395/short <-> "018B")
 * 		- string is coded with its encoded short length followed by the string
 * 		  as is
 * 		  (example "Hello world!" <-> "000CHello World!"
 * 		- date is coded with its encoded byte length followed by its string
 * 		  representation
 * 		  (example Date(837039928046) <-> "1407/10/96 4:5 PM, PDT"
 * @author	P. D?chamboux
 */
public class BasicStringCoder implements PNCStringCoder {
	private String toDecode = null;
	private int nextToDecode = 0;
	private StringBuffer toEncode = null;

	public BasicStringCoder() {
		toEncode = new StringBuffer(64);
	}

	public BasicStringCoder(String en) {
		toDecode = en;
	}

	private byte hex2int(char c) throws PExceptionCoder {
		if ((c >= '0') && (c <= '9'))
			return (byte) (c - '0');
		if ((c >= 'A') && (c <= 'F'))
			return (byte) (c - 'A' + 10);
		throw new PExceptionCoder("Cannot decode: wrong hexadecimal code.");
	}

	private void long2hexIntoSB(long l, char[] ev) {
		for (int i = ev.length - 1; i >= 0; i--) {
			byte b = (byte) (l & 0xF);
			if (b < 10)
				ev[i] = (char) (b + '0');
			else
				ev[i] = (char) (b - 10 + 'A');
			l >>= 4;
		}
		for (int i = 0; i < ev.length; i++)
			toEncode.append(ev[i]);
	}

	public byte getByte() throws PExceptionCoder {
		if (toDecode == null)
			throw new PExceptionCoder("Cannot decode: this is an encoder object.");
		if ((nextToDecode + 2) > toDecode.length())
			throw new PExceptionCoder("Cannot decode: no more data.");
		byte res = (byte) (hex2int(toDecode.charAt(nextToDecode++)) << 4);
		res += hex2int(toDecode.charAt(nextToDecode++));
		return res;
	}

	public Byte getObyte() throws PExceptionCoder {
		return new Byte(getByte());
	}

	public char getChar() throws PExceptionCoder {
		if (toDecode == null)
			throw new PExceptionCoder("Cannot decode: this is an encoder object.");
		if ((nextToDecode + 1) > toDecode.length())
			throw new PExceptionCoder("Cannot decode: no more data.");
		char res = toDecode.charAt(nextToDecode++);
		return res;
	}

	public Character getOchar() throws PExceptionCoder {
		return new Character(getChar());
	}

	public short getShort() throws PExceptionCoder {
		if (toDecode == null)
			throw new PExceptionCoder("Cannot decode: this is an encoder object.");
		if ((nextToDecode + 4) > toDecode.length())
			throw new PExceptionCoder("Cannot decode: no more data.");
		short res = 0;
		for (int i = 0; i < 4; i++) {
			res <<= 4;
			res += hex2int(toDecode.charAt(nextToDecode++));
		}
		return res;
	}

	public Short getOshort() throws PExceptionCoder {
		return new Short(getShort());
	}

	public int getInt() throws PExceptionCoder {
		if (toDecode == null)
			throw new PExceptionCoder("Cannot decode: this is an encoder object.");
		if ((nextToDecode + 8) > toDecode.length())
			throw new PExceptionCoder("Cannot decode: no more data.");
		int res = 0;
		for (int i = 0; i < 8; i++) {
			res <<= 4;
			res += hex2int(toDecode.charAt(nextToDecode++));
		}
		return res;
	}

	public Integer getOint() throws PExceptionCoder {
		return new Integer(getInt());
	}

	public long getLong() throws PExceptionCoder {
		if (toDecode == null)
			throw new PExceptionCoder("Cannot decode: this is an encoder object.");
		if ((nextToDecode + 16) > toDecode.length())
			throw new PExceptionCoder("Cannot decode: no more data.");
		long res = 0;
		for (int i = 0; i < 16; i++) {
			res <<= 4;
			res += hex2int(toDecode.charAt(nextToDecode++));
		}
		return res;
	}

	public Long getOlong() throws PExceptionCoder {
		return new Long(getLong());
	}

	public String getString() throws PExceptionCoder {
		if (toDecode == null)
			throw new PExceptionCoder("Cannot decode: this is an encoder object.");
		int len = getShort();
		if ((nextToDecode + len) > toDecode.length())
			throw new PExceptionCoder("Cannot decode: no more data.");
		String res = toDecode.substring(nextToDecode, nextToDecode + len);
		nextToDecode += len;
		return res;
	}

	public Date getDate() throws PExceptionCoder {
		if (toDecode == null)
			throw new PExceptionCoder("Cannot decode: this is an encoder object.");
		int len = getByte();
		if ((nextToDecode + len) > toDecode.length())
			throw new PExceptionCoder("Cannot decode: no more data.");
		try {
			Date res = DateFormat.getDateTimeInstance().parse(toDecode.substring(nextToDecode, nextToDecode + len));
			nextToDecode += len;
			return res;
		} catch (ParseException pe) {
			throw new PExceptionCoder(pe, "Cannot decode: wrong Date format.");
		}
	}

	public byte[] getByteArray() throws PExceptionCoder {
		if (toDecode == null)
			throw new PExceptionCoder("Cannot decode: this is an encoder object.");
		int len = getByte();
		if ((nextToDecode + len) > toDecode.length())
			throw new PExceptionCoder("Cannot decode: no more data.");
//TODO
		return new byte[0];
	}

	public char[] getCharArray() throws PExceptionCoder {
		if (toDecode == null)
			throw new PExceptionCoder("Cannot decode: this is an encoder object.");
		int len = getByte();
		if ((nextToDecode + len) > toDecode.length())
			throw new PExceptionCoder("Cannot decode: no more data.");
//TODO
		return new char[0];
	}

	public void putByte(byte val) throws PExceptionCoder {
		if (toEncode == null)
			throw new PExceptionCoder("Cannot encode: this is a decoder object.");
		long2hexIntoSB(val, new char[2]);
	}

	public void putObyte(Byte val) throws PExceptionCoder {
		putByte(val.byteValue());
	}

	public void putChar(char val) throws PExceptionCoder {
		if (toEncode == null)
			throw new PExceptionCoder("Cannot encode: this is a decoder object.");
		toEncode.append(val);
	}

	public void putOchar(Character val) throws PExceptionCoder {
		putChar(val.charValue());
	}

	public void putShort(short val) throws PExceptionCoder {
		if (toEncode == null)
			throw new PExceptionCoder("Cannot encode: this is a decoder object.");
		long2hexIntoSB(val, new char[4]);
	}

	public void putOshort(Short val) throws PExceptionCoder {
		putShort(val.shortValue());
	}

	public void putInt(int val) throws PExceptionCoder {
		if (toEncode == null)
			throw new PExceptionCoder("Cannot encode: this is a decoder object.");
		long2hexIntoSB(val, new char[8]);
	}

	public void putOint(Integer val) throws PExceptionCoder {
		putInt(val.intValue());
	}

	public void putLong(long val) throws PExceptionCoder {
		if (toEncode == null)
			throw new PExceptionCoder("Cannot encode: this is a decoder object.");
		long2hexIntoSB(val, new char[16]);
	}

	public void putOlong(Long val) throws PExceptionCoder {
		putLong(val.longValue());
	}

	public void putString(String val) throws PExceptionCoder {
		if (toEncode == null)
			throw new PExceptionCoder("Cannot encode: this is a decoder object.");
		if (val.length() > Short.MAX_VALUE)
			throw new PExceptionCoder("Cannot encode: String field too long.");
		putShort((short) val.length());
		toEncode.append(val);
	}

	public void putDate(Date val) throws PExceptionCoder {
		if (toEncode == null)
			throw new PExceptionCoder("Cannot encode: this is a decoder object.");
		String ds = DateFormat.getDateTimeInstance().format(val);
		if (ds.length() > Byte.MAX_VALUE)
			throw new PExceptionCoder("Cannot encode: Date field too long.");
		putByte((byte) ds.length());
		toEncode.append(ds);
	}

	public void putByteArray(byte[] va) throws PExceptionCoder {
		if (toEncode == null)
			throw new PExceptionCoder("Cannot encode: this is a decoder object.");
//TODO
	}

	public void putCharArray(char[] val) throws PExceptionCoder {
		if (toEncode == null)
			throw new PExceptionCoder("Cannot encode: this is a decoder object.");
//TODO
	}

	public String getStringCode() throws PExceptionCoder {
		if (toEncode == null)
			throw new PExceptionCoder("Cannot encode: this is a decoder object.");
		return toEncode.toString();
	}
}
