/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.xml2mi.api;

import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.ClassProject;
import org.objectweb.jorm.util.io.api.PathExplorer;
import org.objectweb.jorm.xml2mi.lib.ParserErrorHandler;
import org.objectweb.jorm.api.PException;
import org.w3c.dom.Element;

import java.util.Iterator;
import java.util.Collection;
import java.util.ArrayList;


/**
 * The Parser interface defines the behaviour of a JORM component used to
 * parse a set of JORM definition files.
 * A parser is able to parse XML files (these files may be passed by the pmc
 * command line), and build a schema of meta-objects, using a meta-information
 * manager component.
 * @author	P. Dechamboux, X. Spengler
 */
public interface Parser {
	/**
	 * Launches the parsing process. It reads and parses all the XML files
	 * and builds a schema of meta-objects for all of them.
	 * Before to use this method, the following object need to be set : Manager
	 * (with setMetaInfoManager), the PathExplorer (with setPathExplorer), the
	 * mapper name (with setMapperName), the mapping parser object (with setMappingParser).
	 * @param files	an iterator over the name the XML files to be parsed
	 */
	Collection parse(Iterator files) throws PException;

	/**
	 * Launches the parsing process.
	 * It reads and parse an XML file and builds a schema of meta-objects
	 * for it.
	 * Before to use this method, the following object need to be set : Manager
	 * (with setMetaInfoManager), the PathExplorer (with setPathExplorer), the
	 * mapper name (with setMapperName), the mapping parser object (with setMappingParser).
	 * @param file	the name of the XML file to be parsed
	 */
	MetaObject parse(String file) throws PException;


	void parseMapping(Element mappingElem, ClassProject classProject) throws PException;

	/**
	 * Assigns a meta-information manager to a parser in order from it to
	 * create the meta-information related to the XML files it parses.
	 * @param mim	the meta-information Manager to be associated to this parser
	 */
	void setMetaInfoManager(Manager mim);

	/**
	 * Assigns a PathExplorer object for locating files that have to be parsed.
	 * @param pathexpl	the PathExplorer to be used for file location
	 */
	void setPathExplorer(PathExplorer pathexpl);

	/**
	 * Sets the parser which is able to parse the mapping part of the XML file.
	 * @param mappername	the name of the mapper
	 * @param mappingParser	the parser which parses the mapping object
	 */
	void addMappingParser(String mappername,
						  MappingParser mappingParser) throws PException;

	/**
	 * Returns a MappingParser object.
	 * @param  mapperName    a mapper name,
	 * @return a MappingParser object.
	 */
	public MappingParser getMappingParser(String mapperName);

	/**
	 * Initializes the Parser. In order to compile XML files.
	 * @param dtdVerify true, the dtd must be used to validate xml files, else
	 *                  false
	 * @param dtds		A list of usefull DTD.
	 */
	void init(boolean dtdVerify, ArrayList dtds);

	/**
	 * Returns the exception which is reached during the parsing of a file.
	 * @return  an Exception object
	 */
	Exception getParserException();

	/**
	 * Allows to know the list of error the parser encounters.
	 * @return  a ParserErrorHandler object which countains all errors from
	 *          the xml parser
	 */
	ParserErrorHandler getParserErrorHandler();
}
