/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.xml2mi.lib;

import org.apache.xerces.parsers.DOMParser;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.compiler.api.PExceptionCompiler;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassProject;
import org.objectweb.jorm.metainfo.api.ScalarField;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.NameRef;
import org.objectweb.jorm.metainfo.api.Reference;
import org.objectweb.jorm.metainfo.api.Mapping;
import org.objectweb.jorm.metainfo.api.CompositeName;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.Package;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.metainfo.api.ClassRef;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.jorm.util.io.api.PathExplorer;
import org.objectweb.jorm.util.io.lib.DirJavaExplorer;
import org.objectweb.jorm.xml2mi.api.MappingParser;
import org.objectweb.jorm.xml2mi.api.Parser;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;

import java.io.InputStream;
import java.util.Map;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

/**
 * The BasicDomParser class implements the Parser interface.
 * It uses the DOM trees associated to the XML files that are parsed in order
 * to build the meta-objects.
 * @author      P. Dechamboux, X. Spengler
 */
public class BasicDomParser implements Parser, Loggable {
	/**
	 * The meta information manager for managing meta-objects within the
	 * context of a mapper.
	 */
	private Manager metaInfoManager;

	/**
	 * This hashmap associates a mapper name with a MappingParser object.
	 */
	private Map mappingParsers = new HashMap();

	/**
	 * A dom parser (here from xerses)
	 */
	private DOMParser parser;

	/**
	 * a simple logger to log
	 */
	private Logger logger;

	/**
	 * a logger factory to create other loggers if needed
	 */
	private LoggerFactory loggerFactory;

	/**
	 * pathExplorer is able to return input or output stream from a given file.
	 */
	private PathExplorer pathExplorer;

	static String fileSeparator = System.getProperty("file.separator");
	static String pathSeparator = System.getProperty("path.separator");

	/**
	 * the parser exception reached during the parsing of a file
	 */
	private Exception parserException = null;

	/**
	 * This structure contains Class and CompositeName objects.
	 * key = a JORM persistent definition file
	 * value = the associated MetaObject object.
	 */
	private Hashtable motable = null;

	/**
	 * if the genDep flag is set to true, the dependance will be generated
	 */
	private boolean genDep = false;

	private DTDResolver resolver = null;

	private List parsedMO = null;
	private List undefinedMO = null;

	/**
	 * A Map of String objects.
	 * key = a String object representing an  id value.
	 * value = the associated GenClassRef object.
	 */
	private Map idvalue2genclassref;

	// IMPLEMENTATION OF METHODS FROM THE Parser INTERFACE

	/**
	 * Main constructor for BasicDomParser. In order to compile XML files, it
	 * creates a DOMParser and initialzes it.
	 * @param dtdVerify true, the dtd must be used to validate xml files, else
	 *                  false
	 */
	public void init(boolean dtdVerify, ArrayList dtds) {
		parser = new DOMParser();
		parsedMO = new ArrayList();
		undefinedMO = new ArrayList();
		ParserErrorHandler parserErrorHandler = new ParserErrorHandler();
		parser.setErrorHandler(parserErrorHandler);
		motable = new Hashtable();
		mappingParsers = new HashMap();
		idvalue2genclassref = new HashMap();
		resolver = new DTDResolver(dtds, logger);
		try {
			parser.setFeature(
					"http://xml.org/sax/features/validation",
					dtdVerify);
		} catch (SAXNotRecognizedException e) {
			logger.log(BasicLevel.ERROR,
					   "there is a problem (exception) from the SAX module " +
					   "(SAXNotRecognized)");
		} catch (SAXNotSupportedException e) {
			logger.log(BasicLevel.ERROR,
					   "there is a problem (exception) from the SAX module " +
					   "(SAXNotSupported)");
		}
		try {
			parser.setProperty(
					"http://apache.org/xml/properties/internal/entity-resolver",
					resolver);
		} catch (Exception e) {
			logger.log(BasicLevel.ERROR,
					   "Impossible to assign the entity-resolver property", e);
		}
	}

	/**
	 * Adds a MappingParser object.
	 * @param  mapperName    a mapper name,
	 *         mappingParser a MappingParser object.
	 */
	public void addMappingParser(String mapperName,
								 MappingParser mappingParser) throws PException {
		if (mappingParser == null) {
			throw new PExceptionCompiler("<" + mapperName +
										 ">MappingParser has not been created.");
		}
		if (getLogger().isLoggable(BasicLevel.DEBUG)) {
			getLogger().log(BasicLevel.DEBUG,
							"Add a MappingParser (" + mapperName +
							") for the current Parser");
		}
		if (!mappingParsers.containsKey(mapperName)) {
			mappingParsers.put(mapperName, mappingParser);
		}
	}

	/**
	 * Returns a MappingParser object.
	 * @param  mapperName    a mapper name,
	 * @return a MappingParser object.
	 */
	public MappingParser getMappingParser(String mapperName) {
		MappingParser mappingParser = (MappingParser) mappingParsers.get(mapperName);
		return mappingParser;
	}

	/**
	 * Launches the parsing process. It reads and parses all the XML files
	 * and builds a schema of meta-objects for all of them.
	 * Before to use this method, the following object need to be set : Manager
	 * (with setMetaInfoManager), the PathExplorer (with setPathExplorer), the
	 * mapper name (with setMapperName), the mapping parser object (with setMappingParser).
	 *
	 * @param files     an iterator over the name the XML files to be parsed
	 */
	public Collection parse(Iterator files) throws PException {
		Collection res = new ArrayList();
		while (files.hasNext()) {
			String filename = (String) files.next();
			MetaObject mo = (MetaObject) motable.get(filename);
			if (mo == null) {
				mo = parse(filename);
				if (mo != null) {
					motable.put(filename, mo);
				}
			}
			res.add(mo);
		}
		while (!undefinedMO.isEmpty()) {
			MetaObject mo = (MetaObject) undefinedMO.remove(0);
			String fqname = null;
			if (mo instanceof Class) {
				fqname = ((Class) mo).getFQName();
			} else if (mo instanceof CompositeName) {
				fqname = ((CompositeName) mo).getFQName();
			} else {
				throw new PException("Unknown meta object: " + mo);
			}
			fqname = fqname.replace('.', fileSeparator.charAt(0)) + ".pd";
			parse(fqname);
			motable.put(fqname, mo);
		}
		return res;
	}

	/**
	 * Launches the parsing process.
	 * It reads and parses an XML file and builds a schema of meta-objects
	 * for it.
	 * Before to use this method, the following object need to be set : Manager
	 * (with setMetaInfoManager), the PathExplorer (with setPathExplorer), the
	 * mapper name (with setMapperName), the mapping parser object (with setMappingParser).
	 *
	 * @param file      the name of the XML file to be parsed
	 */
	public MetaObject parse(String file) throws PException {
		boolean debug = logger.isLoggable(BasicLevel.DEBUG);
		if (debug)
			logger.log(BasicLevel.DEBUG, "try to get input stream from <"
										 + file + ">");
		InputStream is = pathExplorer.getInputStream(file);
		if (is == null) {
			throw new PException("The file '" + file +
								 "' has not been found in the following classpath \""
								 + pathExplorer.getClassPath() + "\"");
		} else if (debug)
			logger.log(BasicLevel.DEBUG, "File " + file + " found");

		InputSource isource = new InputSource(is);
		if (isource.getByteStream() == null)
			throw new PException("The file '" + file +
								 "' has not been found in the following classpath \""
								 + pathExplorer.getClassPath() + "\"");

		((ParserErrorHandler) parser.getErrorHandler()).initialize();
		// parse the file inputsource
		logger.log(BasicLevel.INFO, "Parsing the file " + file + "...");
		try {
			parser.parse(isource);
		} catch (Exception e) {
			logger.log(BasicLevel.ERROR,
					   "there is a problem (exception) during the parsing of <"
					   + file + "> from the SAX module (SAXNotRecognized)", e);
			parserException = e;
		}
		ParserErrorHandler peh = (ParserErrorHandler) parser.getErrorHandler();
		if (peh.numErrors() != 0) {
			logger.log(BasicLevel.ERROR,
					   "the parser handler has encountered "
					   + peh.numErrors()
					   + " error(s) in [" + file + "], please fix it(them)");

			// display errors
			for (int i = 0; i < peh.numErrors(); i++) {
				logger.log(BasicLevel.ERROR, peh.getErrors().get(i));
			}
			return null;
		}
		if (debug)
			logger.log(BasicLevel.DEBUG, "Parsing done.");
		// Build the meta-information
		return (parser.getDocument().getDocumentElement() != null
				? build(parser.getDocument(), file)
				: null);
	}

	/**
	 * Assigns a meta-information manager to a parser in order from it to
	 * create the meta-information related to the XML files it parses.
	 *
	 * @param mim       the meta-information Manager to be associated to this parser
	 */
	public void setMetaInfoManager(Manager mim) {
		this.metaInfoManager = mim;
	}

	/**
	 * Assigns a PathExplorer object for locating files that have to be parsed.
	 *
	 * @param pathexpl  the PathExplorer to be used for file location
	 */
	public void setPathExplorer(PathExplorer pathexpl) {
		pathExplorer = pathexpl;
		if (pathExplorer == null) {
			pathExplorer = new DirJavaExplorer();
		}
		logger.log(BasicLevel.DEBUG, "jorm classpath=" + pathExplorer.getClassPath());
	}

	/**
	 * Enables or disables the generation of the dependances.
	 *
	 * @param gendep    true, the dependances are generated, else false
	 */
	public void setGenDep(boolean gendep) {
		genDep = gendep;
	}

	/**
	 * Allows to know if the dependances must be generated or not.
	 *
	 * @return  true, if the dependances are generated, else false
	 */
	public boolean isGenDep() {
		return genDep;
	}

	/**
	 * returns the exception which is reached during the parsing of a file
	 *
	 * @return  an Exception object
	 */
	public Exception getParserException() {
		return parserException;
	}

	/**
	 * allows to know the list of error the parser encounters.
	 *
	 * @return  a ParserErrorHandler object which countains all errors from
	 *          the xml parser
	 */
	public ParserErrorHandler getParserErrorHandler() {
		return (ParserErrorHandler) parser.getErrorHandler();
	}

	///////////////////////////////////////////////////////////////////
	//                      local methods
	///////////////////////////////////////////////////////////////////

	/**
	 * builds the meta-information into the meta information manager.
	 * @param doc       the document to read and move into meta information
	 * @param name      the name of the document (file name).
	 *
	 * <!ELEMENT jorm (package?, ((class, mapping*) | composite-name)) >
	 */
	private MetaObject build(Document doc, String name) throws PException {
		if (logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG,
					   "build the meta information for <" + name + "> description file");
		Package schema = null;
		Class clazz = null;
		CompositeName aname = null;
		NodeList nodes = doc.getDocumentElement().getChildNodes();
		for (int i = 0; i < nodes.getLength(); i++) {
			Node node = nodes.item(i);
			String nodeName = node.getNodeName();
			if (nodeName.equals("package")) {
				if (logger.isLoggable(BasicLevel.DEBUG)) {
					logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
					logger.log(BasicLevel.DEBUG, "package name =<" +
												 node.getFirstChild().getNodeValue() + ">");
				}
				schema = metaInfoManager.createPackage(node.getFirstChild().getNodeValue());
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
			} else if (nodeName.equals("class")) {
				String className = ((Element) node).getAttribute("name");
				// compare the name of the file with the name of the class
				int index = name.lastIndexOf(fileSeparator);
				int endIdx = name.lastIndexOf(".");
				if (endIdx <= index)
					endIdx = name.length();
				String sub = name.substring(index + 1, endIdx);
				if (!sub.equals(className))
					logger.log(BasicLevel.WARN, "Be careful, the name of the class" +
												" is different from the name of the file : " +
												sub + " and " + className +
												"(" + className + " is used)");

				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");

				String isAbstractString = ((Element) node).getAttribute("abstract").toLowerCase();
				if (schema == null) {
					schema = metaInfoManager.createPackage("");
				}
				clazz = schema.createClass(className);
				if (!parsedMO.contains(clazz)) {
					// The class has not been already loaded
					parsedMO.add(clazz);
					undefinedMO.remove(clazz);
					clazz.setAbstract(Boolean.getBoolean(isAbstractString));
					processClass(clazz, (Element) node);
				}
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
			} else if (nodeName.equals("mapping")) {
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
				String projectName = ((Element) node).getAttribute("project-name");
				if (clazz == null) {
					logger.log(BasicLevel.DEBUG, "the class definition is missing!");
					break;
				} else {
					if (logger.isLoggable(BasicLevel.DEBUG))
						logger.log(BasicLevel.DEBUG, "project-name =<" + projectName + ">");
					ClassProject classProject = clazz.createClassProject(projectName);
					parseMapping((Element) node, classProject);
				}
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
			} else if (nodeName.equals("composite-name")) {
				String CN_name = ((Element) node).getAttribute("name");
				// compare the name of the file with the name of the class
				int index = name.lastIndexOf(fileSeparator);
				String sub = name.substring(index + 1, name.length() - 3);

				if (sub.compareTo(CN_name) != 0)
					logger.log(BasicLevel.WARN, "Be careful, the name of the composite name" +
												" is different from the name of the file : " +
												name + " and " + CN_name +
												"(" + CN_name + " is used)");

				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");

				if (schema == null) {
					schema = metaInfoManager.createPackage("");
				}
				aname = schema.createCompositeName(CN_name);
				if (!parsedMO.contains(aname)) {
					parsedMO.add(aname);
					undefinedMO.remove(aname);
					// The composite name has not been already loaded
					processCompositeName(aname, (Element) node);
				}
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
				return aname;
			}
		}
		if (clazz != null)
			return clazz;
		return null;
	}

	private void processCompositeName(CompositeName compositeName,
									  Element compositeNameElem) {

		NodeList children = compositeNameElem.getChildNodes();
		for (int i = 0; i < children.getLength(); i++) {
			Node child = children.item(i);
			String childName = child.getNodeName();

			if (childName.equals("scalar-field")) {
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "begin =<" + childName + ">");

				processScalarField(compositeName, (Element) child);

				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "end   =<" + childName + ">");
			} else if (childName.equals("extension")) {
				// Get the compositename extension and add it to the list
				// of inherited compositenames
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "begin =<" + childName + ">");

				String CN_name = ((Element) child).getAttribute("Name");
				CompositeName parentCN = metaInfoManager.getCompositeName(CN_name);
				if (parentCN == null) {
					parentCN = metaInfoManager.createCompositeName(CN_name);
					undefinedMO.add(parentCN);
				}
				compositeName.addInheritedCompositeName(parentCN);
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "end   =<" + childName + ">");
			}
		}
	}


	/**
	 * Reads the subtree enrooted at class element and sets propreties of a
	 * Class object.
	 *
	 * <!ELEMENT class ((field | scalarfield)*, extension*, name-def*)>
	 * The Class object is created by the build() method.
	 *
	 * @param clazz             an instance of Class,
	 *        classElem     a class element
	 */
	private void processClass(Class clazz, Element classElem) throws PException {

		NodeList children = classElem.getChildNodes();

		for (int i = 0; i < children.getLength(); i++) {
			Node child = children.item(i);
			String nodeName = child.getNodeName();
			if (nodeName.equals("field")) {
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
				processField(clazz, (Element) child);
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
			} else if (nodeName.equals("scalar-field")) {
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
				processScalarField(clazz, (Element) child);
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
			} else if (nodeName.equals("extension")) {
				if (logger.isLoggable(BasicLevel.DEBUG)) {
					logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
					logger.log(BasicLevel.DEBUG, "name of the extension=<" +
												 child.getFirstChild().getNodeValue() + ">");
				}
				String fqcn = child.getFirstChild().getNodeValue();
				Class inheritedClass = metaInfoManager.getClass(fqcn);
				// First parse the inherited class
				if (inheritedClass == null) {
					/*c = metaInfoManager.createClass(fqcn);
					undefinedMO.add(c);
					*/
					fqcn = fqcn.replace('.', fileSeparator.charAt(0)) + ".pd";
					inheritedClass = (Class) parse(fqcn);
					motable.put(fqcn, inheritedClass);
				}
				clazz.addInheritedClass(inheritedClass);
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
			} else if (nodeName.equals("name-def")) {
				/* <!ATTLIST name-def
						   id ID #REQUIRED
						   name CDATA #IMPLIED>
				 */
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
				NameDef nameDef = clazz.createNameDef();
				processNameDef(nameDef, ((Element) child));
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");

			}
			/*else if (nodeName.equals("ClassMapping")) {
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");

				mappingParser.setMetaInfoManager(metaInfoManager);
				mappingParser.parseMapping((Element) child, clazz);

				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
			}*/
		}
	}

	/**
	 * Reads the subtree enrooted at scalar-field element and creates
	 * ScalarField objects.
	 *
	 * <!ELEMENT scalar-field (scalar-type, null-value?)>
	 *
	 * @param mo                      an instance of Class, or GenClassRef, or
	 *                            CompositeName
	 *        scalarFieldElem     a scalar-field element
	 */
	private void processScalarField(MetaObject mo, Element scalarFieldElem) {
		/* <!ATTLIST scalar-field
						   name CDATA #REQUIRED>
		 */
		String scalarFieldName = scalarFieldElem.getAttribute("name");
		NodeList children = scalarFieldElem.getChildNodes();
		ScalarField scalarField = null;
		for (int j = 0; j < children.getLength(); j++) {
			Node child = children.item(j);
			String childName = child.getNodeName();
			if (logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG, "begin =<" + childName + ">");
			if (childName.equals("scalar-type")) {
				String type = ((Element) child).getAttribute("type");
				if (mo instanceof Class) {
					scalarField = ((Class) mo).createHiddenField(
							scalarFieldName, getPType(type),
							Integer.parseInt(((Element) child).getAttribute("size")));
				} else if (mo instanceof GenClassRef) {
					scalarField = ((GenClassRef) mo).createHiddenField(
							scalarFieldName, getPType(type),
							Integer.parseInt(((Element) child).getAttribute("size")));
				} else if (mo instanceof CompositeName) {
					scalarField = ((CompositeName) mo).createCompositeNameField(
							scalarFieldName, getPType(type),
							Integer.parseInt(((Element) child).getAttribute("size")));
				}
			} else if (childName.equals("null-value")) {
				String value = ((Element) child).getAttribute("value");
				scalarField.setNullValue(value);
			}
			if (logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG, "end   =<" + childName + ">");
		}
	}

	/**
	 * Reads the subtree enrooted at field element and adds a couple of
	 * field names in the NameRef object.
	 *
	 * <!ELEMENT field (primitive-type | (reference-type, field-ref?)) >
	 * The field-ref element is read in processFieldRef() that is called
	 * from processReferenceType().
	 *
	 * @param clazz             an instance of Class,
	 *        fieldElem     a field element
	 */
	private void processField(Class clazz, Element fieldElem) {
		/* <!ATTLIST field
				   name CDATA #REQUIRED>
		 */
		String fieldName = fieldElem.getAttribute("name");
		NodeList children = fieldElem.getChildNodes();
		for (int i = 0; i < children.getLength(); i++) {
			Node child = children.item(i);
			String nodeName = child.getNodeName();
			if (nodeName.equals("primitive-type")) {
				/* <!ATTLIST primitive-type
						   id ID #REQUIRED
						   type (...|...) #REQUIRED
						   size CDATA #IMPLIED>
				*/
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
				clazz.createPrimitiveElement(fieldName,
											 getPType(((Element) child).getAttribute("type")),
											 Integer.parseInt(((Element) child).getAttribute("size")));
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
			} else if (nodeName.equals("class-ref")) {
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
				processClassRef(clazz, (Element) child);
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
			} else if (nodeName.equals("gen-class-ref")) {
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
				processGenClassRef(clazz, (Element) child);
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
			}
		}
	}

	/**
	 * Returns the referenced Class object.
	 * @param mo            an instance of Class or GenClassRef,
	 *       classRefElem   a class-ref element.
	 * @return a Class object.
	 */
	private Class getReferencedClass(MetaObject mo, Element classRefElem) {
		String fqClassName = classRefElem.getAttribute("class-name");
		Class aClass = metaInfoManager.getClass(fqClassName);
		if (aClass == null) {
			aClass = metaInfoManager.createClass(fqClassName);
			undefinedMO.add(aClass);
		}
		return aClass;
	}

	/**
	 * Reads the subtree enrooted at classRefElem and creates a ClassRef object
	 * and one or several NameDef objects.
	 *
	 * <!ELEMENT class-ref (name-def+) >
	 * <!ATTLIST class-name CDATA #REQUIRED
	 shared (TRUE|FALSE) "TRUE">
	 *
	 * @param mo                an instance of Class or GenClassRef,
	 *        classRefElem      a class-ref elemement.
	 */
	private void processClassRef(MetaObject mo, Element classRefElem) {
		Class referencedClass = getReferencedClass(mo, classRefElem);
		// exception to raise if referencedClass == null !!!
		ClassRef classRef = null;
		if (mo instanceof Class) {
			Node fieldNode = classRefElem.getParentNode();
			String fieldName = ((Element) fieldNode).getAttribute("name");
			classRef = ((Class) mo).createClassRef(fieldName, referencedClass);
		} else if (mo instanceof GenClassRef) {
			classRef = ((GenClassRef) mo).createClassRef(referencedClass);
		}
		NodeList children = classRefElem.getChildNodes();
		for (int i = 0; i < children.getLength(); i++) {
			Node child = children.item(i);
			String childName = child.getNodeName();
			if (childName.equals("name-def")) {
				/* <!ATTLIST name-def
						   id ID #REQUIRED
						   name CDATA #IMPLIED>
				 */
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "begin =<" + childName + ">");
				NameDef nameDef = classRef.createRefNameDef();
				processNameDef(nameDef, ((Element) child));
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "begin =<" + childName + ">");
			}
		}
	}

	/**
	 * Reads the subtree enrooted at field-ref and adds in a Reference object
	 * a field name and its reverse field.
	 *
	 * @param mo                an instance of Reference,
	 *        refTypeElem   a ref-type element
	 */
	private void processFieldRef(Reference mo, Element refTypeElem) {
		Node fieldRefNode = refTypeElem.getNextSibling();
		if (fieldRefNode != null) {
			/* <!ATTLIST field-ref
						   field-name CDATA #REQUIRED
			 */
			String fieldRefNodeName = fieldRefNode.getNodeName();
			if (logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG,
						   "begin =<" + fieldRefNodeName + ">");

			// mo.addFieldRef( fieldName, fieldRefName);
			if (logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG,
						   "end   =<" + fieldRefNodeName + ">");
		}
	}

	/**
	 * returns the PType of a given name. If the type does not exist,
	 * null is returned.
	 * @param name      the string representation of the name.
	 * @return  the corresponding PType of its name.
	 */
	private PType getPType(String name) {
		if (name.equals(PTypeSpace.BOOLEAN.getJormName()))
			return PTypeSpace.BOOLEAN;
		if (name.equals(PTypeSpace.OBJBOOLEAN.getJormName()))
			return PTypeSpace.OBJBOOLEAN;
		else if (name.equals(PTypeSpace.OBJCHAR.getJormName())
				|| name.equals("Character"))
			return PTypeSpace.OBJCHAR;
		else if (name.equals(PTypeSpace.CHAR.getJormName()))
			return PTypeSpace.CHAR;
		else if (name.equals(PTypeSpace.OBJBYTE.getJormName()))
			return PTypeSpace.OBJBYTE;
		else if (name.equals(PTypeSpace.BYTE.getJormName()))
			return PTypeSpace.BYTE;
		else if (name.equals(PTypeSpace.OBJSHORT.getJormName()))
			return PTypeSpace.OBJSHORT;
		else if (name.equals(PTypeSpace.SHORT.getJormName()))
			return PTypeSpace.SHORT;
		else if (name.equals(PTypeSpace.OBJINT.getJormName())
				|| name.equals("Integer"))
			return PTypeSpace.OBJINT;
		else if (name.equals(PTypeSpace.INT.getJormName()))
			return PTypeSpace.INT;
		else if (name.equals(PTypeSpace.OBJLONG.getJormName()))
			return PTypeSpace.OBJLONG;
		else if (name.equals(PTypeSpace.LONG.getJormName()))
			return PTypeSpace.LONG;
		else if (name.equals(PTypeSpace.OBJFLOAT.getJormName()))
			return PTypeSpace.OBJFLOAT;
		else if (name.equals(PTypeSpace.FLOAT.getJormName()))
			return PTypeSpace.FLOAT;
		else if (name.equals(PTypeSpace.OBJDOUBLE.getJormName()))
			return PTypeSpace.OBJDOUBLE;
		else if (name.equals(PTypeSpace.DOUBLE.getJormName()))
			return PTypeSpace.DOUBLE;
		else if (name.equals(PTypeSpace.STRING.getJormName())
				|| name.equals("String"))
			return PTypeSpace.STRING;
		else if (name.equals(PTypeSpace.DATE.getJormName())
				|| name.equals("Date"))
			return PTypeSpace.DATE;
		else if (name.equals(PTypeSpace.SERIALIZED.getJormName())
				|| name.equals("Serializable"))
			return PTypeSpace.SERIALIZED;
		else if (name.equals(PTypeSpace.BYTEARRAY.getJormName()))
			return PTypeSpace.BYTEARRAY;
		else if (name.equals(PTypeSpace.CHARARRAY.getJormName()))
			return PTypeSpace.CHARARRAY;
		else if (name.equals(PTypeSpace.BIGDECIMAL.getJormName()))
			return PTypeSpace.BIGDECIMAL;
		else {
			logger.log(BasicLevel.ERROR, "Invalid type name: " + name);
			return null;
		}
	}


	/**
	 * Reads the subtree enrooted at name-def element and:
	 * - sets system to true in the nameDef object, or
	 * - sets field name in the nameDef object, or
	 * - creates a NameRef object and possibly an hidden CompositeName object.
	 *
	 * <p> <!ELEMENT name-def (system | field-ref| composite-name-ref ) >
	 *
	 * @param nameDef           an instance of NameDef,
	 *        nameDefElem   a name-def element
	 */
	private void processNameDef(NameDef nameDef, Element nameDefElem) {
		String name = nameDefElem.getAttribute("name");
		nameDef.setName(name);
		NodeList children = nameDefElem.getChildNodes();
		for (int i = 0; i < children.getLength(); i++) {
			Node child = children.item(i);
			String childName = child.getNodeName();
			if (childName.equals("system")) {
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "begin =<" + childName + ">");
				nameDef.setSystem(true);
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "end   =<" + childName + ">");
			} else if (childName.equals("field-ref")) {
				/*  <!ATTLIST field-ref
							field-name CDATA #REQUIRED>
				 */
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "begin =<" + childName + ">");
				String fieldName = ((Element) child).getAttribute("field-name");
				nameDef.setFieldName(fieldName);
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "end   =<" + childName + ">");
			} else if (childName.equals("composite-name-ref")) {
				/*  <!ATTLIST composite-name-ref
							composite-name-name CDATA #REQUIRED>
				 */
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "begin =<" + childName + ">");
				String fqCN_Name =
						((Element) child).getAttribute("composite-name-name");
				if (logger.isLoggable(BasicLevel.DEBUG)) {
					logger.log(BasicLevel.DEBUG,
							   "composite-name-ref composite-name-name =<" + fqCN_Name + ">");
				}
				int lastindexofpoint = fqCN_Name.lastIndexOf('.');
				String CN_Name = null;
				Package CN_Schema = null;
				// search the schema of the referenced composite name
				if (lastindexofpoint != -1) {
					// the name of the schema is specified
					String schemaName =
							fqCN_Name.substring(0, fqCN_Name.lastIndexOf('.'));
					CN_Schema = metaInfoManager.createPackage(schemaName);
					CN_Name =
							fqCN_Name.substring(fqCN_Name.lastIndexOf('.') + 1,
												fqCN_Name.length());
				} else {
					// the name of the schema is not specified TJRS = 2 ?
					Object tmp = nameDef.getParent();
					// get the schema
					while (!(tmp instanceof Package)) {
						tmp = ((MetaObject) tmp).getParent();
					}
					CN_Schema = (Package) tmp;
					logger.log(BasicLevel.WARN, "Be careful, " +
												"the name of package (Package name) is not specified, " +
												"continue with " + CN_Schema.getName() + " schema.");
					CN_Name = fqCN_Name;
				}
				if (logger.isLoggable(BasicLevel.DEBUG)) {
					logger.log(BasicLevel.DEBUG,
							   "schema name =<" + CN_Schema.getName() + ">");
				}
				// try to get the class from the schema
				CompositeName aCompositeName = CN_Schema.getCompositeName(CN_Name);
				if (aCompositeName == null) {
					aCompositeName = CN_Schema.createCompositeName(CN_Name);
					undefinedMO.add(aCompositeName);
				}
				NameRef nameRef = nameDef.createNameRef(aCompositeName);
				processNameRef(nameRef, ((Element) child));
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "end   =<" + childName + ">");
			}
		}
	}

	/**
	 * Reads the subtree enrooted at composite-name-ref and adds a couple of
	 * field names in the NameRef object.
	 *
	 * <p> <!ELEMENT composite-name-ref (composite-name-field-projection+) >
	 *
	 * @param nameRef               an instance of NameRef,
	 *        compositeNameRefElem  a composite-name-ref element
	 */
	private void processNameRef(NameRef nameRef, Element compositeNameRefElem) {
		NodeList children = compositeNameRefElem.getChildNodes();
		for (int i = 0; i < children.getLength(); i++) {
			Node child = children.item(i);
			String nodeName = child.getNodeName();
			if (nodeName.equals("composite-name-field-projection")) {
				/*  <!ATTLIST composite-name-field-projection
							composite-name-field-name CDATA #REQUIRED
							class-field-name CDATA #REQUIRED>
				 */
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "begin =<" + nodeName + ">");
				String CN_fieldName =
						((Element) child).getAttribute("composite-name-field-name");
				String classfieldName =
						((Element) child).getAttribute("class-field-name");
				nameRef.addProjection(CN_fieldName, classfieldName);
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "end   =<" + nodeName + ">");
			}
		}
	}

	/**
	 * Reads the subtree enrooted at index and adds a couple of
	 * field names in the NameRef object.
	 *
	 * <p> <!ELEMENT index (field-ref+) >
	 *
	 * @param index          an instance of Index,
	 *        indexElem  an index element
	 */
/*
    private void processIndex(Index index, Element indexElem) {
        NodeList children = indexElem.getChildNodes();
        for (int i = 0; i < children.getLength(); i++) {
            Node child = children.item(i);
            String childName = child.getNodeName();
            if (childName.equals("field-ref")) {
               //  <!ATTLIST field-ref
               //            field-name CDATA #REQUIRED>

                if (logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "begin =<" + childName + ">");
                String fieldName = ((Element) child).getAttribute("field-name");
                index.addFieldName(fieldName);
                if (logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "end   =<" + childName + ">");
            }
        }
    }
*/

	/**
	 * Reads the subtree enrooted at gen-class-ref element and:
	 * - parses a gen-class element
	 *
	 * <!ELEMENT gen-class-ref   (gen-class, name-def+)>
	 *
	 * @param mo                an instance of Class or GenClassRef,
	 *        genClassRefElem   a gen-class-ref element.
	 */
	private void processGenClassRef(MetaObject mo, Element genClassRefElem) {
		NodeList children = genClassRefElem.getChildNodes();
		GenClassRef genClassRef = null;
		for (int i = 0; i < children.getLength(); i++) {
			Node child = children.item(i);
			String childName = child.getNodeName();
			if (childName.equals("gen-class")) {
				/* <!ATTLIST gen-class
						   id ID #REQUIRED
						   gen-class-name CDATA #REQUIRED>
				 */
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "begin =<" + childName + ">");
				String genClassName =
						((Element) child).getAttribute("gen-class-name");
				if (mo instanceof Class) {
					Node fieldNode = genClassRefElem.getParentNode();
					String fieldName = ((Element) fieldNode).getAttribute("name");
					genClassRef =
							((Class) mo).createGenClassRef(fieldName, genClassName);
					/* idvalue2genclassref.put(fieldName, genClassRef);
					if (logger.isLoggable(BasicLevel.DEBUG)) {
						logger.log(BasicLevel.DEBUG, "mo=class, gen class id =<" + fieldName + ">");
					}
					*/
				} else if (mo instanceof GenClassRef) {
					genClassRef =
							((GenClassRef) mo).createGenClassRef(genClassName);
				}
				idvalue2genclassref.put(genClassRef.getGenClassId(), genClassRef);
				if (logger.isLoggable(BasicLevel.DEBUG)) {
					logger.log(BasicLevel.DEBUG,
							   "genclass id =<" + genClassRef.getGenClassId() + ">");
				}
				processGenClass(genClassRef, (Element) child);
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "end   =<" + childName + ">");
			} else if (childName.equals("name-def")) {
				/* <!ATTLIST name-def
						   id ID #REQUIRED
						   name CDATA #IMPLIED>
				 */
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "begin =<" + childName + ">");
				NameDef nameDef = genClassRef.createRefNameDef();
				processNameDef(nameDef, ((Element) child));
				if (logger.isLoggable(BasicLevel.DEBUG))
					logger.log(BasicLevel.DEBUG, "end   =<" + childName + ">");
			}
		}
	}

	/**
	 * Reads the subtree enrooted at gen-class element and:
	 * - parses scalar-field element(s),
	 * - creates NameDef object(s),
	 * - creates an Index object,
	 * - creates a PrimitiveType object,
	 * - parses a reference-type element
	 *
	 * <!ELEMENT gen-class
	 *       (scalarfield*, name-def+, index?, (primitive-type | reference-type)>
	 *
	 * @param genClassRef          an instance of GenClassRef,
	 *        genClassElem     a gen-class element
	 */
	private void processGenClass(GenClassRef genClassRef,
								 Element genClassElem) {
		NodeList children = genClassElem.getChildNodes();
		for (int i = 0; i < children.getLength(); i++) {
			Node child = children.item(i);
			String childName = child.getNodeName();
			if (logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG, "begin =<" + childName + ">");
			if (childName.equals("scalar-field")) {
				/* <!ATTLIST scalar-field
						   name CDATA #REQUIRED>
				 */
				processScalarField(genClassRef, (Element) child);
			} else if (childName.equals("name-def")) {
				/* <!ATTLIST name-def
						   id ID #REQUIRED
						   name CDATA #IMPLIED>
				 */
				NameDef nameDef = genClassRef.createIdNameDef();
				processNameDef(nameDef, ((Element) child));
			} else if (childName.equals("index")) {
				processIndex(genClassRef, ((Element) child));
			} else if (childName.equals("primitive-type")) {
				/* <!ATTLIST primitive-type
						   id ID #REQUIRED
						   type (...|...) #REQUIRED
						   size CDATA #IMPLIED>
				*/
				genClassRef.createPrimitiveElement(
						getPType(((Element) child).getAttribute("type")),
						Integer.parseInt(((Element) child).getAttribute("size")));
			} else if (childName.equals("class-ref")) {
				processClassRef(genClassRef, (Element) child);
			} else if (childName.equals("gen-class-ref")) {
				processClassRef(genClassRef, (Element) child);
			}
			if (logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG, "end   =<" + childName + ">");
		}
	}

	private void processIndex(GenClassRef genClassRef,
							  Element indexElem) {
		NodeList children = indexElem.getChildNodes();
		for (int i = 0; i < children.getLength(); i++) {
			Node child = children.item(i);
			String childName = child.getNodeName();
			if (logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG, "begin =<" + childName + ">");
			if (childName.equals("field-ref")) {
				genClassRef.addIndexField(
						((Element) child).getAttribute("field-name"));
			}
			if (logger.isLoggable(BasicLevel.DEBUG))
				logger.log(BasicLevel.DEBUG, "end   =<" + childName + ">");
		}
	}

	/**
	 * Parses the mapping part of the XML file and builds the corresponding
	 * meta-information for a given Class object.
	 * Manager, PathExplorer, MapperName and MappingName must be set before the
	 * call of this method.
	 *
	 * @param mappingElem           the node to treat
	 */
	public void parseMapping(Element mappingElem, ClassProject classProject)
			throws PException {
		NodeList nodes = mappingElem.getChildNodes();
		MappingParser mp = null;
		Mapping mapping = null;
		String mapperName = null;
		for (int i = 0; i < nodes.getLength(); i++) {
			Node child = nodes.item(i);
			String childName = child.getNodeName();
			if (logger.isLoggable(BasicLevel.DEBUG)) {
				logger.log(BasicLevel.DEBUG, "begin =<" + childName + ">");
			}
			int pos = childName.indexOf("-");
			if (pos == -1) {
				continue;
			}
			mapperName = childName.substring(0, pos);
			mp = getMappingParser(mapperName);
			if (mp == null) {
				if (logger.isLoggable(BasicLevel.DEBUG)) {
					logger.log(BasicLevel.DEBUG,
							   "cannot get the mapping parser!");
				}
				break;
			}
			mp.setPathExplorer(pathExplorer);
			((Loggable) mp).setLogger(logger);
			mp.setMetaInfoManager(metaInfoManager);
			mp.setidvalue2genclassref(idvalue2genclassref);
			mp.setCurrentClass((Class) classProject.getParent());
			mapping = classProject.createMapping(mapperName);
			if (logger.isLoggable(BasicLevel.DEBUG)) {
				logger.log(BasicLevel.DEBUG,
						   "current class =<" + ((Class) classProject.getParent()).getName() + ">");
				logger.log(BasicLevel.DEBUG,
						   "current project =<" + classProject.getProjectName() + ">");
				logger.log(BasicLevel.DEBUG,
						   "current mapping =<" + mapping.getMapperName() + ">");
			}
			mapping.setLogger(logger);
			mp.parseMapping((Element) child, mapping);
			if (logger.isLoggable(BasicLevel.DEBUG)) {
				logger.log(BasicLevel.DEBUG, "end   =<" + childName + ">");
			}
		}
	}


	// IMPLEMENTATION OF METHODS FROM THE Loggable INTERFACE

	/**
	 * Defines a logger object.
	 *
	 * @param logger    the logger object
	 */
	public void setLogger(Logger logger) {
		this.logger = logger;
		if (resolver != null)
			resolver.setLogger(logger);
	}

	/**
	 * Defines the logger factory to obtain new loggers.
	 *
	 * @param loggerfactory     The LoggerFactory object to obtain a logger object
	 */
	public void setLoggerFactory(LoggerFactory loggerfactory) {
		this.loggerFactory = loggerfactory;
	}

	public Logger getLogger() {
		return logger;
	}

	public LoggerFactory getLoggerFactory() {
		return loggerFactory;
	}
}

